/*
 * GEMClip, a Clipboard Manager for FreeMiNT and MagiC.
 *
 * Copyright (C) 2011 Jean-Franois Lemaire (jflemaire@skynet.be)
 *
 * This file is part of GEMClip.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <mintbind.h>

#include "general.h"

#include "about.h"
#include "protocols.h"
#include "utils.h"

#define MARGIN 5

static TEXT *load_license   (void);
static void  scroll_license (TEXT *text, WINDOW *window, short message, OBJECT *tree);

/*
 * Create and open the "About GEMClip dialog.
 * Return the opened status variable TRUE or FALSE on error.
 */
int open_about_form(GEMCLIP *manager, WINDOW *window)
{
	int attribs  = NAME | CLOSER | MOVER | SMALLER;
	OBJECT *tree = get_gem_tree(ABOUT);
	GRECT rect;
	GRECT wrect;

	/* Prepare the build string in the form: "Compiled on XXX 00 0000 with GCC 0.00.0". */
	*tree[ABOUT_BUILD].ob_spec.tedinfo->te_ptext = '\0';
	snprintf(tree[ABOUT_BUILD].ob_spec.tedinfo->te_ptext,
			 tree[ABOUT_BUILD].ob_spec.tedinfo->te_txtlen,
			 "%s %s %s %d.%d.%d",
			 "Compiled on",
			 __DATE__,
			 "with GCC",
			 __GNUC__,
			 __GNUC_MINOR__,
			 __GNUC_PATCHLEVEL__);

	/* Calculate the coordinates to open the dialog in the centre of the screen. */
	wind_get_grect(0, WF_WORKXYWH, &wrect);

	rect.g_x = wrect.g_x + ((wrect.g_w - tree->ob_width)  / 2);
	rect.g_y = wrect.g_y + ((wrect.g_h - tree->ob_height) / 2);
	rect.g_w = tree->ob_width;
	rect.g_h = tree->ob_height;


	/* Open and process the license file. */
	manager->text = load_license();

	if ((window = init_window(window)) != NULL)
	{
		/* Create and open the window. */
		if ((window->handle = wind_create_grect(attribs, &rect)) != 0)
		{
			wind_set_str(window->handle, WF_NAME, get_gstring(STRINGS, STRINGS_ABOUT_NAME));
			strcpy(window->info, get_gstring(STRINGS, STRINGS_ABOUT_NAME)); // Default infoline.
			vst_font(window->wks, 1);
			vst_point(window->wks, 8, &window->wchar, &window->hchar, &window->wcell, &window->hcell);

			/* Get the full area of the window and position the root box. */
			wind_calc_grect(WC_BORDER, attribs, &rect, &wrect);
			tree->ob_x = rect.g_x;
			tree->ob_y = rect.g_y;

			/* Open About window. */
			if (wind_open_grect(window->handle, &wrect) != 0)
			{
				window->open = TRUE;
				vsf_color(window->wks, G_WHITE);
				vst_color(window->wks, G_BLACK);
			}
			else
			{
				form_error(FERR_NOHANDLES);
				window->open = FALSE;
			}
		} else
		{
			form_error(FERR_NOHANDLES);
			window->open = FALSE;
		}
	} else {
		window->open = FALSE;
	}
	return window->open;
}

/*
 * Process button and keyboard actions on the About dialog. We evaluate first
 * the Help and Undo keys then the other objects.
 */
void handle_about_form(GEMCLIP *manager, WINDOW *window, OBJECT *tree, EVMULT_OUT *evmult_out, short object, USERCONF *conf, BUBBLES *bubbles)
{
	if (object == 0) // Key pressed.
	{
		switch (evmult_out->emo_kreturn)
		{
		case SC_UPARROW:
			scroll_license(manager->text, window, WA_UPLINE, tree);
			break;
		case SC_DNARROW:
			scroll_license(manager->text, window, WA_DNLINE, tree);
			break;
		case SC_HELP:
			if (manager->aeslang == AESLANG_GERMAN) {
				call_help(window->handle, tree, object, manager->avserver, "*:\\GEMCLIPD.HYP ", "", manager->system);
			}
			else {
				call_help(window->handle, tree, object, manager->avserver, "*:\\GEMCLIPE.HYP ", "", manager->system);
			}
			break;
		case SC_UNDO:
			tree[object].ob_state &= ~OS_SELECTED;
			send_local_message(WM_CLOSED, window->handle, 0, 0, 0, 0);
			break;
		}
	}
	else if ( ! (evmult_out->emo_mbutton & RIGHT_BUTTON))
	{
		int pgheight = tree[ABOUT_GPL].ob_height / window->hcell - 11; // Remove interline.

		switch (object) // Button clicked.
		{
		case ABOUT_UPARROW:
			if (manager->text != NULL)
			{
				/* Check that we are not already at the top, to avoid flicker. */
				if ((evmult_out->emo_mclicks == 2) && (manager->text->lineidx != 0))
				{
					manager->text->lineidx = 0;
					scroll_license(manager->text, window, WA_UPPAGE, tree);
				}
				else if (manager->text->lineidx > 0)
				{
					manager->text->lineidx--;
					scroll_license(manager->text, window, WA_UPLINE, tree);
				}
			}
			break;
		case ABOUT_DNARROW:
			if (manager->text != NULL)
			{
				/* Check that we are not already at the bottom, to avoid flicker. */
				if ((evmult_out->emo_mclicks == 2) && (manager->text->lineidx != manager->text->nolines - 1))
				{
					manager->text->lineidx = manager->text->nolines - pgheight;
					scroll_license(manager->text, window, WA_DNPAGE, tree);
				}
				else {
					scroll_license(manager->text, window, WA_DNLINE, tree);
				}
			}
			break;
		case ABOUT_CLOSE:
			tree[object].ob_state &= ~OS_SELECTED;
			send_local_message(WM_CLOSED, window->handle, 0, 0, 0, 0);
			break;
		case ABOUT_WEB:
			{
				short state = 0;

				send_local_message(MN_SELECTED, MENUWIN_HELP, MENUWIN_WEB, 0, 0, 0);
				state  =  OS_OUTLINED;
				state &= ~OS_SELECTED;
				change_state(window->handle, tree, object, state);
			}
			break;
		case ABOUT_HELP:
			if (manager->aeslang == AESLANG_GERMAN) {
				call_help(window->handle, tree, object, manager->avserver, "*:\\GEMCLIPD.HYP ", "", manager->system);
			}
			else {
				call_help(window->handle, tree, object, manager->avserver, "*:\\GEMCLIPE.HYP ", "", manager->system);
			}
			break;
		}
	}
	else
	{
		switch (object)
		{
		case ABOUT_CLOSE:
			display_bubble_mouse(bubbles, ABTOKAY, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case ABOUT_WEB:
			display_bubble_mouse(bubbles, ABTSITE, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case ABOUT_HELP:
			display_bubble_mouse(bubbles, ABTHELP, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		}
	}
}

/*
 * Load the license file into memory and set pointers to each line.
 */
static TEXT *load_license(void)
{
	int idx    = 0;
	long size  = 0;
	long nb    = 0;
	char *p    = NULL;
	FILE *fp   = NULL;
	TEXT *text = NULL;

	if ((text = xmalloc(sizeof *text)) != NULL)
	{
		char *env          = NULL;
		char path[LBUFFER] = {0};

		shel_envrn(&env, "ACCPATH=");
		
		if (env != NULL) {
			snprintf(path, sizeof path, "%s%s%s", env, "\\", "GEMCLIP\\COPYING");
		}

		if (((fp = fopen("GEMCLIP\\COPYING", "r")) != NULL) ||
		    ((fp = fopen("gemclip\\copying", "r")) != NULL) ||
			((fp = fopen("COPYING", "r"))          != NULL) ||
			((fp = fopen("copying", "r"))          != NULL) ||
			((fp = fopen(path, "r"))               != NULL))
		{
			if ((fseek(fp, 0, SEEK_END)) != -1)
			{
				if ((size = ftell(fp)) != -1)
				{
					if ((fseek(fp, 0, SEEK_SET) != -1))
					{
						if ((text->buffer = xmalloc(size + 1)) != NULL)
						{
							nb = fread(text->buffer, sizeof (char), size, fp);
							text->buffer[nb] = '\0';
		
							/* Count number of lines in the file. */
							p = text->buffer;
		
							for (text->nolines = 1; *p; p++)
							{
								if (*p == '\n') {
									text->nolines++;
								}
							}
		
							/* Initialize first line index. */
							text->lineidx = 0;
		
							/* Allocate memory for all pointers to char *, set each pointer to a new line. */
							if ((text->line = xmalloc(text->nolines * sizeof *text->line)) != NULL)
							{
								text->line[0] = text->buffer;
								p = text->buffer;
		
								for (idx = 0; *p; p++)
								{
									if (*p == '\n')
									{
										text->line[++idx] = p + 1;
										*p = '\0';
									}
								}
								fclose(fp); // Don't forget to close the file.
							}
							else
							{
								form_alert(1, FA_ERROR "[Error malloc'ing for|'*text->line'.][OK]");
								fclose(fp);
								free(text);
								free(p); // Points to text->buffer.
								text = NULL;
							}
						}
						else
						{
							form_alert(1, FA_ERROR "[Error malloc'ing for|'*text->buffer'.][OK]");
							fclose(fp);
							free(text);
							text = NULL;
						}
					}
					else
					{
						form_alert_string(1, FA_ERROR, "Error:", strerror(errno));
						fclose(fp);
						free(text);
						text = NULL;
					}
				}
				else
				{
					form_alert_string(1, FA_ERROR, "Error:", strerror(errno));
					fclose(fp);
					free(text);
					text = NULL;
				}
			}
			else
			{
				form_alert_string(1, FA_ERROR, "Error:", strerror(errno));
				fclose(fp);
				free(text);
				text = NULL;
			}
		}
		else
		{
			form_alert(1, FA_ERROR "[Error opening file in|load_license().][OK]");
			free(text);
			text = NULL;
		}
	}
	else {
		form_alert(1, FA_ERROR "[Error malloc'ing for|'*text'.][OK]");
	}
	return text;
}

/*
 * Redraw the license inside the About window dialog.
 */
void draw_license(WINDOW *window, TEXT *text, OBJECT *tree)
{
	short xpos	 = 0;
	short ypos	 = 0;
	short pxy[4] = {0, 0, 0, 0};
	int nline	 = 0;
	GRECT rect, rd;

	if (text != NULL) // Has the license been loaded into memory?
	{
		nline = text->lineidx;
		objc_offset(tree, ABOUT_GPL, &rd.g_x, &rd.g_y);
		rd.g_w = tree[ABOUT_GPL].ob_width - window->wchar; // Retract a margin and the scrollbar width.
		rd.g_h = tree[ABOUT_GPL].ob_height - MARGIN;
		xpos += rd.g_x + MARGIN; // Add a left margin.
		ypos += rd.g_y + MARGIN; // Add a top margin.

		graf_mouse(M_OFF, NULL);
		wind_update(BEG_UPDATE);
		wind_get_grect(window->handle, WF_FIRSTXYWH, &rect); // Get first rect.

		/* Walk the rectangle list and redraw the dirtied area. */
		while (rect.g_w && rect.g_h)
		{
			if (rc_intersect(&rd, &rect) != 0)
			{
				grect_to_array(&rect, pxy);
				vs_clip(window->wks, TRUE, pxy);
				v_bar(window->wks, pxy);

				/* Draw to the bottom of the clipped area - 2 lines (to compensate
				for the ypos incrementation below). */
				while ((text->lineidx < text->nolines) && (ypos < (pxy[3] - (window->hcell + 4 * 2))))
				{
					ypos += window->hcell + 4; // Increase interline.
					v_ftext(window->wks, xpos, ypos, (const char *)text->line[nline]);
					nline++;
				}

				ypos += window->hcell + 4;
				v_ftext(window->wks, xpos, ypos, (const char *)text->line[nline]);
				vs_clip(window->wks, FALSE, pxy);
			}
			wind_get_grect(window->wks, WF_NEXTXYWH, &rect); // Get next rect.
		}
		wind_update(END_UPDATE);
		graf_mouse(M_ON, NULL);
	}
}

/*
 * Handle WM_ARROWED messages in the About dialog.
 */
static void scroll_license(TEXT *text, WINDOW *window, short message, OBJECT *tree)
{
	short pxy[8]  = {0};
	short bar[4]  = {0};
	short ypos	  = 0;
	int nline	  = 0;
	int remainln  = 0;
	int pgheight  = 0;
	MFDB src	  = {NULL, 0, 0, 0, 0, 0, 0, 0, 0};
	MFDB dst	  = {NULL, 0, 0, 0, 0, 0, 0, 0, 0};
	GRECT wrect   = {0, 0, 0, 0};
	GRECT rect	  = {0, 0, 0, 0};

	if (text != NULL)
	{
		/* Set the font then calculate the height (in lines of text) of the answer area
		and the number of lines remaining to display, minus the page height. */
		pgheight = tree[ABOUT_GPL].ob_height / window->hcell - 11; // Remove interline.
		remainln = text->nolines - pgheight - text->lineidx;
		objc_offset(tree, ABOUT_GPL, &wrect.g_x, &wrect.g_y);
		wrect.g_w = tree[ABOUT_GPL].ob_width - window->wchar;
		wrect.g_h = tree[ABOUT_GPL].ob_height - 1; // Compensate a bit.

		pxy[0] = pxy[4] = bar[0] = wrect.g_x + MARGIN;
		pxy[2] = pxy[6] = bar[2] = wrect.g_x + wrect.g_w;

		switch (message)
		{
		case WA_UPPAGE:
		case WA_DNPAGE:
			draw_license(window, text, tree);
			break;
		case WA_UPLINE:
			nline = text->lineidx;

			pxy[1] = wrect.g_y + MARGIN;
			pxy[3] = wrect.g_y + wrect.g_h - MARGIN - window->hcell - 4;
			pxy[5] = wrect.g_y + MARGIN + window->hcell + 4;
			pxy[7] = wrect.g_y + wrect.g_h - 4;

			bar[1] = pxy[1];
			bar[3] = pxy[5];

			ypos = wrect.g_y + window->hcell + MARGIN;
			break;
		case WA_DNLINE:
			if ((text->lineidx + pgheight) < text->nolines)
			{
				nline = text->lineidx + pgheight;
				text->lineidx++;

				pxy[1] = wrect.g_y + MARGIN + window->hcell + 4 + 1;
				pxy[3] = wrect.g_y + wrect.g_h;
				pxy[5] = wrect.g_y + MARGIN - 1;
				pxy[7] = wrect.g_y + wrect.g_h - (window->hcell * 2) - 4 - 1;

				bar[1] = pxy[7];
				bar[3] = pxy[3];

				ypos = wrect.g_y + (pgheight * (window->hcell + 4)) + 4;
			}
			break;
		}

		graf_mouse(M_OFF, NULL);
		wind_update(BEG_UPDATE);
		wind_get_grect(window->handle, WF_FIRSTXYWH, &rect);

		while (rect.g_w && rect.g_h)
		{
			if (rc_intersect(&wrect, &rect) != 0)
			{
				vro_cpyfm(window->wks, S_ONLY, pxy, &src, &dst);
				v_bar(window->wks, bar); // Erase bottom line.
				v_ftext(window->wks, (wrect.g_x + MARGIN), ypos, (const char *)text->line[nline]);
			}

			wind_get_grect(window->handle, WF_NEXTXYWH, &rect);
		}

		wind_update(END_UPDATE);
		graf_mouse(M_ON, NULL);
	}
}

/*
 * Free memory used for the display of the license.
 */
void free_license(TEXT *text)
{
	if (text != NULL)
	{
		free(text->line);
		free(text->buffer);
		free(text);
	}
}
