/*
 * GEMClip, a Clipboard Manager for FreeMiNT and MagiC.
 *
 * Copyright (C) 2011 Jean-Franois Lemaire (jflemaire@skynet.be)
 *
 * This file is part of GEMClip.
 *
 * This program is free software; redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <ctype.h>
#include <mintbind.h>
#include <mint/errno.h>
#include <mint/ostruct.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include "general.h"

#include "clipbrd.h"
#include "clips.h"
#include "manager.h"
#include "protocols.h"
#include "sort.h"
#include "utils.h"

typedef struct xattr
{
   uint16_t   mode; 			 /* File mode, statements for */
								 /* - possible filetypes	  */
								 /* - special bits			  */
								 /* - access rights 		  */
   int32_t	  index;			 /* File number 			  */
   uint16_t   dev;				 /* Device number			  */
   uint16_t   rdev; 			 /* Actual device			  */
								 /* (e.g. for BIOS files)	  */
								 /* in MagiC however reserved */
   uint16_t   nlink;			 /* Number of links 		  */
   uint16_t   uid;				 /* User number 			  */
   uint16_t   gid;				 /* Group number			  */
   int32_t	  size; 			 /* File length 			  */
   int32_t	  blksize, nblocks;  /* Blocksize/occupied blocks */
   uint16_t   mtime, mdate; 	 /* Date of last modification */
   uint16_t   atime, adate; 	 /* Date of the last access   */
   uint16_t   ctime, cdate; 	 /* Creation date			  */
   uint16_t   attr; 			 /* TOS file attributes 	  */
   uint16_t   reserved2;		 /* Reserviert				  */
   int32_t	  reserved3[2]; 	 /* Reserviert				  */
} XATTR;

static void reorder_clips    (long clips[], int nclips, const char *clipbrd);
static int  read_directory   (int dirhdl, XATTR *xattr, char buffer[], size_t size, GEMCLIP *manager, USERCONF *conf);
static void read_preview     (GEMCLIP *manager, char buffer[], size_t size, XATTR *xattr);
static void write_preview    (CLIPS *clip, long size, short txtlen, const char *buffer);
static void rename_backwards (const char* clipbrd, int nclips, USERCONF *conf);

/*
 * Overview of how GEMClip Processes the Clipboard
 * -----------------------------------------------
 * When GEMClip starts, either at boot time or after, if the clipboard directory
 * contains a SCRAP.TXT file and nothing else (for example, if this is the 1st time
 * the accessory is executed or if the 'Empty clipboard during boot' option is set
 * in the Preferences) this file is backed up into the file CLIP00. Nothing is
 * displayed in the manager listbox except for the preview of SCRAP.TXT, since
 * there is no clips to display yet.
 *      
 * Then GEMClip polls every second the CLIPBRD directory and updates the clips
 * history, either if clips were deleted or SCRAP.TXT was modified.
 *
 * Note that a GEM application is required to broadcast the SC_CHANGED message when
 * it has modified SCRAP.TXT; therefore, GEMClip waits for this message and
 * processes the clipboard directory accordingly. Unfortunately, many applications
 * do not send this message... Hence the need for polling the clipboard.
 *
 * So, each time SCRAP.TXT is modified (an application has written data into it),
 * CLIP00 is renamed into CLIP01 and displayed as the first clip in the Manager
 * list box and SCRAP.TXT is again backed up into CLIP00. The next time SCRAP.TXT
 * is modified, CLIP01 is renamed into CLIP02 and CLIP00 into CLIP01 and SCRAP.TXT
 * is again backed up into CLIP00 and then it goes on...
 *
 * CLIP00 is only used as a backup of SCRAP.TXT. The 1st clip displayed in the
 * Manager list box (in standard sorting mode) corresponds to CLIP01.
 */ 

/*
 * This function is called each time a change is made on the clips with the Manager listbox or when
 * scrap.txt was modified (therefore we need to update the clips listbox). There may not be any
 * SCRAP.TXT file.
 */
int update_manager(int scraphdl, int dirhdl, GEMCLIP *manager, USERCONF *conf)
{
	int modified = FALSE;
	_DATETIME time;

	/* Was scrap.txt modified? Always true the 1st time the function is called. */
	if (scraphdl >= 0) {
		Fdatime(&time, scraphdl, 0);
	}

	if ((time.hour   != manager->clips[SCRAP]->time.hour)   ||
	    (time.minute != manager->clips[SCRAP]->time.minute) ||
	    (time.second != manager->clips[SCRAP]->time.second) ||
	    (time.year   != manager->clips[SCRAP]->time.year)   ||
	    (time.month  != manager->clips[SCRAP]->time.month)  ||
	    (time.day	 != manager->clips[SCRAP]->time.day))
	{
		char path[LBUFFER] = {0};
		char buffer[(manager->tree[MANAGER_PREV1].ob_spec.tedinfo->te_txtlen + 1) * 6];
		char *content  = NULL; // Contents of scrap.txt.
		short txtlen   = manager->tree[MANAGER_CLIPBOARD].ob_spec.tedinfo->te_txtlen;
		int clipn      = 0;
		int filehdl	   = 0;
		int nclips 	   = -1;
		int del        = FALSE;
		int firstrun   = FALSE;
		int empty      = FALSE;
		long clips[99] = {0}; // 1-99.
		long xr		   = 0;
		_DISKINFO dinfo;
		XATTR xattr;

		memset(buffer, 0, sizeof buffer);

		/* Distinguish between the 1st call of the function after launch and the following calls. */
		if (manager->clips[SCRAP]->time.year == 0) {
			firstrun = TRUE;
		}

		/* Update top and bottom SCRAP.TXT preview. Keep count of the size needed for backuping
		   the file below. We skip empty scraps (not at least 2 bytes long). */
		if (scraphdl >= 0)
		{
			if ((manager->clips[SCRAP]->size = Fseek(0, scraphdl, SEEK_END)) >= 2)
			{
				if ((content = xmalloc(manager->clips[SCRAP]->size)) != NULL) // Will be freed below.
				{
					if (Fseek(0, scraphdl, SEEK_SET) >= 0) // Rewind file pointer.
					{
						if (Fread(scraphdl, manager->clips[SCRAP]->size, content) > 0)
						{
							char *p   = NULL;
							long size = 0;

							content[manager->clips[SCRAP]->size] = '\0';
							p = content;

							/* If size > manager->clips[SCRAP]->size, clip contains only whitespace. */
							while (isspace(*p))
							{
								p++;
								size++;
							}

							if (size < manager->clips[SCRAP]->size)
							{
								modified = TRUE;
								manager->clips[SCRAP]->time = time;
								snprintf(buffer, sizeof buffer, "%s", content);
								remove_whitespace(buffer);
								write_preview(manager->clips[SCRAP], manager->clips[SCRAP]->size, txtlen, buffer);
								strcpy(manager->tree[MANAGER_CLIPBOARD].ob_spec.tedinfo->te_ptext,
									   manager->clips[SCRAP]->preview);
								update_bottom_preview(manager->tree, buffer);
							}
							else
							{
								manager->clips[SCRAP]->time = time;
								empty = TRUE;
								modified = FALSE;
							}
						}
					}
				}
			}
			else
			{
				manager->clips[SCRAP]->time = time;
				empty = TRUE;
				modified = FALSE;

				for (int i = MANAGER_PREV1; i <= MANAGER_PREV6; i++) {
					*manager->tree[i].ob_spec.tedinfo->te_ptext = '\0';
				}
			}
		}
		else // No SCRAP.TXT. Try to recreate it from CLI00.
		{
			char *copy    = NULL;
			int copyhdl   = 0;
			long copysize = 0;

			snprintf(path, sizeof path, "%s%s", manager->clipbrd, "clip00");

			if ((filehdl = Fopen(path, 0)) >= 0)
			{
				copysize = Fseek(0, filehdl, SEEK_END);

				if ((copy = xmalloc(copysize)) != NULL)
				{
					Fseek(0, filehdl, SEEK_SET);
					Fread(filehdl, copysize, copy);

					snprintf(path, sizeof path, "%s%s", manager->clipbrd, "scrap.txt");

					if ((copyhdl = Fcreate(path, 0)) >= 0)
					{
						Fwrite(copyhdl, copysize, copy);
						Fclose(copyhdl);
						av_protokoll_path_update(manager->avserver, manager->clipbrd, manager->system);
					}

					free(copy);
				}
				Fclose(filehdl);
			}
		}

		/*  We only take into account an empty SCRAP.TXT if this is boot time.
		    If we didn't, the Manager would be completely empty. */
		if ((modified == TRUE) || (firstrun == TRUE))
		{
			/* Count # of clips while there are files. Since the user may have reduced the clipboard
			   size, we also delete all other clips, if any, and, just in case, TEMP* files. */
			while (Dxreaddir(sizeof buffer, dirhdl, buffer, &xattr, &xr) != ENMFIL)
			{
 				if ((buffer[4] == 't') && (buffer[5] == 'e') && (buffer[6] == 'm') && (buffer[7] == 'p'))
	 			{
					snprintf(path, sizeof path, "%s%s", manager->clipbrd, &buffer[4]);

					if (Fdelete(path) != E_OK) {
						form_alert(1, FA_ERROR "[Error deleting file|in update_manager().][OK]");
					}
	 			}
				if (nclips < conf->size)
	 			{
		 			if ((buffer[4] == 'c') && (buffer[5] == 'l') && (buffer[6] == 'i') && (buffer[7] == 'p'))
					{
						/* We must distinguish between the lack or presence of clips besides clip00. */
						if ((buffer[8] == '0') && (buffer[9] == '0'))
						{
							if (nclips == -1)
							{
								clips[0] = 0;
								nclips = 0;
							}
							continue;
						}

						/* If this is the 1st proper clip found, we must adjust the #. */
						if (nclips == -1) {
							nclips = 1;
						}
						else if (nclips >= 0) {
							nclips++;
						}
						clips[nclips] = strtol(&buffer[8], NULL, 10);
					}
				}
				else // Delete file (except 'clip00') since limit has been reached. Don't increment nclips.
				{
		 			if ((buffer[4] == 'c') && (buffer[5] == 'l') && (buffer[6] == 'i') && (buffer[7] == 'p') &&
		 			    (buffer[8] != '0') && (buffer[9] != '0'))
					{
						snprintf(path, sizeof path, "%s%s", manager->clipbrd, &buffer[4]);

						if (Fdelete(path) != E_OK) {
							form_alert(1, FA_ERROR "[Error deleting file|in update_manager().][OK]");
						}
					}
				}
			}

			if (nclips > 0)
			{
				qsort(clips, nclips, sizeof clips[0], compare_clip_index);

				/* Find if some clips are missing. */
				for (clipn = 0; clipn <= nclips; clipn++)
				{
					if (clips[clipn] > nclips) {
						del = TRUE;
					}
				}
			}

			/* Reorder clips either by reordering or renaming backwards. Don't create a new clip if this
			   is the 1st time we process the clipboard. */
			if (del == TRUE) {
				reorder_clips(clips, nclips, manager->clipbrd);
			}
			else if ((nclips >= SCRAP) && (firstrun == FALSE)) {
				rename_backwards(manager->clipbrd, nclips, conf);
			}

			if ((empty == FALSE) && (scraphdl >= 0))
			{
		 		/* Backup scrap.txt into clip00. Make sure there is enough free space available on the drive
		 		   (conf->freespace is in MiB). We need the device number. NOTE: it would appear that we have
		 		   to add 1 to xattr.dev because 0 is 'current drive'? */
				snprintf(path, sizeof path, "%s%s", manager->clipbrd, "clip00");
				Dfree(&dinfo, xattr.dev + 1);

				if ((dinfo.b_free * dinfo.b_secsiz * dinfo.b_clsiz) > (conf->freespace * 1048576))
				{
					if ((filehdl = Fcreate(path, 0)) >= 0)
					{
						if (content != NULL)
						{
							Fwrite(filehdl, manager->clips[SCRAP]->size, content);
						}
						Fclose(filehdl);
					}
					else {
						form_alert(1, FA_ERROR "[Error while creating 'clip00'!][OK]");
					}
				}
				else {
					form_alert(1, get_free_string(WN_SPACE));
				}

				av_protokoll_path_update(manager->avserver, manager->clipbrd, manager->system);
			}
			manager->nclips = read_directory(dirhdl, &xattr, buffer, sizeof buffer, manager, conf);
			update_listbox(manager);
		}

		free(content);
	}
	return modified;
}

/*
 * Reorder clips in the CLIPBRD directory.
 */
static void reorder_clips(long clips[], int nclips, const char *clipbrd)
{
	char path[LBUFFER]    = {0};
	char newpath[LBUFFER] = {0};

	/* Reorder clips. 1st backup each clip file in a tempXX file in numerical order. */
	for (int i = 0; i <= nclips; i++)
	{
		if (clips[i] < 10)
		{
			snprintf(path, sizeof path, "%s%s%ld", clipbrd, "clip0", clips[i]);
		}
		else if (clips[i] >= 10)
		{
			snprintf(path, sizeof path, "%s%s%ld", clipbrd, "clip", clips[i]);
		}

		/* Temporary files. */
		if (i < 10)
		{
			snprintf(newpath, sizeof path, "%s%s%d", clipbrd, "temp0", i);
		}
		else if (i >= 10)
		{
			snprintf(newpath, sizeof path, "%s%s%d", clipbrd, "temp", i);
		}

		if (Frename(0, path, newpath) != E_OK) {
			form_alert(1, FA_ERROR "[Error renaming file|in reorder_clips().][OK]");
		}
	}

	/* Rename each tempXX file into a clipXX file. We know they are in numerical order. */
	for (int i = 0; i <= nclips; i++)
	{
		if (i < 10)
		{
			snprintf(path, sizeof path, "%s%s%d", clipbrd, "temp0", i);
			snprintf(newpath, sizeof path, "%s%s%d", clipbrd, "clip0", i);
		}
		else if (i >= 10)
		{
			snprintf(path, sizeof path, "%s%s%d", clipbrd, "temp", i);
			snprintf(newpath, sizeof path, "%s%s%d", clipbrd, "clip", i);
		}

		if (Frename(0, path, newpath) != E_OK) {
			form_alert(1, FA_ERROR "[Error renaming file|in reorder_clips().][OK]");
		}
	}
}

/*
 * Re-read directory and get preview of each clip file.
 */
static int read_directory(int dirhdl, XATTR *xattr, char buffer[], size_t size, GEMCLIP *manager, USERCONF *conf)
{
	int nclips = -1;
	long xr	   = 0;

	Drewinddir(dirhdl);

	while ((Dxreaddir(size, dirhdl, buffer, xattr, &xr) != ENMFIL) && (nclips < conf->size))
	{
		if ((buffer[4] == 'c') && (buffer[5] == 'l') && (buffer[6] == 'i') && (buffer[7] == 'p'))
		{
			/* We must distinguish between the lack or presence of clips besides clip00. */
			if ((buffer[8] == '0') && (buffer[9] == '0'))
			{
				if (nclips == -1) {
					nclips = 0;
				}
				continue;
			}

			/* If this is the 1st proper clip found, we must adjust the #. */
			if (nclips == -1) {
				nclips = 1;
			}
			else {
				nclips++;
			}
			read_preview(manager, buffer, size, xattr);
		}
	}

	return (nclips > 0) ? nclips : SCRAP;
}

/*
 * Read the clip (not scrap.txt) file for writing the preview. Get time and date.
 */
static void read_preview(GEMCLIP *manager, char buffer[], size_t size, XATTR *xattr)
{
	char path[LBUFFER] = {0};
	short txtlen       = manager->tree[MANAGER_CLIP01].ob_spec.tedinfo->te_txtlen;
	int clipn          = strtol(&buffer[8], NULL, 10);;
	int cliphdl	       = 0;
	long clipsize      = 0; // Disk space used by each clip.

	snprintf(path, sizeof path, "%s%s", manager->clipbrd, &buffer[4]); // Complete path.

	if ((cliphdl = Fopen(path, 0)) >= 0)
	{
		/* Since files may be stored in any order on the drive we must get their index. */
		strcpy(manager->clips[clipn]->name, &buffer[4]);

		manager->clips[clipn]->index = clipn;
		manager->clips[clipn]->time.hour = (xattr->ctime >> 11) & 0x1F;
		manager->clips[clipn]->time.minute = (xattr->ctime >> 5) & 0x3F;
		manager->clips[clipn]->time.second = xattr->ctime & 0x1F;
		manager->clips[clipn]->time.year = (xattr->cdate >> 9) & 0x7F;
		manager->clips[clipn]->time.month = (xattr->cdate >> 5) & 0x0F;
		manager->clips[clipn]->time.day = xattr->cdate & 0x1F;

		memset(buffer, 0, size);

		/* Update list box entry. Get the full size but only read the preview. */
		if ((clipsize = Fseek(0, cliphdl, SEEK_END)) >= 0)
		{
			if (Fseek(0, cliphdl, SEEK_SET) >= 0)
			{
				if (Fread(cliphdl, size, buffer) >= 0)
				{
					manager->clips[clipn]->size = clipsize;
					manager->fullsize += clipsize; // Register disk space used.
					remove_whitespace(buffer);
					write_preview(manager->clips[clipn], clipsize, txtlen, buffer);
				}
				else {
					form_alert(1, FA_ERROR "[Error reading file|in read_preview().][OK]");
				}
			}
			else {
				form_alert(1, FA_ERROR "[Error seeking file|in read_preview().][OK]");
			}
		}
		else {
			form_alert(1, FA_ERROR "[Error seeking file|in read_preview().][OK]");
		}
		Fclose(cliphdl);
	}
	else {
		form_alert(1, FA_ERROR "[Error opening file|in read_preview().][OK]");
	}
}

/*
 * Copy preview for the Manager. If content is longer than a listbox line add ellipsis to the string.
 */
static void write_preview(CLIPS *clip, long size, short txtlen, const char *buffer)
{
	char dots[] = "[...]";
	int len    = txtlen - 6;

	if (size > txtlen)
	{
		snprintf(clip->preview, txtlen, "%-*s", txtlen, buffer);
		strcpy(&clip->preview[len], dots); // Right align ellipses.
	}
	else {
		snprintf(clip->preview, txtlen, "%s", buffer);
	}
}

/*
 * Rename backwards. Prepare last clip file name. Account for the user limit.
 * If user limit has been reached, renaming clip is the same as renamed one.
 */
static void rename_backwards(const char* clipbrd, int nclips, USERCONF *conf)
{
	char path[LBUFFER]    = {0};
	char newpath[LBUFFER] = {0};
	int nclip             = (nclips == conf->size) ? nclips - 1 : nclips;
	long prevclipn        = -1;

	/* nclips is always at least 0. */
	if (nclips < 9)
	{
		snprintf(path, sizeof path, "%s%s%d", clipbrd, "clip0", nclip);
		snprintf(newpath, sizeof newpath, "%s%s%d", clipbrd, "clip0", nclip + 1);
	}
	else if (nclips == 9) // Special precaution needed (10 - 1 == 9, not 09).
	{
		snprintf(path, sizeof path, "%s%s%d", clipbrd, "clip0", nclip);
		snprintf(newpath, sizeof newpath, "%s%s%d", clipbrd, "clip", nclip + 1);
	}
	else if (nclips > 9)
	{
		snprintf(path, sizeof path, "%s%s%d", clipbrd, "clip", nclip);
		snprintf(newpath, sizeof newpath, "%s%s%d", clipbrd, "clip", nclip + 1);
	}

	/* If limit has been reached, newpath == the oldest clip. We need to delete it 1st,
	   but only if this is the 1st time the clipboard is processed. */
	if (nclips == conf->size) {
		Fdelete(newpath);
	}

	/* Reorder clips by renaming the last clip read in the clipboard
	   into the next one then loop back until clip00. */
	while (nclips >= 0)
	{
		if (nclips != prevclipn)
		{
			prevclipn = nclips;

			if (Frename(0, path, newpath) == 0) // No errors.
			{
				strcpy(newpath, path); // Old path becomes new path.
				nclips -= 2;

				/* Some needed corrections.	*/
				if (nclips == 0) {
					nclips = 1;
				}

				if (nclips < 0) {
					nclips = 0;
				}

				if ((nclips >= 0) && (nclips < 10)) {
					snprintf(path, sizeof path, "%s%s%d", clipbrd, "clip0", nclips);
				}
				else if (nclips >= 10) {
					snprintf(path, sizeof path, "%s%s%d", clipbrd, "clip", nclips);
				}
			}
			else {
				form_alert(1, FA_ERROR "[Error renaming file|in rename_backwards().][OK]");
			}
		}
		else {
			nclips = -1; // Exit loop.
		}
	}
}

/*
 * If any clip is selected, deselect all selected clips. If none is selected, select all.
 * If mode == MENUWIN_SELECT, all clips are (de)selected; if mode == MENUWIN_SLOTS, only the
 * slots corresponding to the visible clips are taken into account.
 */
void select_all_clips(GEMCLIP *manager, int mode)
{
	int i = MANAGER_CLIP01;
	int j = 0;

	if (mode == MENUWIN_ALL_SLOTS)
	{
		/* If at least 1 clip is selected, deselect all slots. */
		if (manager->selected >= 1) // Deselect all.
		{
			while ((i <= MANAGER_CLIP12) && (j <= manager->nclips))
			{
				if (*manager->tree[i].ob_spec.tedinfo->te_ptext != '\0')
				{
					manager->tree[i].ob_state &= ~OS_SELECTED;
					manager->clips[j]->select = FALSE;
				}
				i++;
				j++;
			}
		}
		else // Select all.
		{
			while ((i <= MANAGER_CLIP12) && (j <= manager->nclips))
			{
				if (*manager->tree[i].ob_spec.tedinfo->te_ptext != '\0')
				{
					manager->tree[i].ob_state |= OS_SELECTED;
					manager->clips[j]->select = TRUE;
				}
				i++;
				j++;
			}
		}
	}
	else if (mode == MENUWIN_ALL_CLIPS)
	{
		if (manager->selected >= 1) // Deselect all.
		{
			while (i <= MANAGER_CLIP12)
			{
				manager->tree[i].ob_state &= ~OS_SELECTED;
				i++;
			}

			for (i = 0; i <= manager->nclips; i++)
			{
				manager->clips[i]->select = FALSE;
			}
		}
		else // Select all.
		{
			while (i <= MANAGER_CLIP12)
			{
				if (*manager->tree[i].ob_spec.tedinfo->te_ptext != '\0') {
					manager->tree[i].ob_state |= OS_SELECTED;
				}
				i++;
			}

			for (i = 0; i <= manager->nclips; i++)
			{
				manager->clips[i]->select = TRUE;
			}
		}
	}

	if (manager->selected >= 1)	{
		manager->selected = 0;
	}
	else
	{
		/*  Increment by either the # of used listbox slots or the total number of clips. */
		if (mode == MENUWIN_ALL_SLOTS)
		{
			manager->selected += (i - 2); // Account for offset between object # and clip #.
		}
		else if (mode == MENUWIN_ALL_CLIPS)
		{
			manager->selected = manager->nclips;
		}
	}

	manager->clipn = SCRAP; // Either way, make SCRAP.TXT the default.
	update_objects(manager);
	update_clips_infoline(manager);
	update_preview(manager);
	redraw_object(manager->window->handle, manager->tree, MANAGER_LISTBOX);
	redraw_object(manager->window->handle, manager->tree, MANAGER_PREVIEW_BOX);
}


/*
 * Save the currently selected clip(s).
 */
void save_clips(GEMCLIP *manager, int clipn)
{
	char *homedir         = NULL;
	char *content         = NULL;
	char *title           = get_gstring(STRINGS, STRINGS_SAVE_AS);
	char inpath[LBUFFER]  = {0};
	char outpath[LBUFFER] = {0};
	char outdir[LBUFFER]  = {0};
	char name[MBUFFER]    = {0}; // Default file name displayed in the selector.
	short button          = -1;
	int infile            = 0;
	int outfile           = 0;
	int size              = 0;
	int selector          = 0;

	/* We default to the $HOME directory. The user can change the location with
	   the file selector. If no $HOME, default to C:. */
	shel_envrn(&homedir, "HOME=");

	if (homedir != NULL) {
		snprintf(outpath, sizeof outpath, "%s%s", homedir, "\\*");
	}
	else {
		strcpy(outpath, "C:\\*");
	}

	if (clipn == SCRAP) {
		strcpy(name, "scrap.txt");
	}
	else if (clipn < 10) {
		snprintf(name, sizeof name, "%s%d", "clip0", clipn);
	}
	else {
		snprintf(name, sizeof name, "%s%d", "clip", clipn);
	}

	wind_update(BEG_MCTRL);
	selector = fsel_exinput(outpath, name, &button, title);
	wind_update(END_MCTRL);

 	/* No error, selector returned with a file selected. name updated. */
 	if ((selector != 0) && (button == FSEL_OK))
	{
		char *p = NULL;

		/* Build up directory path for redrawing the desktop window, if needed. */
		if ((p = strrchr(outpath, '\\')) != NULL)
		{
			*++p = '\0';
			strcpy(outdir, outpath); // Keep dir path for updating via AV server.
			strcat(outpath, name);
		}

		write_clippath(inpath, sizeof inpath, manager->clipbrd, clipn);

		/* Open selected CLIP for reading. */
		if ((infile = Fopen(inpath, S_READ)) >= 0)
		{
			size = Fseek(0, infile, SEEK_END);

			if ((content = xmalloc(size)) != NULL)
			{
				if (Fseek(0, infile, SEEK_SET) >= 0) // Rewind.
				{
					if (Fread(infile, size, content) >= 0)
					{
						/* Open selected FILE for writing into it. */
						if ((outfile = Fopen(outpath, S_WRITE)) >= 0)
						{
							/* File exists. Overwrite its contents. NOTE: we should warn the user. */
							if (Fwrite(outfile, size, content) >= 0) {
								av_protokoll_path_update(manager->avserver, outdir, manager->system);
							}
							else {
								form_alert(1, FA_ERROR "[Error writing file|in save_clips().][OK]");
							}
							Fclose(outfile);
						}
					 	/* File does not exist. */
						else if ((outfile = Fcreate(outpath, 0)) >= 0)
						{
							if (Fwrite(outfile, size, content) >= 0) {
								av_protokoll_path_update(manager->avserver, outdir, manager->system);
							}
							else {
								form_alert(1, FA_ERROR "[Error writing file|in save_clips().][OK]");
							}
							Fclose(outfile);
						}
					}
					else {
						form_alert(1, FA_ERROR "[Error reading file|in save_clips().][OK]");
					}
				}
				else {
					form_alert(1, FA_ERROR "[Error seeking file|in save_clips().][OK]");
				}
				free(content);
			}
			else {
				form_alert(1, FA_ERROR "[Error with malloc |in save_clips().][OK]");
			}
			Fclose(infile);
		}
		else {
			form_alert(1, FA_ERROR "[Error opening file|in save_clips().][OK]");
		}
	}
}

/*
 * Save the currently selected clip(s).
 * <== Parameters:
 *   - selected menu item, which determines how and where the file should be written;
 *   - clip to save.
 */
void append_clips(GEMCLIP *manager, short item, int clipn)
{
	char *homedir         = NULL;
	char *content         = NULL;
	char *title           = NULL;
	char inpath[LBUFFER]  = {0};
	char outpath[LBUFFER] = {0};
	char outdir[LBUFFER]  = {0};
	char name[MBUFFER]    = {0}; // Default file name displayed in the selector.
	short button          = -1;
	short mode            = S_WRITE;
	int infile            = 0;
	int outfile           = 0;
	int size              = 0;

	/* Set extended file modes. */
	if (item == MENUWIN_SAVE) {
		mode |= 0x200; // Create file if it doesn't exist.
	}
	else if (item == MENUWIN_CURRENT) {
		mode |= 0X400; // Empty the file.
	}

	/* We default to the $HOME directory. The user can change the location with
	   the file selector. If no $HOME, default to C:. */
	shel_envrn(&homedir, "HOME=");

	 /* If not appending to or overwriting scrap.txt we need neither the file selector nor $HOME. */
	if ((item != MENUWIN_APPEND_CURRENT) && (item != MENUWIN_CURRENT))
	{
		if (homedir != NULL) {
			snprintf(outpath, sizeof outpath, "%s%s", homedir, "\\*");
		}
		else {
			strcpy(outpath, "C:\\*");
		}

		if (clipn == SCRAP) {
			strcpy(name, "scrap.txt");
		}
		else if (item == MENUWIN_APPEND) {
			*name = '\0';
		}
		else if (clipn < 10) {
			snprintf(name, sizeof name, "%s%d", "clip0", clipn);
		}
		else {
			snprintf(name, sizeof name, "%s%d", "clip", clipn);
		}

		wind_update(BEG_MCTRL);

		if (item == MENUWIN_APPEND) {
			title = get_gstring(STRINGS, STRINGS_APPEND_TO);
		}
		else {
			title = get_gstring(STRINGS, STRINGS_SAVE_AS);
		}

		if (fsel_exinput(outpath, name, &button, title) != 0)
		{
			if (button == FSEL_OK) // Selector returned with a file selected. 'name' is updated.
			{
				char *p = NULL;

				/* Build up directory path for redrawing the desktop window, if appropriate. */
				if ((p = strrchr(outpath, '\\')) != NULL)
				{
					*++p = '\0';
					strcpy(outdir, outpath); // Keep dir path for updating via AV server.
					strcat(outpath, name);
				}
			}
		}
		wind_update(END_MCTRL);
	}

	if (button != FSEL_CANCEL) // If selector used and not canceled (may be -1 or FS_OK).
	{
		write_clippath(inpath, sizeof inpath, manager->clipbrd, clipn);

		if ((item == MENUWIN_APPEND_CURRENT) || (item == MENUWIN_CURRENT)) {
			snprintf(outpath, sizeof outpath, "%s%s", manager->clipbrd, "scrap.txt");
		}

		if ((infile = Fopen(inpath, S_READ)) >= 0)
		{
			size = Fseek(0, infile, SEEK_END);

			if ((content = xmalloc(size)) != NULL)
			{
				if (Fseek(0, infile, SEEK_SET) >= 0) // Rewind.
				{
					if (Fread(infile, size, content) >= 0)
					{
						if ((outfile = Fopen(outpath, mode)) >= 0)
						{
							if ((item != MENUWIN_SAVE) && (item != MENUWIN_CURRENT)) {
								Fseek(0, outfile, SEEK_END); // Move to the end to append.
							}

							if (Fwrite(outfile, size, content) >= 0)
							{
								/* If scrap.txt modified, notify other apps AND GEMClip itself. */
								if ((item == MENUWIN_APPEND_CURRENT) || (item == MENUWIN_CURRENT))
								{
									av_protokoll_path_update(manager->avserver, manager->clipbrd, manager->system);
									broadcast_sc_changed();
									send_local_message(SC_CHANGED, 0x0002, 0, 0, 0, 0);
								}
								else {
									av_protokoll_path_update(manager->avserver, outdir, manager->system);
								}
							}
							else {
								form_alert(1, FA_ERROR "[Error writing file|in save_clips().][OK]");
							}

							Fclose(outfile);
						}
						else {
							form_alert(1, FA_ERROR "[Error opening file|in save_clips().][OK]");
						}
					}
					else {
						form_alert(1, FA_ERROR "[Error reading file|in save_clips().][OK]");
					}
				}
				else {
					form_alert(1, FA_ERROR "[Error seeking file|in save_clips().][OK]");
				}
				free(content);
			}
			else
			{
				form_alert(1, FA_ERROR "[Error with malloc |in save_clips().][OK]");

			}
			Fclose(infile);
		}
		else {
			form_alert(1, FA_ERROR "[Error opening file|in save_clips().][OK]");
		}
	}
}

/*
 * Append the content of a clip at the end of another clip.
 */
void merge_clips(GEMCLIP *manager)
{
	char *content      = NULL;
	char clippath[LBUFFER] = {0};
	char destpath[LBUFFER] = {0};
	int clipn = 1; // Skip SCRAP.TXT.
	int dest  = 0;
	int clip  = 0;
	int size  = 0;

	/* Get 1st selected clip to write pathname with. */
	while (manager->clips[clipn]->select == FALSE) {
		clipn++;
	}

	write_clippath(destpath, sizeof destpath, manager->clipbrd, clipn);
	clipn++; // Skip to next clip.

	/* Append whatever clips may be selected to the first one. */
	for (int i = clipn; i <= manager->nclips; i++)
	{
		if (manager->clips[i]->select == TRUE)
		{
			write_clippath(clippath, sizeof clippath, manager->clipbrd, i);

			if ((clip = Fopen(clippath, S_READ)) >= 0)
			{
				size = Fseek(0, clip, SEEK_END);
				Fseek(0, clip, SEEK_SET);

				if ((size > 0) && ((content = xmalloc(size)) != NULL))
				{
					if (Fread(clip, size, content) >= 0)
					{
						if ((dest = Fopen(destpath, S_WRITE)) >= 0)
						{
							Fseek(0, dest, SEEK_END);
							Fwrite(dest, size, content);
							Fclose(dest);
						}
						else {
							form_alert(1, FA_ERROR "[Error opening file|in merge_clips().][OK]");
						}
					}
					else {
						form_alert(1, FA_ERROR "[Error reading file|in marge_clips().][OK]");
					}
					free(content);
				}
				Fclose(clip);
			}
			else {
				form_alert(1, FA_ERROR "[Error opening file|in merge_clips().][OK]");
			}
		}
	}

	deselect_all(manager);

	/* Force an update of the Manager. */
    manager->clips[SCRAP]->time.year = -1;
	broadcast_sc_changed();
	send_local_message(SC_CHANGED, 0x0002, 0, 0, 0, 0);
}

/*
 * Find the 1st (or only) selected clip in the array.
 */
int find_first_selected(GEMCLIP *manager)
{
	int clipn = SCRAP; // Will default to SCRAP.TXT if no clip selected.

	if (manager->selected >= 1)
	{
		for (clipn = 1; clipn <= manager->nclips; clipn++)
		{
			if (manager->clips[clipn]->select == TRUE) {
				break;
			}
		}
	}

	return clipn;
}

/*
 * Find if any clips are selected. SCRAP.TXT counts as a valid selection so we loop from 0.
 * Return TRUE or FALSE.
 */
int selection(CLIPS *clips[])
{
	int i = 0;
	int selected = FALSE;

	while ((selected == FALSE) && (i < NCLIPS))
	{
		if (clips[i++]->select == TRUE) {
			selected = TRUE;
		}
	}

	return selected;
}

/*
 * Delete selected clip(s).
 */
void delete_clips(GEMCLIP *manager, USERCONF *conf)
{
	char path[LBUFFER] = {0};
	int delete         = FALSE;

	if (manager->selected >= 1)
	{
		if (conf->warnings == FALSE) // Warnings not disabled.
		{
			if (manager->selected >= WARNCLIPS) // At least 12 clips selected.
			{
				if (form_alert(2, get_free_string(WN_TWELVE)) == 1) // Yes button pressed.
				{
					if (form_alert(1, get_free_string(WN_DELETE)) == 1) {
						delete = TRUE;
					}
				}
			}
			else if (form_alert(1, get_free_string(WN_DELETE)) == 1) {
				delete = TRUE;
			}
		}
		else {
			delete = TRUE;
		}

		if (delete == TRUE)
		{
			for (int i = 1; i <= manager->nclips; i++)
			{
				if (manager->clips[i]->select == TRUE)
				{
					write_clippath(path, sizeof path, manager->clipbrd, i);

					if (Fdelete(path) != E_OK) {
						form_alert(1, FA_ERROR "[Error deleting file|in delete_clips().][OK]");
					}

					manager->fullsize -= manager->clips[i]->size;
					manager->nclips--;
					manager->clips[i]->index = 0;
					manager->clips[i]->select = FALSE;
				}
			}

			/* Deselect all, reset default to SCRAP.TXT. */
			for (int i = 0; i < NCLIPS; i++) {
				manager->clips[i]->select = FALSE;
			}

			deselect_all(manager);

			/* Force an update of the Manager. */
		    manager->clips[SCRAP]->time.year = 0; // Set update_manager() to work as 'first run'.
			send_local_message(SC_CHANGED, 0x0002, 0, 0, 0, 0);
		}
	}
}

/*
 * Write path into buffer. Receive the buffer to write into and the clip index. Return the size
 * of the string written + 1 for the terminating null. Returns 0 if an error occurred.
 */
int write_clippath(char *path, size_t pathsize, const char *clipbrd, int clipn)
{
	int size = 0;

	if (clipn == SCRAP) {
		size = snprintf(path, pathsize, "%s%s", clipbrd, "scrap.txt");
	}
	else if (clipn < 10) {
		size = snprintf(path, pathsize, "%s%s%d", clipbrd, "clip0", clipn);
	}
	else if (clipn >= 10) {
		size = snprintf(path, pathsize, "%s%s%d", clipbrd, "clip", clipn);
	}

	if (size < 0) {
		size = 0;
	}
	else if ((size > 0) && (size < (int)pathsize)) {
		size = size + 1;
	}

	return size;
}

/*
 * Send preview (actual content of the clip after pruning) to a DICT client (namely, GEMDict).
 */
void define_clips(const char *command)
{
	char *path           = NULL;
	char buffer[MBUFFER] = {'x'}; // Workaround for an issue with shell_write() eating the 1st char.

	snprintf(&buffer[1], sizeof buffer, "%s", command);
	shel_envrn(&path, "DICT=");
	shel_write(SWM_LAUNCH, 0, CL_NORMAL, path, buffer);
}


/*
 * Deselect all clips.
 */
void deselect_all(GEMCLIP *manager)
{
	manager->selected = 1;
	select_all_clips(manager, MENUWIN_ALL_CLIPS);
}
                                                                          