/*
 * GEMClip, a Clipboard Manager for FreeMiNT and MagiC.
 *
 * Copyright (C) 2011 Jean-Franois Lemaire (jflemaire@skynet.be)
 *
 * This file is part of GEMClip.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <mint/osbind.h>
#include <errno.h>
#include <dirent.h>

#include "general.h"

#include "clipbrd.h"
#include "config.h"
#include "protocols.h"
#include "utils.h"

/* Sizes Defaults. */
enum { SIZE = 32, CLIPSIZE = 32708, FREESPACE = 32, ICONTIMEOUT = 5 };

struct input {
	int dbclick;
	int rmbclick;
	int enter;
	int space;
	int insert;
	long icontimeout;
};

static FILE *get_config_file	(const char *mode, const char *cfgname);
static void  reset_config		(USERCONF *conf);
static void  init_config_form	(USERCONF *conf, OBJECT *tree);
static void  save_config        (GEMCLIP *manager, WINDOW *window, USERCONF *conf, OBJECT *tree, short object, struct input *input);
static int   handle_popup_combo (OBJECT *tree, OBJECT *popup, short whandle, short object, short gstring);

/*
 * Look for a 'gemclip.cfg' configuration file. If none is found, reset default settings.
 */
void init_config(USERCONF *conf)
{
	FILE *fp  = NULL;

	if (((fp = get_config_file("r", "gemclip.cfg")) != NULL) || ((fp = get_config_file("r", "GEMCLIP.CFG")) != NULL))
	{
		/* Check that the file is not empty before reading from it. */
		if ((fgetc(fp)) != EOF) {
			read_config(fp, conf);
		}
		else {
			 reset_config(conf);
		}
	}
	else {
		reset_config(conf);
	}
}

/*
 * Find the directory from where to read (and where to write) the config file. We first
 * look for the folder 'defaults' in $HOME, and if none is found, create it. If there
 * is no $HOME at all, we look in the directory the program was run from.
 * ==> Return the file or NULL.
 */
static FILE *get_config_file(const char *mode, const char *name)
{
	int found	  = TRUE;
	char *homedir = NULL;
	char path[LBUFFER] = {0};
	FILE *fp = NULL;
	DIR *dp  = NULL;

	shel_envrn(&homedir, "HOME=");

	if (homedir != NULL)
	{
		snprintf(path, sizeof path, "%s%s%s", homedir, "\\defaults\\", name);

		if ((fp = fopen(path, mode)) == NULL) // No config in 'defaults' subdir.
		{
			snprintf(path, sizeof path, "%s%s", homedir, "\\defaults");

			if ((dp = opendir(path)) == NULL) // No 'defaults' dir.
			{
				if ((Dcreate(path)) != 0) {
					form_alert(1, FA_ERROR "[Error creating directory!][OK]");
				}
			}
			else
			{
				if ((closedir(dp)) != 0) {
					form_alert_string(1, FA_ERROR, "[Error: %s][OK]", strerror(errno));
				}
			}

			found = FALSE;
		}
	}
	else {
		found = FALSE;
	}

	if (found == FALSE) // Last resort. Look in the app dir.
	{
		if ((fp = fopen(name, mode)) == NULL) {
			fp = NULL;
		}
	}

	return fp;
}

/*
 * Parse the config file and set the variables to the values stored into it.
 */
void read_config(FILE *fp, USERCONF *conf)
{
	char value[LBUFFER] = {0};
	size_t size    = sizeof value;
	OBJECT *tree   = get_gem_tree(CONFIG);
	OBJECT *mngr   = get_gem_tree(MANAGER);
	GRECT deskrect = {0, 0, 0, 0};

	wind_get_grect(0, WF_WORKXYWH, &deskrect);

	if (tree != NULL)
	{
		if ((get_value(fp, "WINDOW_XPOS", value, size)) != NULL)
		{
			if ((conf->winx = strtol(value, NULL, 10)) >= deskrect.g_w) {
				conf->winx = deskrect.g_x + ((deskrect.g_w - mngr->ob_width) / 2);
			}
		}
		else {
			conf->winx = deskrect.g_x + ((deskrect.g_w - mngr->ob_width) / 2);
		}

		if ((get_value(fp, "WINDOW_YPOS", value, size)) != NULL)
		{
			if ((conf->winy = strtol(value, NULL, 10)) >= deskrect.g_h) {
				conf->winy = deskrect.g_y + ((deskrect.g_h - (mngr->ob_height + 24)) / 2);
			}
		}
		else {
			conf->winy = deskrect.g_y + ((deskrect.g_h - (mngr->ob_height + 24)) / 2);
		}

		if ((get_value(fp, "BOOT_CLEANING", value, size)) != NULL)
		{
			conf->cleaning = (strcmp(value, "true") == 0) ? TRUE : FALSE;
		}
		else {
			conf->cleaning = FALSE;
		}

		if ((get_value(fp, "BOOT_ICONIFY", value, size)) != NULL)
		{
			conf->iconify = (strcmp(value, "true") == 0) ? TRUE : FALSE;
		}
		else {
			conf->iconify = TRUE;
		}

		if ((get_value(fp, "AUTO_ICONIFY", value, size)) != NULL)
		{
			conf->autoicon = (strcmp(value, "true") == 0) ? TRUE : FALSE;
		}
		else {
			conf->autoicon = TRUE;
		}

		if ((get_value(fp, "ICONIFY_TIMEOUT", value, size)) != NULL)
		{
			if ((conf->icontimeout = strtol(value, NULL, 10)) > 30) {
				conf->icontimeout = ICONTIMEOUT;
			}
		}
		else {
			conf->icontimeout = ICONTIMEOUT;
		}

		/* Must go 1st because we need to know if we restore window settings. */
		if ((get_value(fp, "SETTINGS", value, size)) != NULL)
		{
			conf->settings = (strcmp(value, "true") == 0) ? TRUE : FALSE;
		}
		else {
			conf->settings = TRUE;
		}

		if (conf->settings == TRUE)
		{
			if ((get_value(fp, "VIEWMODE", value, size)) != NULL)
			{
				long viewmode = strtol(value, NULL, 10);

				if ((viewmode > POP_VIEWMODE_NOLIST) || (viewmode < POP_VIEWMODE_STANDARD)) {
					conf->viewmode = POP_VIEWMODE_STANDARD;
				}
				else {
					conf->viewmode = viewmode;
				}
			}
			else {
				conf->viewmode = POP_VIEWMODE_STANDARD; // Standard mode is default.
			}
		}
		else {
			conf->viewmode = POP_VIEWMODE_STANDARD;
		}

		if ((get_value(fp, "CLIPMODE", value, size)) != NULL)
		{
			if ((conf->clipmode = strtol(value, NULL, 10)) > LIST_INFO) {
				conf->clipmode = LIST_TEXT;
			}
		}
		else {
			conf->clipmode = LIST_TEXT; // Preview is default.
		}

		if ((get_value(fp, "INFOMODE", value, size)) != NULL)
		{
			if ((conf->infomode = strtol(value, NULL, 10)) > LIST_SIZE) {
				conf->infomode = LIST_NAME;
			}
		}
		else {
			conf->infomode = LIST_NAME;
		}

		if ((get_value(fp, "ORDER", value, size)) != NULL)
		{
			if ((conf->order = strtol(value, NULL, 10)) > LIST_BOTTOM) {
				conf->order = LIST_TOP;
			}
		}
		else {
			conf->order = LIST_TOP;
		}

		if ((get_value(fp, "CYCLING", value, size)) != NULL)
		{
			conf->cycling = (strcmp(value, "true") == 0) ? TRUE : FALSE;
		}
		else {
			conf->cycling = TRUE;
		}

		if ((get_value(fp, "BUBBLES", value, size)) != NULL)
		{
			conf->bubbles = (strcmp(value, "true") == 0) ? TRUE : FALSE;
		}
		else {
			conf->bubbles = TRUE;
		}

		if ((get_value(fp, "WARNINGS", value, size)) != NULL) {
			conf->warnings = (strcmp(value, "false") == 0) ? FALSE : TRUE;
		}
		else {
			conf->warnings = FALSE;
		}

		if ((get_value(fp, "DOUBLE_CLICK", value, size)) != NULL)
		{
			conf->dbclick = strtol(value, NULL, 10);
			if ((conf->dbclick > POP_ACTIONS_VIEW) || (conf->dbclick < POP_ACTIONS_MENU)) {
				conf->dbclick = POP_ACTIONS_CURRENT;
			}
		}
		else {
			conf->dbclick = POP_ACTIONS_CURRENT;
		}

		if ((get_value(fp, "RMB_CLICK", value, size)) != NULL)
		{
			conf->rmbclick = strtol(value, NULL, 10);
			if ((conf->rmbclick > POP_ACTIONS_VIEW) || (conf->rmbclick < POP_ACTIONS_MENU)) {
				conf->rmbclick = POP_ACTIONS_MENU;
			}
		}
		else {
			conf->rmbclick = POP_ACTIONS_MENU;
		}

		if ((get_value(fp, "ENTER_KEY", value, size)) != NULL)
		{
			conf->enter = strtol(value, NULL, 10);
			if ((conf->enter > POP_ACTIONS_VIEW) || (conf->enter < POP_ACTIONS_MENU)) {
				conf->enter = POP_ACTIONS_CURRENT;
			}
		}
		else {
			conf->enter = POP_ACTIONS_CURRENT;
		}

		if ((get_value(fp, "SPACE_KEY", value, size)) != NULL)
		{
			conf->space = strtol(value, NULL, 10);
			if ((conf->space > POP_ACTIONS_VIEW) || (conf->space < POP_ACTIONS_MENU)) {
				conf->space = POP_ACTIONS_VIEW;
			}
		}
		else {
			conf->space = POP_ACTIONS_VIEW;
		}

		if ((get_value(fp, "INSERT_KEY", value, size)) != NULL)
		{
			conf->insert = strtol(value, NULL, 10);
			if ((conf->insert > POP_ACTIONS_VIEW) || (conf->insert < POP_ACTIONS_MENU)) {
				conf->insert = POP_ACTIONS_INFO;
			}
		}
		else {
			conf->insert = POP_ACTIONS_INFO;
		}

		if ((get_value(fp, "CLIPBRD_SIZE", value, size)) != NULL)
		{
			conf->size = strtol(value, NULL, 10);

			if (conf->size > 99) {
				conf->size = 99;
			}
			write_popup_boxtext(tree, CONFIG_SIZE, conf->size);
		}
		else {
			conf->size = SIZE;
		}

		if ((get_value(fp, "CLIP_SIZE", value, size)) != NULL)
		{
			conf->clipsize = strtol(value, NULL, 10);
			write_popup_boxtext(tree, CONFIG_CLIP_SIZE, conf->clipsize);
		}
		else {
			conf->clipsize = CLIPSIZE;
		}

		if ((get_value(fp, "FREE_SPACE", value, size)) != NULL)
		{
			conf->freespace = strtol(value, NULL, 10);
			write_popup_boxtext(tree, CONFIG_FREE_SPACE, conf->freespace);
		}
		else {
			conf->freespace = FREESPACE;
		}

		if ((get_value(fp, "CLIPBRD_PATH", value, size)) != NULL)
		{
			if (strchr(value, '\\') != NULL) { // Check that we have some sort of correct path.
				strcpy(conf->clipbrd, value);
			}
			else {
				strcpy(conf->clipbrd, "C:\\CLIPBRD\\");
			}
		}
		else {
			strcpy(conf->clipbrd, "C:\\CLIPBRD\\");
		}

		if ((get_value(fp, "OVERRIDE_PATH", value, size)) != NULL)
		{
			conf->override = (strcmp(value, "false") == 0) ? FALSE : TRUE;
		}
		else {
			conf->override = FALSE;
		}
	}

	if (fclose(fp) == EOF) {
		form_alert(1, FA_ERROR "[Error closing config file!][OK]");
	}
}

/*
 * Write config file when the accessory is terminated by the system.
 */
void write_config(WINDOW *window, USERCONF *conf)
{
	FILE *fp       = NULL;
	GRECT deskrect = {0, 0, 0, 0};

	wind_get_grect(0, WF_WORKXYWH, &deskrect);

	if ((fp = get_config_file("w", "gemclip.cfg")) != NULL)
	{
		short x       = -1;
		short y       = -1;
		short foo	  = 0;
		OBJECT *about = get_gem_tree(ABOUT);

		/* Get window coordinates. If terminating, the win* variables are saved.
		If closing the window, the conf->win* variables are set for when we need
		to re-open the window. */

 		/* If application not hidden and Manager window not iconified. Also, the system
 		   may have closed the window, so we check if the handle is still valid. */
 		if ((window->icon == FALSE) && (window->handle >= 0)) {
			wind_get(window->handle, WF_CURRXYWH, &x, &y, &foo, &foo);
		}

		fprintf(fp, "# Configuration file for GEMClip, a FreeMiNT/MagiC clipboard manager.\n");
		fprintf(fp, "# Lines starting with a hash character (#) are ignored.\n");

		fprintf(fp, "\n# Program Version #\n");
		fprintf(fp, "VERSION = %s\n", about[ABOUT_VERSION].ob_spec.free_string);

		if (foo >= 0) // We suppose that 0 or less means the window was closed by the system.
		{
			/* We take care not to register win coords if accessory is hidden. */
			fprintf(fp, "\n# Window Position #\n");
			if ((x >= 0) && (x < deskrect.g_w)) {
				fprintf(fp, "WINDOW_XPOS = %d\n", x);
			}
			else {
				fprintf(fp, "WINDOW_XPOS = %d\n", conf->winx);
			}
	
			if ((y >= 0) && (y < deskrect.g_h)) {
				fprintf(fp, "WINDOW_YPOS = %d\n", y);
			}
			else {
				fprintf(fp, "WINDOW_YPOS = %d\n", conf->winy);
			}
		}

		fprintf(fp, "\n# Miscellanous Settings #\n");
		fprintf(fp, "BOOT_CLEANING = %s\n", (conf->cleaning == TRUE) ? "true" : "false");
		fprintf(fp, "BOOT_ICONIFY = %s\n",	(conf->iconify == FALSE) ? "false" : "true");
		fprintf(fp, "AUTO_ICONIFY = %s\n",	(conf->autoicon == TRUE) ? "true" : "false");
		fprintf(fp, "ICONIFY_TIMEOUT = %d\n", conf->icontimeout);
		fprintf(fp, "VIEWMODE = %d\n",		 conf->viewmode);
		fprintf(fp, "CLIPMODE = %d\n",		 conf->clipmode);
		fprintf(fp, "INFOMODE = %d\n",		 conf->infomode);
		fprintf(fp, "ORDER = %d\n",			 conf->order);
		fprintf(fp, "CYCLING = %s\n",		(conf->cycling == TRUE) ? "true" : "false");
		fprintf(fp, "BUBBLES = %s\n",		(conf->bubbles == TRUE) ? "true" : "false");
		fprintf(fp, "SETTINGS = %s\n",		(conf->settings == TRUE) ? "true" : "false");
		fprintf(fp, "CLIPBRD_PATH = %s\n",	 conf->clipbrd);
		fprintf(fp, "OVERRIDE_PATH = %s\n", (conf->override == FALSE) ? "false" : "true");

		fprintf(fp, "\n# Keyboard/Mouse Settings #\n");
		fprintf(fp, "DOUBLE_CLICK = %d\n",	 conf->dbclick);
		fprintf(fp, "RMB_CLICK = %d\n", 	 conf->rmbclick);
		fprintf(fp, "ENTER_KEY = %d\n", 	 conf->enter);
		fprintf(fp, "SPACE_KEY = %d\n", 	 conf->space);
		fprintf(fp, "INSERT_KEY = %d\n", 	 conf->insert);

		fprintf(fp, "\n# Advanced Settings #\n");
		fprintf(fp, "CLIPBRD_SIZE = %d\n",	 conf->size);
		fprintf(fp, "CLIP_SIZE = %ld\n",	 conf->clipsize);
		fprintf(fp, "FREE_SPACE = %ld\n",	 conf->freespace);
		fprintf(fp, "WARNINGS = %s\n",		(conf->warnings == FALSE) ? "false" : "true");

		if (fclose(fp) == EOF) {
			form_alert(1, FA_ERROR "[Error closing config file.][OK]");
		}
	}
	else {
		form_alert(1, FA_ERROR "[Error opening config file.][OK]");
	}
}

/*
 * Set default settings if no config file could be found or the user clicked the 'Defaults'.
 */
void reset_config(USERCONF *conf)
{
	GRECT deskrect = {0, 0, 0, 0};
	OBJECT *tree = get_gem_tree(MANAGER);

	wind_get_grect(0, WF_WORKXYWH, &deskrect);

	/* Calculate the coordinates to open the dialog in the centre of the screen. */
	conf->winx = deskrect.g_x + ((deskrect.g_w - tree->ob_width) / 2);
	conf->winy = deskrect.g_y + ((deskrect.g_h - (tree->ob_height + 24)) / 2);
	conf->cleaning = FALSE;
	conf->iconify = TRUE;
	conf->autoicon = TRUE;
	conf->icontimeout = ICONTIMEOUT;
	conf->viewmode = POP_VIEWMODE_STANDARD;
	conf->clipmode = LIST_TEXT;
	conf->infomode = LIST_NAME;
	conf->order = LIST_TOP;
	conf->settings = TRUE;
	conf->cycling = TRUE;
	conf->bubbles = TRUE;
	strcpy(conf->clipbrd, "C:\\CLIPBRD\\");
	conf->override = FALSE;
	conf->dbclick = POP_ACTIONS_CURRENT;
	conf->rmbclick = POP_ACTIONS_MENU;
	conf->enter = POP_ACTIONS_CURRENT;
	conf->space = POP_ACTIONS_INFO;
	conf->insert = POP_ACTIONS_INFO;
	conf->size = SIZE;
	conf->clipsize = CLIPSIZE;
	conf->freespace = FREESPACE;
	conf->warnings = FALSE;
}

/*
 * Create and open the configuration dialog. Display the last opened sub-dialog.
 * Return NULL on error.
 */
int open_config_form(GEMCLIP *manager, WINDOW *window, USERCONF *conf)
{
	int attribs = NAME | INFO | CLOSER | MOVER | SMALLER;
	static int reopen = FALSE;
	GRECT inrect	  = {0, 0, 0, 0};
	GRECT outrect	  = {0, 0, 0, 0};
	OBJECT *tree	  = get_gem_tree(CONFIG);
	OBJECT *icons	  = get_gem_tree(ICONS);

	init_config_form(conf, tree);

	/* Set the height of the background box (0), relative to the BOTTOM box. */
	tree->ob_height = tree[CONFIG_BOTTOM].ob_y;

	/* Hide the other boxes. */
	tree[CONFIG_KEYS].ob_flags |= OF_HIDETREE;
	tree[CONFIG_SIZES].ob_flags |= OF_HIDETREE;

	/* If the dialog has already been opened, reset it to its last coordinates. */
	if (reopen == FALSE) // First opening.
	{
		 /* Defaults to MISC sub-dialog. */
		tree[CONFIG_SIDE_MISC].ob_state = OS_SELECTED;
		wind_get_grect(0, WF_WORKXYWH, &outrect);

		/* Calculate the coordinates to open the dialog in the centre of the screen. */
		inrect.g_x = outrect.g_x + ((outrect.g_w - tree->ob_width)  / 2);
		inrect.g_y = outrect.g_y + ((outrect.g_h - tree->ob_height) / 2);
		inrect.g_w = tree->ob_width;
		inrect.g_h = tree->ob_height;

		wind_calc_grect(WC_BORDER, attribs, &inrect, &outrect); // Calc the full coords of the win.
	}
	else { // Re-opening.
		outrect = manager->cfgrect;
	}

	/* Set the other boxes to the same position as that of the default one. */
	tree[CONFIG_KEYS].ob_x = tree[CONFIG_SIZES].ob_x = tree[CONFIG_MISC].ob_x;
	tree[CONFIG_KEYS].ob_y = tree[CONFIG_SIZES].ob_y = tree[CONFIG_MISC].ob_y;

	/* Iinitialize the structure and create a virtual workstation. */
	if ((window = init_window(window)) != NULL)
	{
		/* Create and open the window. */
		if ((window->handle = wind_create_grect(attribs, &outrect)) != 0)
		{
			wind_set_str(window->handle, WF_NAME, get_gstring(STRINGS, STRINGS_CONFIG_NAME));
			wind_set_str(window->handle, WF_INFO, ""); // Will be written later.

			/* Left side icons are in another tree (we avoid duplicatING icons in the resource). */
			tree[CONFIG_SIDE_MISC].ob_spec	= icons[ICONS_MISC].ob_spec;
			tree[CONFIG_SIDE_KEYS].ob_spec	= icons[ICONS_KEYS].ob_spec;
			tree[CONFIG_SIDE_SIZES].ob_spec = icons[ICONS_SIZES].ob_spec;

			if (wind_open_grect(window->handle, &outrect) != 0)
			{
				window->open = TRUE;

				/* If the dialog has already been opened, display the last selected
				sub-dialog (we check the state of the last selected side-panel icon). */
				if (tree[CONFIG_SIDE_MISC].ob_state & OS_SELECTED)
				{
					tree[CONFIG_MISC].ob_flags &= ~OF_HIDETREE;
					tree[CONFIG_KEYS].ob_flags  |= OF_HIDETREE;
					tree[CONFIG_SIZES].ob_flags |= OF_HIDETREE;

					wind_set_str(window->handle, WF_INFO, get_gstring(CONFIG, CONFIG_MANAGER_INFO_MISC));
					window->strindex = CONFIG_MANAGER_INFO_MISC;
				}
				else if (tree[CONFIG_SIDE_KEYS].ob_state & OS_SELECTED)
				{
					tree[CONFIG_MISC].ob_flags	|= OF_HIDETREE;
					tree[CONFIG_KEYS].ob_flags &= ~OF_HIDETREE;
					tree[CONFIG_SIZES].ob_flags |= OF_HIDETREE;

					wind_set_str(window->handle, WF_INFO, get_gstring(CONFIG, CONFIG_MANAGER_INFO_KEYS));
					window->strindex = CONFIG_MANAGER_INFO_KEYS;
				}
				else if (tree[CONFIG_SIDE_SIZES].ob_state & OS_SELECTED)
				{
					tree[CONFIG_MISC].ob_flags	 |= OF_HIDETREE;
					tree[CONFIG_KEYS].ob_flags	 |= OF_HIDETREE;
					tree[CONFIG_SIZES].ob_flags &= ~OF_HIDETREE;

					wind_set_str(window->handle, WF_INFO, get_gstring(CONFIG, CONFIG_MANAGER_INFO_SIZES));
					window->strindex = CONFIG_MANAGER_INFO_SIZES;
				}

 				/* Store the current infoline. */
 				strcpy(window->info, tree[window->strindex].ob_spec.free_string);
				reopen = TRUE;
			}
			else
			{
				form_error(FERR_NOHANDLES);
				window->open = FALSE;
			}
		} else
		{
			form_error(FERR_NOHANDLES);
			window->open = FALSE;
		}
	} else {
		window->open = FALSE;
	}

	return window->open;
}

/*
 * Initialize values and set up objects in the configuration dialog before opening it.
 * Also called when selecting the "Defaults" button.
 */
static void init_config_form(USERCONF *conf, OBJECT *tree)
{
	char *str = NULL;
	int len   = 0;
	int item  = 0;
	OBJECT *autoicon = get_gem_tree(POP_AUTOICON);

	/* Write clipboard path in editable field. */
	snprintf(tree[CONFIG_PATH].ob_spec.free_string, 41, "%s", conf->clipbrd);

	/* Write popup labels. */
	len = tree[CONFIG_AUTOICON_POP].ob_spec.tedinfo->te_txtlen + 1;

	if (conf->icontimeout > 9) {
		snprintf(tree[CONFIG_AUTOICON_POP].ob_spec.tedinfo->te_ptext, len, "%3d", conf->icontimeout);
	}
	else {
		snprintf(tree[CONFIG_AUTOICON_POP].ob_spec.tedinfo->te_ptext, len, "%3d", conf->icontimeout);
	}

	/* Check the autoicon popup item. Uncheck previous selection before. */
	uncheck_prev_item(autoicon);

	switch (conf->icontimeout)
	{
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
		item = conf->icontimeout;
		break;
	case 10:
		item = 6;
		break;
	case 15:
		item = 7;
		break;
	case 20:
		item = 8;
		break;
	case 25:
		item = 9;
		break;
	case 30:
		item = 10;
		break;
	}

	menu_icheck(autoicon, item, CHECK);

	/* These are strings so handled differently. */
	len = tree[CONFIG_DB_CLICK].ob_spec.tedinfo->te_txtlen + 1;

	str = get_gstring(POP_ACTIONS, conf->dbclick);
	snprintf(tree[CONFIG_DB_CLICK].ob_spec.tedinfo->te_ptext, len, "%s", &str[1]);

	str = get_gstring(POP_ACTIONS, conf->rmbclick);
	snprintf(tree[CONFIG_RMB_CLICK].ob_spec.tedinfo->te_ptext, len, "%s", &str[1]);

	str = get_gstring(POP_ACTIONS, conf->enter);
	snprintf(tree[CONFIG_ENTER].ob_spec.tedinfo->te_ptext, len, "%s", &str[1]);

	str = get_gstring(POP_ACTIONS, conf->space);
	snprintf(tree[CONFIG_SPACE].ob_spec.tedinfo->te_ptext, len, "%s", &str[1]);

	str = get_gstring(POP_ACTIONS, conf->insert);
	snprintf(tree[CONFIG_INSERT].ob_spec.tedinfo->te_ptext, len, "%s", &str[1]);

	/* Update LIMITS' FBOXTEXTs with the values saved in the config file. */
	write_popup_boxtext(tree, CONFIG_SIZE, conf->size);
	write_popup_boxtext(tree, CONFIG_CLIP_SIZE, conf->clipsize);
	write_popup_boxtext(tree, CONFIG_FREE_SPACE, conf->freespace);

	/* Check buttons are not selected in the resource. Set them now
		according to the values stored in the config file. */
	if (conf->cleaning == TRUE) {
		tree[CONFIG_CLEANING].ob_state |= OS_SELECTED;
	}

	if (conf->iconify != FALSE) { // Can be TRUE, FALSE, or -1.
		tree[CONFIG_ICONIFY].ob_state |= OS_SELECTED;
	}

	if (conf->autoicon == TRUE)	{
		tree[CONFIG_AUTOICON].ob_state |= OS_SELECTED;
	}
	else // Disable popup.
	{
		tree[CONFIG_AUTOICON_POP].ob_state |= OS_DISABLED;
		tree[CONFIG_AUTOICON_EQUIV].ob_state |= OS_DISABLED;
	}

	if (conf->cycling == TRUE) {
		tree[CONFIG_CYCLING].ob_state |= OS_SELECTED;
	}

	if (conf->bubbles == TRUE) {
		tree[CONFIG_BUBBLES].ob_state |= OS_SELECTED;
	}

	if (conf->settings == TRUE) {
		tree[CONFIG_SETTINGS].ob_state |= OS_SELECTED;
	}

	if (conf->warnings == TRUE) {
		tree[CONFIG_WARNINGS].ob_state |= OS_SELECTED;
	}

	if (conf->override == TRUE) {
		tree[CONFIG_OVERRIDE].ob_state |= OS_SELECTED;
	}
}

/*
 * Handle user interaction with the configuration dialog.
 */
void handle_config_form(GEMCLIP *manager, WINDOW *window, USERCONF *conf, OBJECT *tree, EVMULT_OUT *evmult_out, short object, BUBBLES *bubbles)
{
	static char path[LBUFFER] = {0}; // Must be static for comparing after an OK/Apply.
	char name[SBUFFER] = {0}; // Unused but needed by fsel_exinput().
	short button       = 0;
	int item           = 0;
	static struct input input = {0, 0, 0, 0, 0, 0};
	OBJECT *autopop    = get_gem_tree(POP_AUTOICON);
	OBJECT *actionpop  = get_gem_tree(POP_ACTIONS);

	if (object == 0) // Key pressed.
	{
		switch (evmult_out->emo_kreturn)
		{
		case SC_HELP:
			if (manager->aeslang == AESLANG_GERMAN) {
				call_help(window->handle, tree, object, manager->avserver, "*:\\GEMCLIPD.HYP ", "Konfiguration", manager->system);
			}
			else {
				call_help(window->handle, tree, object, manager->avserver, "*:\\GEMCLIPE.HYP ", "Configuration", manager->system);
			}
			break;
		case SC_UNDO:
			tree[CONFIG_CANCEL].ob_state &= ~OS_SELECTED;
			send_local_message(WM_CLOSED, window->handle, 0, 0, 0, 0);
			break;
		case SC_RETURN:
		case SC_NMPAD_ENTER:
			save_config(manager, window, conf, tree, CONFIG_OK, &input);
			break;
		}
	}
	else if ( !(evmult_out->emo_mbutton & RIGHT_BUTTON))
	{
		switch (object)
		{
		/* Left-side buttons. Avoid flickering by testing if the sub-dialog is already drawn. */
		case CONFIG_SIDE_MISC:
			if (tree[CONFIG_MISC].ob_flags & OF_HIDETREE)
			{
				tree[CONFIG_MISC].ob_flags &= ~OF_HIDETREE;
				tree[CONFIG_KEYS].ob_flags |= OF_HIDETREE;
				tree[CONFIG_SIZES].ob_flags |= OF_HIDETREE;
				redraw_object(window->handle, tree, CONFIG_MISC);
				wind_set_str(window->handle, WF_INFO, tree[CONFIG_MANAGER_INFO_MISC].ob_spec.free_string);
				window->strindex = CONFIG_MANAGER_INFO_MISC;
				strcpy(window->info, tree[window->strindex].ob_spec.free_string);
			}
			break;
		case CONFIG_SIDE_KEYS:
			if (tree[CONFIG_KEYS].ob_flags & OF_HIDETREE)
			{
				tree[CONFIG_MISC].ob_flags |= OF_HIDETREE;
				tree[CONFIG_KEYS].ob_flags &= ~OF_HIDETREE;
				tree[CONFIG_SIZES].ob_flags |= OF_HIDETREE;
				redraw_object(window->handle, tree, CONFIG_KEYS);
				wind_set_str(window->handle, WF_INFO, tree[CONFIG_MANAGER_INFO_KEYS].ob_spec.free_string);
				window->strindex = CONFIG_MANAGER_INFO_KEYS;
				strcpy(window->info, tree[window->strindex].ob_spec.free_string);
			}
			break;
		case CONFIG_SIDE_SIZES:
			if (tree[CONFIG_SIZES].ob_flags & OF_HIDETREE)
			{
				tree[CONFIG_MISC].ob_flags |= OF_HIDETREE;
				tree[CONFIG_KEYS].ob_flags |= OF_HIDETREE;
				tree[CONFIG_SIZES].ob_flags &= ~OF_HIDETREE;
				redraw_object(window->handle, tree, CONFIG_SIZES);
				wind_set_str(window->handle, WF_INFO, tree[CONFIG_MANAGER_INFO_SIZES].ob_spec.free_string);
				window->strindex = CONFIG_MANAGER_INFO_SIZES;
				strcpy(window->info, tree[window->strindex].ob_spec.free_string);
			}
			break;

			/* MISC sub-dialog. */
		case CONFIG_AUTOICON:
			if (tree[object].ob_state & OS_SELECTED)
			{
				change_state(window->handle, tree, CONFIG_AUTOICON_POP,   0);
				change_state(window->handle, tree, CONFIG_AUTOICON_EQUIV, 0);
			}
			else
			{
				change_state(window->handle, tree, CONFIG_AUTOICON_POP,   OS_DISABLED);
				change_state(window->handle, tree, CONFIG_AUTOICON_EQUIV, OS_DISABLED);
			}
			break;
		case CONFIG_AUTOICON_POP:
		case CONFIG_AUTOICON_EQUIV:
			if (tree[CONFIG_AUTOICON].ob_state & OS_SELECTED) // If option actually selected.
			{
				if ((item = handle_popup_combo(tree, autopop, window->handle, CONFIG_AUTOICON_POP, CONFIG_AUTOICON_STR)) != 0)
				{
					input.icontimeout = strtol(autopop[item].ob_spec.free_string, NULL, 10);
				}
			}
			break;

		/* ACTIONS sub-dialog. Handle popups in a smart fashion. Since we are using the same menu
		   for the 4 popups, we need to check/uncheck the current item for each one of them. Also,
		   we use a temp struct to avoid having the popups selections taking effect without the
		   user confirming with the Apply/OK buttons. */
		case CONFIG_DB_CLICK:
		case CONFIG_DB_CLICK_EQUIV:
			uncheck_prev_item(actionpop);

			if (input.dbclick != 0) {
				menu_icheck(actionpop, input.dbclick, CHECK);
			}
			else {
				menu_icheck(actionpop, conf->dbclick, CHECK);
			}

			if ((item = handle_popup_combo(tree, actionpop, window->handle, CONFIG_DB_CLICK, CONFIG_DB_CLICK_STR)) != 0)
			{
				input.dbclick = item;
			}
			break;
		case CONFIG_RMB_CLICK:
		case CONFIG_RMB_CLICK_EQUIV:
			uncheck_prev_item(actionpop);

			if (input.rmbclick != 0) {
				menu_icheck(actionpop, input.rmbclick, CHECK);
			}
			else {
				menu_icheck(actionpop, conf->rmbclick, CHECK);
			}
			
			if ((item = handle_popup_combo(tree, actionpop, window->handle, CONFIG_RMB_CLICK, CONFIG_RMB_CLICK_STR)) != 0)
			{
				input.rmbclick = item;
			}
			break;
		case CONFIG_ENTER:
		case CONFIG_ENTER_EQUIV:
			uncheck_prev_item(actionpop);
			
			if (input.enter != 0) {
				menu_icheck(actionpop, input.enter, CHECK);
			}
			else {
				menu_icheck(actionpop, conf->enter, CHECK);
			}

			if ((item = handle_popup_combo(tree, actionpop, window->handle, CONFIG_ENTER, CONFIG_ENTER_STR)) != 0)
			{
				input.enter = item;
			}
			break;
		case CONFIG_SPACE:
		case CONFIG_SPACE_EQUIV:
			uncheck_prev_item(actionpop);
			
			if (input.space != 0) {
				menu_icheck(actionpop, input.space, CHECK);
			}
			else {
				menu_icheck(actionpop, conf->space, CHECK);
			}
			
			if ((item = handle_popup_combo(tree, actionpop, window->handle, CONFIG_SPACE, CONFIG_SPACE_STR)) != 0)
			{
				input.space = item;
			}
			break;
		case CONFIG_INSERT:
		case CONFIG_INSERT_EQUIV:
			uncheck_prev_item(actionpop);

			if (input.insert != 0) {
				menu_icheck(actionpop, input.insert, CHECK);
			}
			else {
				menu_icheck(actionpop, conf->insert, CHECK);
			}
			
			if ((item = handle_popup_combo(tree, actionpop, window->handle, CONFIG_INSERT, CONFIG_INSERT_STR)) != 0)
			{
				input.insert = item;
			}
			break;
		case CONFIG_PATH:
			strcpy(path, conf->clipbrd);

			if (fsel_exinput(path, name, &button, get_gstring(STRINGS, STRINGS_SELECT_PATH)) == 1)
			{
				if (button == FSEL_OK) // Selector returned with a selection.
				{
					/* At least with XaAES, path may end with '*'. */
					char *p = strchr(path, '*');
					
					if (p != NULL) {
						*p = '\0';
					}
					snprintf(tree[CONFIG_PATH].ob_spec.free_string, 40, "%s", path);
					strcpy(manager->clipbrd, path);
				}
			}
			tree[object].ob_state &= ~OS_SELECTED;
			redraw_object(window->handle, tree, CONFIG_PATH_BOX);
			break;

		/* Buttons at the bottom of the dialog. */
		case CONFIG_HELP:
			if (manager->aeslang == AESLANG_GERMAN) {
				call_help(window->handle, tree, object, manager->avserver, "*:\\GEMCLIPD.HYP ", "Konfiguration", manager->system);
			}
			else {
				call_help(window->handle, tree, object, manager->avserver, "*:\\GEMCLIPE.HYP ", "Configuration", manager->system);
			}
			break;
		case CONFIG_DEFAULTS:
			reset_config(conf);
			init_config_form(conf, tree);
			tree[object].ob_state &= ~OS_SELECTED;
			send_redraw_message(window->handle);
			break;
		case CONFIG_CANCEL:
			strcpy(manager->clipbrd, conf->clipbrd);
			send_local_message(WM_CLOSED, window->handle, 0, 0, 0, 0);
			tree[object].ob_state &= ~OS_SELECTED;
			break;
		case CONFIG_APPLY:
		case CONFIG_OK:
			save_config(manager, window, conf, tree, object, &input);
			break;
		}
	}
	else // Right mouse click.
	{
		switch (object)
		{
		case CONFIG_SIDE_MISC:
			display_bubble_mouse(bubbles, CNFMISC, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_SIDE_KEYS:
			display_bubble_mouse(bubbles, CNFKEYS, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_SIDE_SIZES:
			display_bubble_mouse(bubbles, CNFLIMT, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_ICONIFY:
			display_bubble_mouse(bubbles, CNFICON, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_AUTOICON:
		case CONFIG_AUTOICON_POP:
		case CONFIG_AUTOICON_EQUIV:
			display_bubble_mouse(bubbles, CNFAUTO, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_SETTINGS:
			display_bubble_mouse(bubbles, CNFSETT, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_CYCLING:
			display_bubble_mouse(bubbles, CNFCYCL, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_BUBBLES:
			display_bubble_mouse(bubbles, CNFBUBL, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_CLEANING:
			display_bubble_mouse(bubbles, CNFMPTY, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_WARNINGS:
			display_bubble_mouse(bubbles, CNFWRNG, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_DB_CLICK:
		case CONFIG_DB_CLICK_EQUIV:
			display_bubble_mouse(bubbles, CNFDBCK, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_RMB_CLICK:
		case CONFIG_RMB_CLICK_EQUIV:
			display_bubble_mouse(bubbles, CNFRIGT, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_ENTER:
		case CONFIG_ENTER_EQUIV:
			display_bubble_mouse(bubbles, CNFRTRN, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_SPACE:
		case CONFIG_SPACE_EQUIV:
			display_bubble_mouse(bubbles, CNFSPAC, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_INSERT:
		case CONFIG_INSERT_EQUIV:
			display_bubble_mouse(bubbles, CNFINSR, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_SIZE:
			display_bubble_mouse(bubbles, CNFSIZE, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_CLIP_SIZE:
			display_bubble_mouse(bubbles, CNFCLIP, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_FREE_SPACE:
			display_bubble_mouse(bubbles, CNFFREE, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_PATH:
			display_bubble_mouse(bubbles, CNFPATH, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_OVERRIDE:
			display_bubble_mouse(bubbles, CNFOVER, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_OK:
			display_bubble_mouse(bubbles, CNFOKAY, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_CANCEL:
			display_bubble_mouse(bubbles, CNFCNCL, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_APPLY:
			display_bubble_mouse(bubbles, CNFAPLY, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_DEFAULTS:
			display_bubble_mouse(bubbles, CNFDFLT, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		case CONFIG_HELP:
			display_bubble_mouse(bubbles, CNFHELP, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
			break;
		}
	}
}

static void save_config(GEMCLIP *manager, WINDOW *window, USERCONF *conf, OBJECT *tree, short object, struct input *input)
{
	/* The tick buttons. */
	conf->iconify  = ((tree[CONFIG_ICONIFY].ob_state & OS_SELECTED)	 ? TRUE : FALSE);
	conf->autoicon = ((tree[CONFIG_AUTOICON].ob_state & OS_SELECTED) ? TRUE : FALSE);
	conf->cleaning = ((tree[CONFIG_CLEANING].ob_state & OS_SELECTED) ? TRUE : FALSE);
	conf->cycling  = ((tree[CONFIG_CYCLING].ob_state & OS_SELECTED)	 ? TRUE : FALSE);
	conf->bubbles  = ((tree[CONFIG_BUBBLES].ob_state & OS_SELECTED)	 ? TRUE : FALSE);
	conf->settings = ((tree[CONFIG_SETTINGS].ob_state & OS_SELECTED) ? TRUE : FALSE);
	conf->warnings = ((tree[CONFIG_WARNINGS].ob_state & OS_SELECTED) ? TRUE : FALSE);
	conf->override = ((tree[CONFIG_OVERRIDE].ob_state & OS_SELECTED) ? TRUE : FALSE);

	/* Popup menus. */
	conf->icontimeout = (input->icontimeout != 0) ? input->icontimeout : conf->icontimeout;
	conf->dbclick = (input->dbclick != 0) ? input->dbclick : conf->dbclick;
	conf->rmbclick = (input->rmbclick != 0) ? input->rmbclick : conf->rmbclick;
	conf->enter = (input->enter != 0) ? input->enter : conf->enter;
	conf->space = (input->space != 0) ? input->space : conf->space;
	conf->insert = (input->insert != 0) ? input->insert : conf->insert;

	/* Editable fields. */
	conf->size      = get_number_from_label(tree, CONFIG_SIZE);
	conf->clipsize  = get_number_from_label(tree, CONFIG_CLIP_SIZE);
	conf->freespace = get_number_from_label(tree, CONFIG_FREE_SPACE);
	
	/* If the size of the clipboard was reduced, some cleanup is needed. */
	resize_clipboard(manager, get_number_from_label(tree, CONFIG_SIZE));

	/* Clipboard path. If different, write the new path. */
	if ((strcmp(conf->clipbrd, manager->clipbrd)) != 0)
	{
		strcpy(conf->clipbrd, manager->clipbrd);
		
		if (conf->override == TRUE) {
			scrp_write(manager->clipbrd); // Set new clipboard path.
		}
	}

	write_config(manager->window, conf);
	tree[object].ob_state &= ~OS_SELECTED;

	if (object == CONFIG_OK) {
		send_local_message(WM_CLOSED, window->handle, 0, 0, 0, 0);
	}
	else {
		redraw_object(window->handle, tree, 0);
	}
}

/*
 * Handle tick button and associated popup objects.
 */
static int handle_popup_combo(OBJECT *tree, OBJECT *popup, short whandle, short object, short gstring)
{
	char buffer[tree[object].ob_spec.tedinfo->te_txtlen + 1];
	int item = 0;

	change_state(whandle, tree, gstring, OS_SELECTED);	// Highlight associated G_STRING.
	item = handle_popup(whandle, tree, popup, object, SCROLL_NO, buffer, sizeof buffer);
	change_state(whandle, tree, gstring, OS_NORMAL);	// Un-highlight associated G_STRING.

	return item;
}
