/*
 * GEMClip, a Clipboard Manager for FreeMiNT and MagiC.
 *
 * Copyright (C) 2011 Jean-Franois Lemaire (jflemaire@skynet.be)
 *
 * This file is part of GEMClip.
 *
 * This program is free software; redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <mint/cookie.h>
#include <mint/ostruct.h>
#include <mintbind.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "general.h"

#include "av.h"
#include "about.h"
#include "clipbrd.h"
#include "clips.h"
#include "config.h"
#include "info.h"
#include "manager.h"
#include "print.h"
#include "protocols.h"
#include "scroll.h"
#include "utils.h"

static void do_event_loop   (GEMCLIP *manager, WINDOW *windows[], USERCONF *conf, BUBBLES *bubbles, int *accterm);
static void handle_event    (short handle, GEMCLIP *manager, WINDOW *windows[], OBJECT *about, OBJECT *config, OBJECT *info, USERCONF *conf, BUBBLES *bubbles, short event, EVMULT_IN *evmult_in, EVMULT_OUT *evmult_out);
static void handle_shortcut (GEMCLIP *manager, WINDOW *windows[], EVMULT_OUT *evmult_out, USERCONF *conf);
static void handle_menuwin  (GEMCLIP *manager, WINDOW *windows[], USERCONF *conf, short title, short item, EVMULT_IN *evmult_in);

enum { NWINDOWS = 4 };

int main(void)
{
	short accid = -1;
	int accterm = FALSE;
	int rtrn	= EXIT_SUCCESS;
	int rscok	= FALSE;
	long cookie = 0;
	OBJECT *rmb = NULL;
	MENU rmbopt;
	MENU rmbhelp;

	WINDOW windvar[NWINDOWS]; // GEMClip uses 3 windows + the Manager.
	WINDOW *windows[NWINDOWS];
	CLIPS clips[NCLIPS];	  // 1 for scrap.txt, 99 for the clips.
	USERCONF conf;
	GEMCLIP manager;
	BUBBLES bubbles;

	const char *failure[] = { "Unable to run GEMClip.",
							  "GEMClip konnte nicht gestartet werden.",
							  "Impossible d'excuter GEMClip." };

	const char *norsc[]   = { "Resource file not found.",
							  "Resource-Datei nicht gefunden.",
							  "Fichier resource non trouv." };

	const char *badrsc[]  = { "Wrong resource file!",
							  "Falsche Resource-Datei!",
							  "Mauvais fichier resource!" };

	for (int i = 0; i < NWINDOWS; i++)
	{
		windows[i] = &windvar[i];
		windows[i]->open = FALSE;
	}

	manager.window = &windvar[WINAPP];

	for (int i = 0; i <= NCLIPS; i++)
	{
		manager.clips[i] = &clips[i];
		manager.clips[i]->index = 0;
		manager.clips[i]->select  = FALSE;
	}

	bubbles.p        = NULL;
	bubbles.buffer   = NULL;

	manager.tree	 = NULL;
	manager.menu	 = NULL;
	manager.menuid   = -1;
	manager.text	 = NULL;
	*manager.avserver = '\0';
	manager.suspend  = FALSE;
	manager.size     = LBUFFER;
	manager.fullsize = 0;
	manager.nclips   = 0;
	manager.selected = 0; // No clips selected.
	manager.clipn    = NOSCRAP;
	manager.topclip  = 1; // Start with 1st clip on top of Manager.
	strcpy(manager.clips[SCRAP]->name, "SCRAP.TXT");
	manager.clips[SCRAP]->index = SCRAP;
	manager.clips[SCRAP]->time.hour = 0;
	manager.clips[SCRAP]->time.minute = 0;
	manager.clips[SCRAP]->time.second = 0;
	manager.clips[SCRAP]->time.year = 0;
	manager.clips[SCRAP]->time.month = 0;
	manager.clips[SCRAP]->time.day = 0;
	*manager.clips[SCRAP]->preview = '\0';
	manager.clips[SCRAP]->size = 0;

	/* Get the language of the system from the _AKP cookie. We can't call
		appl_getinfo(AES_LANGUAGE, ...) since appl_init() hasn't been called yet.
		However, we need the language now. We do use the AES defines, though.
		Default is AESLANG_ENGLISH. */
	if(Getcookie(C__AKP, &cookie) == C_FOUND)
	{
		if ((cookie & 0x0F) == AESLANG_GERMAN)
		{
			manager.aeslang = AESLANG_GERMAN;
		}
		else if ((cookie & 0x0F) == AESLANG_FRENCH)
		{
			manager.aeslang = AESLANG_FRENCH;
		}
		else {
			manager.aeslang = AESLANG_ENGLISH;
		}
	}

	/* Get the formatting of date and time. */
	if(Getcookie(C__IDT, &cookie) == C_FOUND)
	{
		manager.datesep = manager.dformat = cookie;
		manager.datesep &= 0xFF;	// Date separator.
		manager.dformat &= 0xF00;	// Date format.
	}

	Pdomain(1); // GEMClip knows about long filenames.

	if ((accid = appl_init()) != -1)
	{
		/* Initialize Manager window here for the all duration of the session.
		   The workstation will be closed at program exit or system shutdown. */
		if ((manager.window = init_window(windows[WINAPP])) != NULL)
		{
			/* Load resource file from subdirectory 'gemclip'. English is the default for all
			   languages except German. Look in $ACCPATH too (mostly for TeraDesk). */
			char *env          = NULL;
			char path[LBUFFER] = {0};

			shel_envrn(&env, "ACCPATH=");

			if (manager.aeslang == AESLANG_GERMAN)
			{
				if (env != NULL ) { 
					snprintf(path, sizeof path, "%s%s%s", env, "\\", "GEMCLIP\\GEMCLIPD.RSC");
				}

				if ((rsrc_load("gemclip\\gemclipd.rsc") != 0) ||
				    (rsrc_load("GEMCLIP\\GEMCLIPD.RSC") != 0) ||
					(rsrc_load("gemclipd.rsc") != 0)          ||
					(rsrc_load("GEMCLIPD.RSC") != 0)          ||
					(rsrc_load(path) != 0))
				{
					rscok = TRUE;
				}
			}
			else
			{
				if (env != NULL) {
					snprintf(path, sizeof path, "%s%s%s", env, "\\", "GEMCLIP\\GEMCLIPE.RSC");
				}

				if ((rsrc_load("gemclip\\gemclipe.rsc") != 0) ||
				    (rsrc_load("GEMCLIP\\GEMCLIPE.RSC") != 0) ||
				    (rsrc_load("gemclipe.rsc") != 0)          ||
				    (rsrc_load("GEMCLIPE.RSC") != 0)          ||
				    (rsrc_load(path) != 0))
				{
					rscok = TRUE;
				}
			}

			if (rscok == TRUE)
			{
				/* Initialize Manager infoline. */
				strcpy(manager.window->info, get_gstring(STRINGS, STRINGS_EMPTY));

			    /* Have we loaded the correct version? */
				if ((strcmp(get_gstring(ABOUT, ABOUT_VERSION), VERSION)) == 0)
				{
			 		/* GEMClip needs MiNT or MagiC. No window menu for MagiC. */
					if (Ssystem(-1, 0L, 0L) == 0) {
			 			manager.system = MINT;
			 		}
			 		else if (Getcookie(C_MagX, &cookie) == C_FOUND) {
			 			manager.system = MAGIC;
			 		}

					if ((manager.system == MINT) || (manager.system == MAGIC))
					{
						char *accname = get_gstring(STRINGS, STRINGS_REGISTER);

						manager.tree = get_gem_tree(MANAGER);
						manager.menu = get_gem_tree(MENUWIN);
						manager.menuid = menu_register(accid, accname);

						if ((manager.tree != NULL) && (manager.menu != NULL))
						{
							if (init_bubbles(&bubbles, manager.aeslang, manager.system) == TRUE) // See protocols.c
							{
								init_config(&conf); // See config.c
								
								/* Tell the AES we understand AP_TERM. */
								shel_write(SWM_NEWMSG, NM_APTERM, 0, 0, 0);

								/* Attach sub-menus to right mouse button contextual menu. */
								rmb = get_gem_tree(RMB_MENU);

								rmbopt.mn_tree   = get_gem_tree(RMB_OPTIONS);
								rmbopt.mn_menu   = ROOT;
								rmbopt.mn_item   = 1;
								rmbopt.mn_scroll = SCROLL_NO;

								rmbhelp.mn_tree	 = get_gem_tree(RMB_HELP);
								rmbhelp.mn_menu	 = ROOT;
								rmbhelp.mn_item	 = 1;
								rmbhelp.mn_scroll = SCROLL_NO;

								menu_attach(ME_ATTACH, rmb, RMB_MENU_OPTIONS, &rmbopt);
								menu_attach(ME_ATTACH, rmb, RMB_MENU_HELP, &rmbhelp);

								menu_ienable(manager.menu, MENUWIN_CURRENT, DISABLE);
								menu_ienable(manager.menu, MENUWIN_DELETE, DISABLE);
								menu_icheck(get_gem_tree(POP_VIEWMODE), conf.viewmode, CHECK);

								/* Radio buttons don't need to be set manually later,
								  so we do it here (the AES will modify them later). */
								if (conf.clipmode == LIST_TEXT) {
									manager.tree[MANAGER_PREVIEW].ob_state |= OS_SELECTED;
								}
								else if (conf.clipmode == LIST_INFO) {
									manager.tree[MANAGER_PROPERTIES].ob_state |= OS_SELECTED;
								}

								if (conf.infomode == LIST_NAME) {
									manager.tree[MANAGER_NAME].ob_state |= OS_SELECTED;
								}
								else if (conf.infomode == LIST_DATE) {
									manager.tree[MANAGER_DATE].ob_state |= OS_SELECTED;
								}
								else if (conf.infomode == LIST_SIZE) {
									manager.tree[MANAGER_SIZE].ob_state |= OS_SELECTED;
								}

								/* Set userconf clipboard path now if needed. */
								if (conf.override == TRUE) {
									scrp_write(conf.clipbrd);
								}

								/* Delete all clips if wanted by the user. */
								if (conf.cleaning) {
									empty_clipboard(conf.clipbrd);
								}

								/* If we don't process the clipboard before entering the event loop,
								   the Manager won't have all needed info if iconify == TRUE,  */
								handle_clipboard(&manager, &conf);

								/* Wait for an AC_OPEN message while we poll the clipboard. */
								while (accterm == FALSE)
								{
									/* Process all events. Will only return when AP_TERM is received. */
				 					do_event_loop(&manager, windows, &conf, &bubbles, &accterm);
					 			}

								/* Quitting: destroy all opened windows after some special care. */
								if (windows[WINABT]->open == TRUE) {
									free_license(manager.text);
								}

								if (manager.window->open == TRUE) {
									av_protokoll_window(manager.window->handle, manager.avserver, AV_ACCWINDCLOSED);
								}

								write_config(manager.window, &conf); // Must be called before closing the Manager.

								for (int i = 0; i < NWINDOWS; i++)
								{
									if (windows[i]->open == TRUE)
									{
										v_clsvwk(windows[i]->wks);
										wind_close(windows[i]->handle);
										wind_delete(windows[i]->handle);
									}
								}

								if (*manager.avserver != '\0')
								{
									av_protokoll_exit(manager.avserver);
								}

								Mfree(bubbles.p);
								Mfree(bubbles.buffer);

								rsrc_free();
								appl_exit();
							}
							else
							{
								form_alert(1, FA_STOP "[Error initializing help bubbles.][Stop]");
								rsrc_free();
								appl_exit();
								rtrn = EXIT_FAILURE;
							}
						}
						else
						{
							form_alert(1, FA_STOP "[Error getting object trees!][Stop]");
							rsrc_free();
							appl_exit();
							rtrn = EXIT_FAILURE;
						}
					}
					 else
					 {
						form_alert(1, get_free_string(WN_SYSTEM));
						rsrc_free();
						appl_exit();
						rtrn = EXIT_FAILURE;
					 }
				}
				 else
				 {
					form_alert_string(1, FA_STOP, badrsc[manager.aeslang], "");
					rsrc_free();
					appl_exit();
					rtrn = EXIT_FAILURE;
				 }
			}
			else
			{
				form_alert_string(1, FA_STOP, norsc[manager.aeslang], "");
				appl_exit();
				rtrn = EXIT_FAILURE;
			}
		}
		else
		{
			form_alert_string(1, FA_STOP, failure[manager.aeslang], "");
			appl_exit();
			rtrn = EXIT_FAILURE;
		}
	}
	else {
		rtrn = EXIT_FAILURE;
	}

	return rtrn;
}

/*
 * GEM event loop. We start by watching for MU_MESAG and MU_TIMER only, until the manager window is
 * opened. Then we update the events watched. When the Manager is closed or inconified, we reset the
 * events flag to its original value.
 */
static void do_event_loop(GEMCLIP *manager, WINDOW *windows[], USERCONF *conf, BUBBLES *bubbles, int *accterm)
{
	const char *shortname = "Clips";
	short event   = 0;
	short msg[8]  = {0};
	short handle  = -1;
	short foo     = 0;
	int timer     = 0;
	int top       = FALSE;
	static int av = FALSE;
	GRECT rect;
	OBJECT *about         = get_gem_tree(ABOUT);
	OBJECT *icons         = get_gem_tree(ICONS);
	OBJECT *info          = get_gem_tree(INFOS);
	OBJECT *config        = get_gem_tree(CONFIG);
	EVMULT_OUT evmult_out = {0, {0, 0}, 0, 0, 0, 0};
	EVMULT_IN evmult_in   = {MU_MESAG | MU_TIMER,						// Events watched.
							 0x102, LEFT_BUTTON | RIGHT_BUTTON, 0x00,   // Buttons.
							 MO_ENTER, {-1, -1, 1, 1}, 					// Mouse rect 1.
							 MO_ENTER, {0, 0, 0, 0}, 					// Mouse rect 2.
							 1000, 0};									// Timer set to 1 sec.

	if (conf->iconify == TRUE) {
		send_local_message(AC_OPEN, 0, manager->menuid, 0, 0, 0);
	}

	do {
		event = evnt_multi_fast(&evmult_in, msg, &evmult_out);

		if (event & MU_MESAG)
		{
			switch (msg[0])
			{
			case AC_OPEN:
				if (manager->window->open == FALSE)
				{
					if (open_manager(manager, conf) == TRUE)
					{
						/* Update watched events. */
						evmult_in.emi_flags = MU_MESAG | MU_BUTTON | MU_TIMER | MU_KEYBD | MU_M1;
						set_mouse_rect(msg[3], &evmult_in);

						if (conf->iconify == TRUE)
						{
							wind_set(manager->window->handle, WF_ICONIFY, -1, -1, -1, -1);
							manager->window->icon = TRUE;
							wind_set_str(manager->window->handle, WF_NAME, shortname);
							conf->iconify = -1; // That's it with boot-iconification.
						}
					}
					else {
						*accterm = TRUE; // Serious error. Probably best to exit.
					}
				}
				else // Already opened. Bring Manager then any other window to the top or handle iconification.
				{
					wind_set(manager->window->handle, WF_TOP, 0, 0, 0, 0);

					/* Do not re-top the Manger by starting with the 2nd window. */
					for (int i = 1; i < NWINDOWS; i++)
					{
						if (windows[i]->open == TRUE) {
							wind_set(windows[i]->handle, WF_TOP, 0, 0, 0, 0);
						}
					}
				}
				break;
			case MN_SELECTED:
				handle_menuwin(manager, windows, conf, msg[3], msg[4], &evmult_in);
				break;
			case WM_TOPPED:
			case WM_NEWTOP:
				wind_set(msg[3], WF_TOP, 0, 0, 0, 0);
				set_mouse_rect(msg[3], &evmult_in);
				break;
			case WM_BOTTOM:
				wind_set(msg[3], WF_BOTTOM, 0, 0, 0, 0);
				break;
			case WM_REDRAW:
				if (msg[3] == manager->window->handle)
				{
					if (manager->window->icon == FALSE)
					{
						draw_window_form(manager->tree, msg[3], (GRECT*)&msg[4]);

						if (top == TRUE) // TRUE if uniconified with a LMB click.
						{
							wind_set(msg[3], WF_TOP, 0, 0, 0, 0);
							top = FALSE;
						}
					}
					else {
						update_icon_preview(manager);
					}
				}
				else if (msg[3] == windows[WINABT]->handle) // About window form.
				{
					if (windows[WINABT]->icon == FALSE)
					{
						draw_window_form(about, msg[3], (GRECT*)&msg[4]);
						draw_license(windows[WINABT], manager->text, about);
					}
					else {
						draw_icon(icons, ICONS_LOGO, icons[ICONS_LOGO].ob_width, icons[ICONS_LOGO].ob_height, msg[3]);
					}
				}
				else if (msg[3] == windows[WININF]->handle) // Information window form.
				{
					if (windows[WININF]->icon == FALSE) {
						draw_window_form(info, msg[3], (GRECT*)&msg[4]);
					}
					else {
						draw_icon(icons, ICONS_LOGO, icons[ICONS_LOGO].ob_width, icons[ICONS_LOGO].ob_height, msg[3]);
					}
				}
				else if (msg[3] == windows[WINCFG]->handle)// Config window form.
				{
					if (windows[WINCFG]->icon == FALSE) {
						draw_window_form(config, msg[3], (GRECT*)&msg[4]);
					}
                    else // Different icon according to the sub-form displayed.
                    {
                        if (config[CONFIG_SIDE_MISC].ob_state & OS_SELECTED) {
                            draw_icon(icons, ICONS_MISC, icons[ICONS_MISC].ob_width, icons[ICONS_MISC].ob_height, msg[3]);
                        }
                        else if (config[CONFIG_SIDE_KEYS].ob_state & OS_SELECTED) {
                            draw_icon(icons, ICONS_KEYS, icons[ICONS_KEYS].ob_width, icons[ICONS_KEYS].ob_height, msg[3]);
                        }
                        else if (config[CONFIG_SIDE_SIZES].ob_state & OS_SELECTED) {
                            draw_icon(icons, ICONS_SIZES, icons[ICONS_SIZES].ob_width, icons[ICONS_SIZES].ob_height, msg[3]);
                        }
                    }
				}
				break;
			case WM_MOVED:
				/* Set window to new position, get new work area, reposition object tree. */
				wind_set(msg[3], WF_CURRXYWH, msg[4], msg[5], msg[6], msg[7]);
				wind_get_grect(msg[3], WF_WORKXYWH, &rect);

				if (msg[3] == manager->window->handle)
				{
					manager->tree[ROOT].ob_x = rect.g_x;
					manager->tree[ROOT].ob_y = rect.g_y;

 					/* Keep win coords for config file, but only if window is not being hidden (off-screen). */
 					wind_get_grect(0, WF_WORKXYWH, &rect);

					if (msg[4] < rect.g_w) {
						conf->winx = msg[4];
					}
					if (msg[5] < rect.g_h) {
						conf->winy = msg[5];
					}
				}
				else if (msg[3] == windows[WINCFG]->handle)
				{
					config[ROOT].ob_x = rect.g_x;
					config[ROOT].ob_y = rect.g_y;
					manager->cfgrect.g_x = msg[4];
					manager->cfgrect.g_y = msg[5];
					manager->cfgrect.g_w = msg[6];
					manager->cfgrect.g_h = msg[7];
				}
				else if (msg[3] == windows[WININF]->handle)
				{
					info[ROOT].ob_x = rect.g_x;
					info[ROOT].ob_y = rect.g_y;
					manager->infrect.g_x = msg[4];
					manager->infrect.g_y = msg[5];
					manager->infrect.g_w = msg[6];
					manager->infrect.g_h = msg[7];
				}
				else if (msg[3] == windows[WINABT]->handle)
				{
					about[ROOT].ob_x = rect.g_x;
					about[ROOT].ob_y = rect.g_y;
				}
				set_mouse_rect(msg[3], &evmult_in);
				break;
			case WM_FULLED:
				{
					const char *str = NULL;
					OBJECT *popup   = get_gem_tree(POP_VIEWMODE);

					if (conf->viewmode == POP_VIEWMODE_MINIMAL)
					{
						resize_manager(manager, POP_VIEWMODE_STANDARD, conf);
						conf->viewmode = POP_VIEWMODE_STANDARD;
					}
					else if (conf->viewmode == POP_VIEWMODE_STANDARD)
					{
						resize_manager(manager, POP_VIEWMODE_MINIMAL, conf);
						conf->viewmode = POP_VIEWMODE_MINIMAL;
					}

					/* Get the current viewmode, write new string. */
					str = get_gstring(POP_VIEWMODE, conf->viewmode);
					snprintf(manager->tree[MANAGER_VIEWMODE_POP].ob_spec.tedinfo->te_ptext,
							 manager->tree[MANAGER_VIEWMODE_POP].ob_spec.tedinfo->te_txtlen,
							 "%s", &str[1]);
					uncheck_prev_item(popup);
					popup[conf->viewmode].ob_state |= OS_CHECKED;
					update_objects(manager);
				}
				break;
			case WM_ICONIFY:
				wind_set(msg[3], WF_ICONIFY, msg[4], msg[5], msg[6], msg[7]);

				if (msg[3] == manager->window->handle) {
					manager->window->icon = TRUE;
				}
				else if (msg[3] == windows[WINABT]->handle) {
					windows[WINABT]->icon = TRUE;
				}
				else if (msg[3] == windows[WININF]->handle) {
					windows[WININF]->icon = TRUE;
				}
				else {
					windows[WINCFG]->icon = TRUE;
				}
				wind_set_str(msg[3], WF_NAME, shortname);
				break;
			case WM_ALLICONIFY:
				// TODO.
				break;
			case WM_UNICONIFY:
				if (msg[3] == manager->window->handle)
				{
					if (msg[4] == -1) // If auto-iconify was used (see 'manager.c').
					{
						int height = 0;

						if (conf->viewmode != POP_VIEWMODE_STANDARD) {
							height = resize_manager(manager, conf->viewmode, conf);
						}
						else {
							height = manager->height;
						}

						wind_set(msg[3], WF_UNICONIFY, conf->winx, conf->winy, manager->tree[ROOT].ob_width, height);
					}
					else {
						wind_set(msg[3], WF_UNICONIFY, msg[4], msg[5], msg[6], msg[7]);
					}

 					/* Top window by default. Don't top on Control+click or RMB. See WM_REDRAW. */
 					if (( ! (evmult_out.emo_kmeta & K_CTRL)) || ( ! (evmult_out.emo_mbutton & RIGHT_BUTTON))) {
						top = TRUE;
					}

					wind_set_str(msg[3], WF_NAME, get_gstring(STRINGS, STRINGS_MANAGER_NAME));
					manager->window->icon = FALSE;
					evmult_in.emi_flags = MU_MESAG | MU_BUTTON | MU_TIMER | MU_KEYBD | MU_M1;
					set_mouse_rect(msg[3], &evmult_in);
				}
				else if (msg[3] == windows[WINABT]->handle)
				{
					wind_set(msg[3], WF_UNICONIFY, msg[4], msg[5], msg[6], msg[7]);
					wind_set_str(msg[3], WF_NAME, get_gstring(STRINGS, STRINGS_ABOUT_NAME));
					windows[WINABT]->icon = FALSE;
				}
				else if (msg[3] == windows[WININF]->handle)
				{
					wind_set(msg[3], WF_UNICONIFY, msg[4], msg[5], msg[6], msg[7]);
					wind_set_str(msg[3], WF_NAME, get_gstring(STRINGS, STRINGS_INFO_NAME));
					windows[WININF]->icon = FALSE;
				}
				else if (msg[3] == windows[WINCFG]->handle)
				{
    				wind_set(msg[3], WF_UNICONIFY, msg[4], msg[5], msg[6], msg[7]);
    	            wind_set_str(msg[3], WF_NAME, get_gstring(STRINGS, STRINGS_CONFIG_NAME));
					windows[WINCFG]->icon = FALSE;
				}
				break;
			case WM_CLOSED:
				/* Some specific precautions before closing and deleting the window. */
				if (msg[3] == windows[WINABT]->handle)
				{
					free_license(manager->text);
					v_clsvwk(windows[WINABT]->wks);
					windows[WINABT]->open = FALSE;
					windows[WINABT]->handle = -1;
					wind_close(msg[3]);
					wind_delete(msg[3]);
				}
				else if (msg[3] == windows[WINCFG]->handle)
				{
			        wind_get_grect(msg[3], WF_CURRXYWH, &manager->cfgrect);
					wind_set_str(msg[3], WF_NAME, get_gstring(STRINGS, windows[WINCFG]->strindex));
					v_clsvwk(windows[WINCFG]->wks);
					windows[WINCFG]->open = FALSE;
					windows[WINCFG]->handle = -1;
					wind_close(msg[3]);
					wind_delete(msg[3]);
				}
				else if (msg[3] == windows[WININF]->handle)
				{
			        wind_get_grect(msg[3], WF_CURRXYWH, &manager->infrect);
					v_clsvwk(windows[WININF]->wks);
					windows[WININF]->open = FALSE;
					windows[WININF]->handle = -1;
					wind_close(msg[3]);
					wind_delete(msg[3]);
				}
				else // Manager window.
				{
					/* Shift-clicking the CLOSE button quits. */
					if (evmult_out.emo_kmeta & (K_LSHIFT | K_RSHIFT)) {
						*accterm = TRUE;
					}
					else
					{
						write_config(manager->window, conf);
				        wind_get_grect(msg[3], WF_CURRXYWH, &manager->mgrrect);
						av_protokoll_window(manager->window->handle, manager->avserver, AV_ACCWINDCLOSED);
						manager->window->open = FALSE;
						manager->window->handle = -1;
						wind_close(msg[3]);
						wind_delete(msg[3]);
						evmult_in.emi_flags = MU_MESAG | MU_TIMER;
					}
				}

				/* Reset mouse rect to Manager if needed. */
				if (manager->window->open == TRUE) {
					set_mouse_rect(manager->window->handle, &evmult_in);
				}
				break;
			case SC_CHANGED:
				handle_clipboard(manager, conf);
				break;

			/* AV server answers. */
            case VA_PROTOSTATUS:
                /* Keep name of the AV server for later use. */
                strcpy(manager->avserver, (*(char **)&msg[6]));
                break;
			case AP_TERM:
				*accterm = TRUE;
				break;
			}
		}

		/* Since accs are launched before the desktop, we wait for a keyboard or button event before
		   registering with the AV server. Also, not everyone uses a desktop. */
		if (event & MU_BUTTON)
		{
			if (av == FALSE)
			{
				init_av_protokoll(manager->system, manager->avserver);
				av = TRUE;
			}

			handle = wind_find(evmult_out.emo_mouse.p_x, evmult_out.emo_mouse.p_y);
			handle_event(handle, manager, windows, about, config, info, conf, bubbles, event, &evmult_in, &evmult_out);
		}

		if (event & MU_KEYBD)
		{
			if (av == FALSE)
			{
				init_av_protokoll(manager->system, manager->avserver);
				av = TRUE;
			}

			wind_get(0, WF_TOP, &handle, &foo, &foo, &foo);

		 	if (handle == manager->window->handle)
		 	{
		 		/* Circumvent XaAES's keyboard navigation for keyboard moving/selecting within the list box. */
				if ((evmult_out.emo_kreturn == SC_UPARROW)      || (evmult_out.emo_kreturn == SC_DNARROW)      ||
				    (evmult_out.emo_kreturn == SC_LFARROW)      || (evmult_out.emo_kreturn == SC_RTARROW)      ||
				    (evmult_out.emo_kreturn == SC_SHFT_UPARROW) || (evmult_out.emo_kreturn == SC_SHFT_DNARROW) ||
				    (evmult_out.emo_kreturn == SC_CLRHOME)      || (evmult_out.emo_kreturn == SC_SHFT_CLRHOME) ||
				    (evmult_out.emo_kreturn == SC_CTRL_CLRHOME))
				{
					scroll_keyboard(manager, &evmult_out);
				}
				else
				{
					if (evmult_out.emo_kmeta & K_CTRL) /* Keyboard shortcut. */
					{
						handle_shortcut(manager, windows, &evmult_out, conf);
					}
					else {
						handle_event(handle, manager, windows, about, config, info, conf, bubbles, event, &evmult_in, &evmult_out);
					}
				}
			}
			else
			{
				if (evmult_out.emo_kmeta & K_CTRL) /* Keyboard shortcut. */
				{
					handle_shortcut(manager, windows, &evmult_out, conf);
				}
				else {
					handle_event(handle, manager, windows, about, config, info, conf, bubbles, event, &evmult_in, &evmult_out);
				}
			}
		}

		if (event & MU_M1) // We don't use the do_window_form() handler.
		{
			wind_get(0, WF_TOP, &handle, &foo, &foo, &foo);

		 	if (handle == manager->window->handle) {
				handle_mouse_rect(manager, manager->window, manager->tree, bubbles, &evmult_in, &evmult_out);
			}
			else if (handle == windows[WINABT]->handle) {
				handle_mouse_rect(manager, windows[WINABT], about, bubbles, &evmult_in, &evmult_out);
			}
			else if (handle == windows[WINCFG]->handle) {
				handle_mouse_rect(manager, windows[WINCFG], config, bubbles, &evmult_in, &evmult_out);
			}
			else if (handle == windows[WININF]->handle) {
				handle_mouse_rect(manager, windows[WININF], info, bubbles, &evmult_in, &evmult_out);
			}
		}

		if (event & MU_TIMER)
		{
			handle_clipboard(manager, conf);

			/* Increment time needed before auto-iconifying. */
			if ((conf->autoicon == TRUE) && (manager->window->open == TRUE))
			{
				if ((manager->window->icon == FALSE) && (timer++ >= (conf->icontimeout * 60)))
				{
					/* Do not autoiconify while there is user interaction. FIXME: not working. */
					if (( ! (event & MU_MESAG))  &&
					    ( ! (event & MU_BUTTON)) &&
					    ( ! (event & MU_KEYBD))  &&
					    ( ! (event & MU_M1)))
					{
						timer = 0;
						wind_set(manager->window->handle, WF_ICONIFY, -1, -1, -1, -1);
						wind_set_str(manager->window->handle, WF_NAME, shortname);
						manager->window->icon = TRUE;
					}
				}
			}
		}
	} while (*accterm == FALSE);
}

/*
 * React to mouse and keyboard events on window dialogs and call the handling functions.
 * Takes care of MU_BUTTON, MU_KEYBD and MU_M1 events (although differently). NOTE: is this
 * function really useful?
 */
static void handle_event(short handle,
						 GEMCLIP *manager,
						 WINDOW *windows[],
						 OBJECT *about,
						 OBJECT *config,
						 OBJECT *info,
						 USERCONF *conf,
						 BUBBLES *bubbles,
						 short event,
						 EVMULT_IN *evmult_in,
						 EVMULT_OUT *evmult_out)
{
	short object = 0;

 	if (handle == manager->window->handle)
	{
		if ((object = do_window_form(manager->tree, manager->window, evmult_out, event)) >= 0)
		{
			handle_manager(manager, conf, evmult_in, evmult_out, object, bubbles);
		}
	}
	else if (handle == windows[WINABT]->handle)
	{
		if ((object = do_window_form(about, windows[WINABT], evmult_out, event)) >= 0)
		{
			handle_about_form(manager, windows[WINABT], about, evmult_out, object, conf, bubbles);
		}
	}
	else if (handle == windows[WINCFG]->handle)
	{
		if ((object = do_window_form(config, windows[WINCFG], evmult_out, event)) >= 0)
		{
			handle_config_form(manager, windows[WINCFG], conf, config, evmult_out, object, bubbles);
		}
	}
	else if (handle == windows[WININF]->handle)
	{
		if ((object = do_window_form(info, windows[WININF], evmult_out, event)) >= 0)
		{
			handle_info_dialog(manager, windows[WININF], info, evmult_out, object, conf, bubbles);
		}
	}
}

/*
 * Process keyboard shortcuts, whatever window may be on top. With/without Shift key pressed.
 */
static void handle_shortcut(GEMCLIP *manager, WINDOW *windows[], EVMULT_OUT *evmult_out, USERCONF *conf)
{
	unsigned char key = get_ascii_value(evmult_out->emo_kreturn, evmult_out->emo_kmeta & (K_LSHIFT | K_RSHIFT));
	short foo    = 0;
	short handle = 0;

	wind_get(0, WF_TOP, &handle, &foo, &foo, &foo);

	if (handle == manager->window->handle) // Manager.
	{
		/* Standard keyboard shortcuts. */
		switch (key)
		{
		case 'w':
			if (conf->cycling == TRUE) {
				av_protokoll_send_key(manager->avserver, evmult_out->emo_kreturn >> 8);
			}
			break;
		case 'o':
			send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_VIEW, 0, 0, 0);
			break;
		case 'i':
			send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_INFO, 0, 0, 0);
			break;
		case 'p':
			send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_PRINT, 0, 0, 0);
			break;
		case 's':
			send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_SAVE, 0, 0, 0);
			break;
		case 'm':
			send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_APPEND, 0, 0, 0);
			break;
		case 'S':
			send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_APPEND_CURRENT, 0, 0, 0);
			break;
		case 'r':
			send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_MERGE, 0, 0, 0);
			break;
		case 'a':
			send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_ALL_SLOTS, 0, 0, 0);
			break;
		case 'A':
			send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_ALL_CLIPS, 0, 0, 0);
			break;
		case 'e':
			send_local_message(MN_SELECTED, MENUWIN_OPTIONS, MENUWIN_PREFS, 0, 0, 0);
			break;
		case 'u':
			/* Close the window, but only if it is not iconified. */
			if (manager->window->icon == FALSE) {
				send_local_message(WM_CLOSED, handle, 0, 0, 0, 0);
			}
			break;
		}

		/* Listbox navigation. Control key allows multiple selections. */
		switch (evmult_out->emo_kreturn)
		{
		case SC_DNARROW:
			scroll_keyboard(manager, evmult_out);
			break;
		}
	}
	else // Other windows only respond to the "Close" command (if not iconified).
	{
		if (key == 'u')
		{
			wind_get(0, WF_TOP, &handle, &foo, &foo, &foo);
			
			if ((handle == windows[WININF]->handle) && (windows[WININF]->icon == FALSE)) {
				send_local_message(WM_CLOSED, handle, 0, 0, 0, 0);
			}
			if ((handle == windows[WINCFG]->handle) && (windows[WINCFG]->icon == FALSE)) {
				send_local_message(WM_CLOSED, handle, 0, 0, 0, 0);
			}
			if ((handle == windows[WINABT]->handle) && (windows[WINABT]->icon == FALSE)) {
				send_local_message(WM_CLOSED, handle, 0, 0, 0, 0);
			}
		}
	}
}

/*
 * When the action asked applies to at least 12 clips (variable manager->selected), warn user before
 * proceding.
 */
static void handle_menuwin(GEMCLIP *manager, WINDOW *windows[], USERCONF *conf, short title, short item, EVMULT_IN *evmult_in)
{
	char path[LBUFFER] = {0};
	int size           = 0;

	switch (title)
	{
	case MENUWIN_SELECTION:
		switch (item)
		{
		case MENUWIN_VIEW:
			if (manager->selected >= WARNCLIPS)
			{
				if ((conf->warnings == FALSE) && (form_alert(2, get_free_string(WN_TWELVE))) == 1)
				{
					for (int clipn = 1; clipn <= manager->nclips; clipn++)
					{
						if (manager->clips[clipn]->select == TRUE)
						{
							 if ((size = write_clippath(path, sizeof path, manager->clipbrd, clipn)) > 0) {
								av_protokoll_open_file(path, size, NULL, manager->avserver, manager->system);
							}
						}
					}
				}
			}
			else if (manager->selected >= 1)
			{
				for (int clipn = 1; clipn <= manager->nclips; clipn++)
				{
					if (manager->clips[clipn]->select == TRUE)
					{
						if ((size = write_clippath(path, sizeof path, manager->clipbrd, clipn)) > 0) {
							av_protokoll_open_file(path, size, NULL, manager->avserver, manager->system);
						}
					}
				}
			}
			else if (manager->selected == 0)
			{
				if (manager->clipn == SCRAP)
				{
					if ((size = write_clippath(path, sizeof path, manager->clipbrd, SCRAP)) > 0) {
						av_protokoll_open_file(path, size, NULL, manager->avserver, manager->system);
					}
				}
			}
			break;
		case MENUWIN_INFO:
			if (windows[WININF]->open == FALSE)
			{
				windows[WININF]->open = open_info_dialog(windows[WININF], manager);
				set_mouse_rect(windows[WININF]->handle, evmult_in);
			}
			else // Already open. Bring it to the top, if needed, and update dialog.
			{
				OBJECT *tree = get_gem_tree(INFOS);

				if (manager->clipn == SCRAP) {
					process_info(manager, windows[WININF], tree);
				}
				else if (manager->clipn > SCRAP) {
					process_info(manager, windows[WININF], tree);
				}
				wind_set(windows[WININF]->handle, WF_TOP, windows[WININF]->handle, 0, 0, 0);
				redraw_object(windows[WININF]->handle, tree, ROOT);
			}
			break;
		case MENUWIN_PRINT:
			print_clips(manager);
			break;
		case MENUWIN_SAVE:				// Save to any file.
			save_clips(manager, manager->clipn);
			break;
		case MENUWIN_CURRENT:			// Overwrite 'scrap.txt'.
		case MENUWIN_APPEND:			// Append to any file.
		case MENUWIN_APPEND_CURRENT:	// Append to 'scrap.txt'.
			append_clips(manager, item, manager->clipn);
			break;
		case MENUWIN_MERGE:
			merge_clips(manager);
			break;
		case MENUWIN_ALL_SLOTS:
		case MENUWIN_ALL_CLIPS:
			select_all_clips(manager, item);
			break;
		case MENUWIN_DELETE:
			for (int i = MANAGER_CLIP01; i <= MANAGER_CLIP12; i++) {
				manager->tree[i].ob_state &= ~OS_SELECTED;
			}
			delete_clips(manager, conf);
			update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
			break;
		case MENUWIN_DICT:
			if (manager->system == MINT) {
				define_clips(manager->clips[manager->clipn]->preview);
			}
			break;
		}
		break;
	case MENUWIN_OPTIONS:
		switch (item)
		{
		case MENUWIN_SUSPEND:
			if (manager->suspend == TRUE)
			{
				menu_text(manager->menu, MENUWIN_SUSPEND, get_gstring(STRINGS, STRINGS_SUSPEND));
				menu_text(get_gem_tree(RMB_OPTIONS), RMB_OPTIONS_SUSPEND, get_gstring(STRINGS, STRINGS_SUSPEND));
				manager->suspend = FALSE;
			}
			else
			{
				menu_text(manager->menu, MENUWIN_SUSPEND, get_gstring(STRINGS, STRINGS_RESUME));
				menu_text(get_gem_tree(RMB_OPTIONS), RMB_OPTIONS_SUSPEND, get_gstring(STRINGS, STRINGS_RESUME));
				manager->suspend = TRUE;
			}
			break;
		case MENUWIN_PREFS:
			if (windows[WINCFG]->open == FALSE)
			{
				windows[WINCFG]->open = open_config_form(manager, windows[WINCFG], conf);
				set_mouse_rect(windows[WINCFG]->handle, evmult_in);
			}
			else {
				wind_set(windows[WINCFG]->handle, WF_TOP, windows[WINCFG]->handle, 0, 0, 0);
			}
			break;
		}
		break;
	case MENUWIN_HELP:
		switch (item)
		{
		case MENUWIN_DOC: // Open guide on title page.
			if (manager->aeslang == AESLANG_GERMAN) {
				call_help(manager->window->handle, manager->tree, 0, manager->avserver, "*:\\GEMCLIPD.HYP ", "", manager->system);
			}
			else {
				call_help(manager->window->handle, manager->tree, 0, manager->avserver, "*:\\GEMCLIPE.HYP ", "", manager->system);
			}
			break;
		case MENUWIN_WEB: // Open web browser on GEMClip's website.
			av_protokoll_open_file("http://gemdict.org/gemclip", 27, NULL, manager->avserver, manager->system);
			break;
		case MENUWIN_ABOUT:
			if (windows[WINABT]->open == FALSE) {
				windows[WINABT]->open = open_about_form(manager, windows[WINABT]);
			}
			else {	// Already opened. Bring it to the top.
				wind_set(windows[WINABT]->handle, WF_TOP, windows[WINABT]->handle, 0, 0, 0);
			}
			break;
		}
		break;
	}

	menu_tnormal(manager->menu, title, UNHIGHLIGHT);
	redraw_object(manager->window->handle, manager->menu, title); // menu_tnormal() does not trigger a redraw.
}
