/*
 * GEMClip, a Clipboard Manager for FreeMiNT and MagiC.
 *
 * Copyright (C) 2011 Jean-Franois Lemaire (jflemaire@skynet.be)
 *
 * This file is part of GEMClip.
 *
 * This program is free software; redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <mintbind.h>
#include <mint/ostruct.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "general.h"

#include "av.h"
#include "clipbrd.h"
#include "clips.h"
#include "info.h"
#include "manager.h"
#include "protocols.h"
#include "scroll.h"
#include "sort.h"
#include "utils.h"

static void handle_userconf_selection (EVMULT_OUT *evmult_out, int action);
static void handle_rmb_menu           (EVMULT_OUT *evmult_out);
static void switch_listbox_mode       (GEMCLIP *manager, CLIPS *clips[]);

/*
 * Open manager window after an AC_OPEN message.
 */
int open_manager(GEMCLIP *manager, USERCONF *conf)
{
	#define ATTRIBS ((manager->system == MINT) ? MINTATTRIBS : MAGICATTRIBS)
	static int reopen = FALSE;
	GRECT grect    = {conf->winx, conf->winy, manager->tree->ob_width, manager->tree->ob_height};
	GRECT calcrect = {0, 0, 0, 0};

	if (reopen == FALSE) // First opening.
	{
		wind_calc_grect(WC_BORDER, ATTRIBS, &grect, &calcrect);
	
		/* Keep track of the diff in height (the height of the title bar and infoline).
		   Needed for resizing the window. */
		manager->widgeth = calcrect.g_h - grect.g_h;
		manager->height = calcrect.g_h;
		reopen = TRUE;
		
		calcrect.g_y += manager->widgeth; //
	}
	else {
		calcrect = manager->mgrrect;
	}

	if ((manager->window->handle = wind_create_grect(ATTRIBS, &calcrect)) != 0)
	{
		if (manager->system == MINT) { // No window menu for MagiC.
			wind_set(manager->window->handle, WF_MENU, (int)((long)(manager->menu) >> 16), (int)(long)(manager->menu), 0, 0);
		}

		wind_set_str(manager->window->handle, WF_NAME, get_gstring(STRINGS, STRINGS_MANAGER_NAME));
		set_listbox_order(manager, conf);
		update_objects(manager);
		update_clips_infoline(manager); // Must be called before opening the window (wind_set_str() call).

		/* Get the Manager out of the way before iconifying it. */
		if (conf->iconify == TRUE)
		{
			wind_get_grect(0, WF_WORKXYWH, &grect); // Desktop coords.
			calcrect.g_x = -1; // With this we will know what to do when uniconifying.
			calcrect.g_y = grect.g_h + 100; // Open window off screen.
		}

		if (wind_open_grect(manager->window->handle, &calcrect) != 0)
		{
			const char *str = get_gstring(POP_VIEWMODE, conf->viewmode);
			int len = manager->tree[MANAGER_VIEWMODE_POP].ob_spec.tedinfo->te_txtlen;

			/* Skip the 1st leading space from the popup items. */
			snprintf(manager->tree[MANAGER_VIEWMODE_POP].ob_spec.tedinfo->te_ptext, len, "%s", &str[1]);
			av_protokoll_window(manager->window->handle, manager->avserver, AV_ACCWINDOPEN);
			update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
			adapt_slider(manager->window, manager->tree, manager->nclips, manager->topclip);
			manager->window->open = TRUE;

		}
	}
	return manager->window->open;
}

/*
 * Process button and keyboard events received by the Manager window. Beep when the user clicks on
 * a disable object with the left mouse button. A right click always displays a bubble.
 */
void handle_manager(GEMCLIP *manager, USERCONF *conf, EVMULT_IN *evmult_in, EVMULT_OUT *evmult_out, short object, BUBBLES *bubbles)
{
	if ((manager->tree[object].ob_state & OS_DISABLED) && (evmult_out->emo_mbutton & LEFT_BUTTON)) {
		sound_bell();
	}
	else
	{
		if (object == 0) // Key pressed.
		{
			switch (evmult_out->emo_kreturn)
			{
			case SC_DEL:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_DELETE, 0, 0, 0);
				break;
			case SC_HELP:
				send_local_message(MN_SELECTED, MENUWIN_HELP, MENUWIN_DOC, 0, 0, 0);
				break;
			case SC_RETURN:
			case SC_NMPAD_ENTER:
				handle_userconf_selection(evmult_out, conf->enter);
				break;
			case SC_SPACE:
				handle_userconf_selection(evmult_out, conf->space);
				break;
			case SC_INSERT:
				handle_userconf_selection(evmult_out, conf->insert);
				break;
			case SC_ESC:
				/* Deselect all list box slots. */
				manager->selected = 1; // Must be >= 1. Will be set to 0 by the next call.
				select_all_clips(manager, MENUWIN_ALL_CLIPS);
				break;
			}
		}
		else if ( ! (evmult_out->emo_mbutton & RIGHT_BUTTON))
		{
			switch (object) // Button clicked.
			{
			case MANAGER_CURRENT:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_CURRENT, 0, 0, 0);
				change_state(manager->window->handle, manager->tree, object, 0);
				manager->tree[manager->clipn + 1].ob_state &= ~OS_SELECTED; // Deselect list box item.
				break;
			case MANAGER_VIEW:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_VIEW, 0, 0, 0);
				change_state(manager->window->handle, manager->tree, object, 0);
				break;
 			/* Either save file or append to scrap.txt (with SHIFT key) or any file (with CTRL). */
 			case MANAGER_SAVE:
				if (evmult_out->emo_kmeta & (K_LSHIFT | K_RSHIFT)) {
					send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_APPEND_CURRENT, 0, 0, 0);
				}
				else if (evmult_out->emo_kmeta & K_CTRL) {
					send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_APPEND, 0, 0, 0);
				}
				else {
					send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_SAVE, 0, 0, 0);
				}
				change_state(manager->window->handle, manager->tree, object, 0);
				break;
			case MANAGER_DELETE:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_DELETE, 0, 0, 0);
				change_state(manager->window->handle, manager->tree, object, 0);
				break;
			case MANAGER_VIEWMODE_POP:
			case MANAGER_VIEWMODE_EQUIV:
				{
					char buffer[manager->tree[MANAGER_VIEWMODE_POP].ob_spec.tedinfo->te_txtlen];
					int viewmode = 0;
					OBJECT *pop  = get_gem_tree(POP_VIEWMODE);

					viewmode = handle_popup(manager->window->handle,
										    manager->tree,
										    pop,
										    MANAGER_VIEWMODE_POP,
										    SCROLL_NO,
										    buffer,
										    sizeof buffer);

					if (viewmode != 0)
					{
						resize_manager(manager, viewmode, conf);
						conf->viewmode = viewmode;
					}
				}
				update_objects(manager);
				break;
			/* Avoid unnecessary redraws by checking the variables aren't already set. */
			case MANAGER_PREVIEW:
				if (conf->clipmode != LIST_TEXT)
				{
					conf->clipmode = LIST_TEXT;
					switch_listbox_mode(manager, manager->clips);
				}
				break;
			case MANAGER_PROPERTIES:
				if (conf->clipmode != LIST_INFO)
				{
					conf->clipmode = LIST_INFO;
					switch_listbox_mode(manager, manager->clips);
				}
				break;
			case MANAGER_NAME:
				if (conf->infomode != LIST_NAME)
				{
					conf->infomode = LIST_NAME;
					reorder_listbox(manager, object);
					switch_listbox_mode(manager, manager->clips);
					update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
				}
				break;
			case MANAGER_DATE:
				if (conf->infomode != LIST_DATE)
				{
					conf->infomode = LIST_DATE;
					reorder_listbox(manager, object);
					switch_listbox_mode(manager, manager->clips);
					update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
				}
				break;
			case MANAGER_SIZE:
				if (conf->infomode != LIST_SIZE)
				{
					conf->infomode = LIST_SIZE;
					reorder_listbox(manager, object);
					switch_listbox_mode(manager, manager->clips);
					update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
				}
				break;
			case MANAGER_ORDER:
				conf->order = (conf->order == LIST_TOP) ? LIST_BOTTOM : LIST_TOP;
				reverse_order(manager);
				switch_listbox_mode(manager, manager->clips);
				update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
			break;
			case MANAGER_CLIP_ICON:
				/* Shift-click the icon to save scrap.txt. */
				if (evmult_out->emo_kmeta & (K_LSHIFT | K_RSHIFT)){
					save_clips(manager, SCRAP);
				}
				else
				{
					char path[LBUFFER] = {0};
					size_t size        = 0;
					size = write_clippath(path, sizeof path, manager->clipbrd, SCRAP);
					av_protokoll_open_file(path, size, NULL, manager->avserver, manager->system);
				}
				change_state(manager->window->handle, manager->tree, object, 0);
				break;
			case MANAGER_CLIP01:
			case MANAGER_CLIP02:
			case MANAGER_CLIP03:
			case MANAGER_CLIP04:
			case MANAGER_CLIP05:
			case MANAGER_CLIP06:
			case MANAGER_CLIP07:
			case MANAGER_CLIP08:
			case MANAGER_CLIP09:
			case MANAGER_CLIP10:
			case MANAGER_CLIP11:
			case MANAGER_CLIP12:
				if (evmult_out->emo_mclicks == 2) // DB-click.
				{
					handle_userconf_selection(evmult_out, conf->dbclick);
				}
				else if (evmult_out->emo_mbutton & RIGHT_BUTTON)
				{
					handle_userconf_selection(evmult_out, conf->rmbclick);
				}
				else // Standard LMB click.
				{
					manager->clipn = handle_selection(manager, object, evmult_out);
				}
				update_clips_infoline(manager);
				update_preview(manager);
				update_objects(manager);
				break;
			/* Even if we redraw via the VDI we still need to update the listbox. */
			case MANAGER_UPARROW:
				scroll_uparrow(manager, evmult_out);
				break;
			case MANAGER_DNARROW:
				scroll_dnarrow(manager, evmult_out);
				break;
			case MANAGER_SLIDER:
				scroll_slider(manager, evmult_in, object);
				break;
			case MANAGER_SCROLLBAR:
				scroll_page(manager, evmult_out);
				update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
				update_listbox_slots(manager);
				update_listbox(manager);
				redraw_object(manager->window->handle, manager->tree, MANAGER_LISTBOX);
				break;
			}
		}
		else // Right button.
		{
			switch (object)
			{
			/* Open contextual menu. Don't display help bubble. */
			case MANAGER_CLIP01:
			case MANAGER_CLIP02:
			case MANAGER_CLIP03:
			case MANAGER_CLIP04:
			case MANAGER_CLIP05:
			case MANAGER_CLIP06:
			case MANAGER_CLIP07:
			case MANAGER_CLIP08:
			case MANAGER_CLIP09:
			case MANAGER_CLIP10:
			case MANAGER_CLIP11:
			case MANAGER_CLIP12:
				handle_userconf_selection(evmult_out, conf->rmbclick);
				break;
			/* Do as XaAES: reverse order in case or right click. */
			case MANAGER_UPARROW:
				scroll_dnarrow(manager, evmult_out);
				break;
			case MANAGER_DNARROW:
				scroll_uparrow(manager, evmult_out);
			case MANAGER_SCROLLBAR:
				scroll_page(manager, evmult_out);
				update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
				update_listbox_slots(manager);
				update_listbox(manager);
				redraw_object(manager->window->handle, manager->tree, MANAGER_LISTBOX);
				break;
			/* Display help bubble for the other objects. */
			case MANAGER_CURRENT:
				display_bubble_mouse(bubbles, MGRCURR, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case MANAGER_VIEW:
				display_bubble_mouse(bubbles, MGRVIEW, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
 			case MANAGER_SAVE:
				display_bubble_mouse(bubbles, MGRSAVE, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case MANAGER_DELETE:
				display_bubble_mouse(bubbles, MGRDELE, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case MANAGER_VIEWMODE_POP:
			case MANAGER_VIEWMODE_EQUIV:
				display_bubble_mouse(bubbles, MGRMODE, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case MANAGER_PREVIEW:
				display_bubble_mouse(bubbles, MGRTEXT, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case MANAGER_PROPERTIES:
				display_bubble_mouse(bubbles, MGRINFO, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case MANAGER_NAME:
				display_bubble_mouse(bubbles, MGRNAME, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case MANAGER_DATE:
				display_bubble_mouse(bubbles, MGRDATE, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case MANAGER_SIZE:
				display_bubble_mouse(bubbles, MGRSIZE, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case MANAGER_ORDER:
				display_bubble_mouse(bubbles, MGRORDR, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case MANAGER_CLIP_ICON:
				display_bubble_mouse(bubbles, MGRICON, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case MANAGER_CLIPBOARD:
				display_bubble_mouse(bubbles, MGRSCRP, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			case MANAGER_PREV1:
			case MANAGER_PREV2:
			case MANAGER_PREV3:
			case MANAGER_PREV4:
			case MANAGER_PREV5:
			case MANAGER_PREV6:
				display_bubble_mouse(bubbles, MGRPREV, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, conf);
				break;
			}
		}
	}
}

/*
 * Handle single/multiple selections in the Manager list box. Making it into its own function seems
 * easier to maintain, even though it is not called anywhere else than in handle_manager().
 * ==> Return selected clip index or SCRAP.
 */
int handle_selection(GEMCLIP *manager, short object, EVMULT_OUT *evmult_out)
{
	int offset = MANAGER_CLIP_N01 - MANAGER_CLIP01; // Offset betwen slots and left side index #.
	int idx    = 0;
	long clipn = 0;

	idx = object + offset;
	clipn = strtol(manager->tree[idx].ob_spec.tedinfo->te_ptext, NULL, 10);

	if (evmult_out->emo_kmeta & (K_LSHIFT | K_RSHIFT)) // Add selection to array.
	{
		if (! (manager->tree[object].ob_state & OS_SELECTED)) // Not if de-selecting.
		{
			redraw_object(manager->window->handle, manager->tree, object);
			manager->selected--;
		}
		else
		{
			manager->clips[clipn]->select = TRUE;
			manager->selected++;
		}
	}
	else // Unique selection.
	{
		/*  Disable all previously selected clips... */
		for (idx = 1; idx <= manager->nclips; idx++)
		{
			manager->clips[idx]->select = FALSE;
		}

		/* ...and listbox slots. */
		for (idx = MANAGER_CLIP01; idx <= MANAGER_CLIP12; idx++)
		{
			if (idx != object) {
				manager->tree[idx].ob_state &= ~OS_SELECTED;
			}
		}

		redraw_object(manager->window->handle, manager->tree, MANAGER_LISTBOX);

		/* If de-selecting a previously selected slot, default to scrapt.txt preview. */
		if (manager->tree[object].ob_state & OS_SELECTED)
		{
			manager->clips[clipn]->select = TRUE;
			manager->selected = 1;
		}
		else
		{
			redraw_object(manager->window->handle, manager->tree, object);
			manager->clips[clipn]->select = FALSE;
			manager->selected = 0;
			clipn = SCRAP;
		}
	}
	return clipn;
}

/*
 * Handle RMB/DB-click and some such events according to the user config.
 */
static void handle_userconf_selection(EVMULT_OUT *evmult_out, int action)
{
	switch (action)
	{
	case ACTN_MENU:
		handle_rmb_menu(evmult_out);
		break;
	case ACTN_CURRENT:
		send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_CURRENT, 0, 0, 0);
		break;
	case ACTN_INFO:
		send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_INFO, 0, 0, 0);
		break;
	case ACTN_VIEW:
		send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_VIEW, 0, 0, 0);
		break;
	}
}

/*
 * Handle contextual menu selections.
 */
static void handle_rmb_menu(EVMULT_OUT *evmult_out)
{
	OBJECT *rmbopt  = get_gem_tree(RMB_OPTIONS);
	OBJECT *rmbhelp = get_gem_tree(RMB_HELP);
	MENU menu;
	MENU mdata;

	menu.mn_tree   = get_gem_tree(RMB_MENU);
	menu.mn_menu   = ROOT;
	menu.mn_item   = 1;
	menu.mn_scroll = SCROLL_NO;

	if (menu_popup(&menu, evmult_out->emo_mouse.p_x, evmult_out->emo_mouse.p_y, &mdata) == 1)
	{
        if (mdata.mn_tree == menu.mn_tree) // If pop up menu.
		{
			switch (mdata.mn_item)
			{
			case RMB_MENU_CURRENT:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_CURRENT, 0, 0, 0);
				break;
			case RMB_MENU_VIEW:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_VIEW, 0, 0, 0);
				break;
			case RMB_MENU_INFO:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_INFO, 0, 0, 0);
				break;
			case RMB_MENU_DELETE:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_DELETE, 0, 0, 0);
				break;
			case RMB_MENU_PRINT:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_PRINT, 0, 0, 0);
				break;
			case RMB_MENU_SAVE:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_SAVE, 0, 0, 0);
				break;
			case RMB_MENU_APPEND:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_APPEND, 0, 0, 0);
				break;
			case RMB_MENU_APPEND_CURRENT:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_APPEND_CURRENT, 0, 0, 0);
				break;
			case RMB_MENU_MERGE:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_MERGE, 0, 0, 0);
				break;
			case RMB_MENU_SLOTS:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_ALL_SLOTS, 0, 0, 0);
				break;
			case RMB_MENU_SELECT:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_ALL_CLIPS, 0, 0, 0);
				break;
			case RMB_MENU_DICT:
				send_local_message(MN_SELECTED, MENUWIN_SELECTION, MENUWIN_DICT, 0, 0, 0);
				break;
			}
		}
        else if (mdata.mn_tree == rmbopt) // If options sub-menu.
		{
			switch (mdata.mn_item)
			{
			case RMB_OPTIONS_SUSPEND:
				send_local_message(MN_SELECTED, MENUWIN_OPTIONS, MENUWIN_SUSPEND, 0, 0, 0);
				break;
			case RMB_OPTIONS_PREFS:
				send_local_message(MN_SELECTED, MENUWIN_OPTIONS, MENUWIN_PREFS, 0, 0, 0);
				break;
			}
		}
		else if (mdata.mn_tree == rmbhelp) // If help sub-menu.
		{
			switch (mdata.mn_item)
			{
			case RMB_HELP_DOC:
				send_local_message(MN_SELECTED, MENUWIN_HELP, MENUWIN_DOC, 0, 0, 0);
				break;
			case RMB_HELP_WEB:
				send_local_message(MN_SELECTED, MENUWIN_HELP, MENUWIN_WEB, 0, 0, 0);
				break;
			case RMB_HELP_ABOUT:
				send_local_message(MN_SELECTED, MENUWIN_HELP, MENUWIN_ABOUT, 0, 0, 0);
				break;
			}
		}
	}
}


/*
 * Update preview slots in the list box. May be called either when the Manager is open or not.
 */
void update_listbox(GEMCLIP *manager)
{
	enum { OFFSET = (MANAGER_CLIP_N01 - MANAGER_CLIP01) }; // Offset betwen slots and left side index #.

	OBJECT *tree = manager->tree;
	char *bytes  = get_gstring(STRINGS, STRINGS_BYTES);
	int slot	 = MANAGER_CLIP01;
	int clip	 = manager->topclip;
	int len 	 = tree[MANAGER_CLIP01].ob_spec.tedinfo->te_txtlen;
	int dformat  = manager->dformat;
	int datesep  = manager->datesep;
	CLIPS *clips = *manager->clips;

	if (manager->nclips > SCRAP) // We must at least have one clip file to display.
	{
		while ((slot <= MANAGER_CLIP12) && (clip <= manager->nclips))
		{
			/* Enable both slots and left side indexes. */
			tree[slot].ob_state &= ~OS_DISABLED;
			tree[slot + OFFSET].ob_state &= ~OS_DISABLED;

			strcpy(tree[slot].ob_spec.tedinfo->te_ptext, clips[clip].preview);

			/* 'Info' mode selected in the Manager. */
			if (tree[MANAGER_PROPERTIES].ob_state & OS_SELECTED) {
				write_date_info(tree[slot].ob_spec.tedinfo->te_ptext, &clips[clip], dformat, datesep, bytes, len);
			}
			slot++;
			clip++;
		}

		/* Disable unused slots and left side indexes. */
		if (manager->nclips < 12)
		{
			for (slot = MANAGER_CLIP01 + manager->nclips; slot <= MANAGER_CLIP12; slot++)
			{
				*tree[slot].ob_spec.tedinfo->te_ptext = '\0';
				tree[slot].ob_state |= OS_DISABLED;
				tree[slot + OFFSET].ob_state |= OS_DISABLED;
			}
		}

		if (manager->window->open == TRUE) {
			adapt_slider(manager->window, manager->tree, manager->nclips, manager->topclip);
		}
	}
	else // Disable and empty all slots.
	{
		for (slot = MANAGER_CLIP01; slot <= MANAGER_CLIP12; slot++)
		{
			*tree[slot].ob_spec.tedinfo->te_ptext = '\0';
			tree[slot].ob_state &= ~OS_SELECTED;
			tree[slot].ob_state |= OS_DISABLED;
			tree[slot + OFFSET].ob_state |= OS_DISABLED;
		}
	}
}

/*
 * Update listbox slots according to the 'select' variable. No redraw happens in this function so
 * it has to be called before functions responsible for redrawing the listbox.
 */
void update_listbox_slots(GEMCLIP *manager)
{
	int slot = MANAGER_CLIP01;
	int clip = manager->topclip; // 1st visible clip in the listbox.

	/* Unselect all slots. */
	while (slot <= MANAGER_CLIP12) {
		manager->tree[slot++].ob_state &= ~OS_SELECTED;
	}

	while (clip <= manager->nclips)
	{
		if (manager->clips[clip]->select == TRUE)
		{
			slot = clip - manager->topclip + 1;

			/* This shouldn't happen, but it happened that an object from the preview box got selected. */
			if (((slot + 1) >= MANAGER_CLIP01) && ((slot + 1) <= MANAGER_CLIP12)) {
				manager->tree[slot + 1].ob_state |= OS_SELECTED;
			}
		}
		clip++;
	}
}

/*
 * Switch between the 2 list box display modes: Preview or Properties.
 */
static void switch_listbox_mode(GEMCLIP *manager, CLIPS *clips[])
{
	char *bytes = get_gstring(STRINGS, STRINGS_BYTES);
	int txtlen	= 0;
	int n		= manager->topclip;

	if (manager->clipn != NOSCRAP)
	{
		txtlen = manager->tree[MANAGER_CLIPBOARD].ob_spec.tedinfo->te_txtlen;

		/* Update clipboard preview, then list box. */
		if (manager->tree[MANAGER_PREVIEW].ob_state & OS_SELECTED)
		{
			snprintf(manager->tree[MANAGER_CLIPBOARD].ob_spec.tedinfo->te_ptext, txtlen, "%s", clips[SCRAP]->preview);
		}
		else if (manager->tree[MANAGER_PROPERTIES].ob_state & OS_SELECTED)
		{
			write_date_info(manager->tree[MANAGER_CLIPBOARD].ob_spec.tedinfo->te_ptext, clips[SCRAP],
							manager->dformat, manager->datesep, bytes, txtlen);
		}

		txtlen = manager->tree[MANAGER_CLIP01].ob_spec.tedinfo->te_txtlen;

		for (int i = MANAGER_CLIP01; i <= MANAGER_CLIP12; i++, n++)
		{
			if (*manager->tree[i].ob_spec.tedinfo->te_ptext != '\0')
			{
				if (manager->tree[MANAGER_PREVIEW].ob_state & OS_SELECTED)
				{
					snprintf(manager->tree[i].ob_spec.tedinfo->te_ptext, txtlen, "%s", clips[n]->preview);
				}
				else if (manager->tree[MANAGER_PROPERTIES].ob_state & OS_SELECTED)
				{
					write_date_info(manager->tree[i].ob_spec.tedinfo->te_ptext, clips[n],
									manager->dformat, manager->datesep, bytes, txtlen);
				}
			}
		}
		redraw_object(manager->window->handle, manager->tree, MANAGER_LISTBOX);
		redraw_object(manager->window->handle, manager->tree, MANAGER_CLIPBOARD);
	}
}

/*
 * Set listbox sorting order according to the user config.
 */
void set_listbox_order(GEMCLIP *manager, USERCONF *conf)
{
	if (conf->infomode == LIST_NAME)
	{
		reorder_listbox(manager, MANAGER_NAME);
		switch_listbox_mode(manager, manager->clips);
	}
	else if (conf->infomode == LIST_DATE)
	{
		reorder_listbox(manager, MANAGER_DATE);
		switch_listbox_mode(manager, manager->clips);
	}
	else if (conf->infomode == LIST_SIZE)
	{
		reorder_listbox(manager, MANAGER_SIZE);
		switch_listbox_mode(manager, manager->clips);
	}

	if (conf->order == LIST_TOP) {
		manager->tree[MANAGER_ORDER].ob_state &= ~OS_SELECTED;
	}
	else if (conf->order == LIST_BOTTOM) {
		manager->tree[MANAGER_ORDER].ob_state |= OS_SELECTED;
	}
	reset_order(manager, conf);
	update_listbox(manager);
}

/*
 * Write date and time info in the Manager when in "Info" mode and in the Information dialog.
 */
void write_date_info(char *str, CLIPS *clip, int dformat, int datesep, const char *bytes, int txtlen)
{
	if (dformat == 0)
	{
		snprintf(str, txtlen, "%02d%c%02d%c%d, %02d:%02d:%02d, %6ld %s",
		clip->time.month, datesep, clip->time.day, datesep, clip->time.year + 1980,
		clip->time.hour, clip->time.minute, clip->time.second * 2,
		clip->size, bytes);
	}
	else if (dformat == 256)
	{
		snprintf(str, txtlen, "%02d%c%02d%c%d, %02d:%02d:%02d, %6ld %s",
		clip->time.day, datesep, clip->time.month, datesep, clip->time.year + 1980,
		clip->time.hour, clip->time.minute, clip->time.second * 2,
		clip->size, bytes);
	}
	else if (dformat == 512)
	{
		snprintf(str, txtlen, "%d%c%02d%c%02d, %02d:%02d:%02d, %6ld %s",
		clip->time.year + 1980, datesep, clip->time.month, datesep, clip->time.day,
		clip->time.hour, clip->time.minute, clip->time.second * 2,
		clip->size, bytes);
	}
	else if (dformat == 768)
	{
		snprintf(str, txtlen, "%d%c%02d%c%02d, %02d:%02d:%02d, %6ld %s",
		clip->time.year + 1980, datesep, clip->time.day, datesep, clip->time.month,
		clip->time.hour, clip->time.minute, clip->time.second * 2,
		clip->size, bytes);
	}
}

/*
 * Update Manager infoline with the number of clips.
 */
void update_clips_infoline(GEMCLIP *manager)
{
	char buffer[MBUFFER] = {0};
	char *clips = NULL;
	char *bytes = get_gstring(STRINGS, STRINGS_BYTES);
	char *selec = get_gstring(STRINGS, STRINGS_SELECTED);

	if ((selection(manager->clips) == FALSE) && (manager->nclips == 0)) { // If clipboard empty.
		wind_set_str(manager->window->handle, WF_INFO, get_gstring(STRINGS, STRINGS_EMPTY));
	}
	else
	{
		if (manager->nclips <= 1) {
			clips = get_gstring(STRINGS, STRINGS_CLIPS_S); // Singular form.
		}
		else {
			clips = get_gstring(STRINGS, STRINGS_CLIPS_P); // Plural.
		}

		snprintf(buffer, sizeof buffer,
				 "%d %s, %d %s, %d %s",
				 manager->nclips, clips, manager->fullsize, bytes, manager->selected, selec);

		wind_set_str(manager->window->handle, WF_INFO, buffer);
		strcpy(manager->window->info, buffer); // Keep new infoline.
	}
}

/*
 * Update the left side indices of the Manager.
 */
void update_indices(CLIPS *clips[], int topclip, OBJECT *tree, short handle)
{
	for (int idx = MANAGER_CLIP_N01; idx <= MANAGER_CLIP_N12; idx++, topclip++)
	{
		if (clips[topclip]->index != 0)
		{
			tree[idx].ob_state &= ~OS_DISABLED;
			snprintf(tree[idx].ob_spec.tedinfo->te_ptext, 3, "%2d", clips[topclip]->index);
		}
		else {
			tree[idx].ob_state |= OS_DISABLED;
		}
	}
	redraw_object(handle, tree, MANAGER_CLIP_NO);
}

/*
 * Update the bottom preview box when an item is selected in the Manager listbox.
 */
void update_preview(GEMCLIP *manager)
{
	char preview[(manager->tree[MANAGER_PREV1].ob_spec.tedinfo->te_txtlen + 1) * 6]; // Preview box.
	char path[LBUFFER] = {0};
	int file   = -1;

	memset(preview, 0, sizeof preview);
	write_clippath(path, sizeof path, manager->clipbrd, manager->clipn);

	if ((file = Fopen(path, 0)) >= 0)
	{
		if (Fread(file, sizeof preview, preview) > 0)
		{
			remove_whitespace(preview);
			update_bottom_preview(manager->tree, preview);
		}
		else {
			form_alert(1, FA_ERROR "[Error reading file|in update_preview().][OK]");
		}
		
		if (Fclose(file) < 0) {
			form_alert(1, FA_ERROR "[Error closing file|in update_preview().][OK]");
		}
	}
	else if (manager->clipn != SCRAP) { // SCRAP.TXT may be missing.
		form_alert(1, FA_ERROR "[Error opening file|in update_preview().][OK]");
	}
	redraw_object(manager->window->handle, manager->tree, MANAGER_PREVIEW_BOX);
}

void update_bottom_preview(OBJECT *tree, const char *preview)
{
	short txtlen = tree[MANAGER_PREV1].ob_spec.tedinfo->te_txtlen;
	int pos 	 = 0;

	/* These object in the resource MUST be in incremental order. */
	for (int i = MANAGER_PREV1; i <= MANAGER_PREV6; i++)
	{
		if (snprintf(tree[i].ob_spec.tedinfo->te_ptext, txtlen, "%s", &preview[pos]) <= 0)	{
			*tree[i].ob_spec.tedinfo->te_ptext = '\0';
		}
		pos += (txtlen - 1); // -1 for the terminating character, since we are overwriting it.
	}
}


/*
 * Update enabled/disabled status of GEM objects according to which clip(s) is(are)
 * currently selected. Redraw toolbar objects if window is open.
 */
void update_objects(GEMCLIP *manager)
{
	OBJECT *popup = get_gem_tree(RMB_MENU);

	/* SCRAP.TXT selected. Might be a multiple selection. */
	if (manager->clipn == SCRAP)
	{
		menu_ienable(manager->menu, MENUWIN_CURRENT, DISABLE);
		menu_ienable(manager->menu, MENUWIN_VIEW, ENABLE);
		menu_ienable(manager->menu, MENUWIN_INFO, ENABLE);
		menu_ienable(manager->menu, MENUWIN_DELETE, DISABLE);
		//menu_ienable(manager->menu, MENUWIN_PRINT, ENABLE);
		menu_ienable(manager->menu, MENUWIN_SAVE, ENABLE);
		menu_ienable(manager->menu, MENUWIN_APPEND, ENABLE);
		menu_ienable(manager->menu, MENUWIN_APPEND_CURRENT, ENABLE);
		menu_ienable(manager->menu, MENUWIN_MERGE, DISABLE);

		/* Popup menu. */
		popup[RMB_MENU_CURRENT].ob_state |= OS_DISABLED;
		popup[RMB_MENU_DELETE].ob_state |= OS_DISABLED;
		popup[RMB_MENU_MERGE].ob_state |= OS_DISABLED;

		if (manager->selected == 0) // Only SCRAP.TXT selected.
		{
			manager->tree[MANAGER_CURRENT].ob_state |= OS_DISABLED;
			manager->tree[MANAGER_DELETE].ob_state |= OS_DISABLED;
		}
		else
		{
			manager->tree[MANAGER_CURRENT].ob_state &= ~OS_DISABLED;
			manager->tree[MANAGER_DELETE].ob_state &= ~OS_DISABLED;
		}
	}
	else if (manager->clipn == NOSCRAP) // No SCRAP.TXT.
	{
		menu_ienable(manager->menu, MENUWIN_CURRENT, DISABLE);
		menu_ienable(manager->menu, MENUWIN_VIEW, DISABLE);
		menu_ienable(manager->menu, MENUWIN_INFO, DISABLE);
		//menu_ienable(manager->menu, MENUWIN_PRINT, DISABLE);
		menu_ienable(manager->menu, MENUWIN_SAVE, DISABLE);
		menu_ienable(manager->menu, MENUWIN_APPEND, DISABLE);
		menu_ienable(manager->menu, MENUWIN_APPEND_CURRENT, DISABLE);
		menu_ienable(manager->menu, MENUWIN_MERGE, DISABLE);
		menu_ienable(manager->menu, MENUWIN_ALL_SLOTS, DISABLE);
		menu_ienable(manager->menu, MENUWIN_ALL_CLIPS, DISABLE);
		menu_ienable(manager->menu, MENUWIN_DELETE, DISABLE);
		menu_ienable(manager->menu, MENUWIN_DICT, DISABLE);

		manager->tree[MANAGER_VIEW].ob_state |= OS_DISABLED;
		manager->tree[MANAGER_SAVE].ob_state |= OS_DISABLED;

		/* Popup menu. */
		popup[RMB_MENU_CURRENT].ob_state |= OS_DISABLED;
		popup[RMB_MENU_DELETE].ob_state |= OS_DISABLED;
		popup[RMB_MENU_MERGE].ob_state |= OS_DISABLED;
	}
	else // 1 or more clip(s) selected.
	{
		manager->tree[MANAGER_CURRENT].ob_state &= ~OS_DISABLED;
		manager->tree[MANAGER_DELETE].ob_state &= ~OS_DISABLED;
		menu_ienable(manager->menu, MENUWIN_CURRENT, ENABLE);
		menu_ienable(manager->menu, MENUWIN_DELETE, ENABLE);
		menu_ienable(manager->menu, MENUWIN_MERGE, ENABLE);

		/* Popup menu. */
		popup[RMB_MENU_CURRENT].ob_state &= ~OS_DISABLED;
		popup[RMB_MENU_DELETE].ob_state &= ~OS_DISABLED;

		/* No merging possible if only 1 clip selected. */
		if (manager->selected == 1)
		{
			menu_ienable(manager->menu, MENUWIN_MERGE, DISABLE);
		}
		else if (manager->selected > 1)
		{
			menu_ienable(manager->menu, MENUWIN_MERGE, ENABLE);
			popup[RMB_MENU_MERGE].ob_state &= ~OS_DISABLED;
		}
	}
	
	if ((manager->window->open == TRUE) && (manager->window->icon == FALSE)) {
		redraw_object(manager->window->handle, manager->tree, MANAGER_TOOLBAR);
	}
}

/*
 * If there are no scrap.txt or clips at all, completely clear the Manager window and disable objects.
 * This is written from top to bottom (infoline -> preview box). Since this function will be called
 * every second until a new scrap.txt is created, we take care not to do anything undeeded.
 */
void clear_manager(GEMCLIP *manager)
{
	int i	= 0;
	int sel = FALSE;

	if (manager->nclips != 0)
	{
		manager->nclips = 0;
		manager->selected = 0;
		manager->fullsize = 0;
		manager->clipn = NOSCRAP;
		update_clips_infoline(manager);
	}

	if ( ! (manager->menu[MENUWIN_CURRENT].ob_state & OS_DISABLED))
	{
		menu_ienable(manager->menu, MENUWIN_CURRENT, DISABLE);
		menu_ienable(manager->menu, MENUWIN_VIEW, DISABLE);
		menu_ienable(manager->menu, MENUWIN_INFO, DISABLE);
		//menu_ienable(manager->menu, MENUWIN_PRINT, DISABLE);
		menu_ienable(manager->menu, MENUWIN_SAVE, DISABLE);
		menu_ienable(manager->menu, MENUWIN_APPEND, DISABLE);
		menu_ienable(manager->menu, MENUWIN_APPEND_CURRENT, DISABLE);
		menu_ienable(manager->menu, MENUWIN_MERGE, DISABLE);
		menu_ienable(manager->menu, MENUWIN_ALL_SLOTS, DISABLE);
		menu_ienable(manager->menu, MENUWIN_ALL_CLIPS, DISABLE);
		menu_ienable(manager->menu, MENUWIN_DELETE, DISABLE);
	}

	if ( ! (manager->tree[MANAGER_CURRENT].ob_state & OS_DISABLED))
	{
		change_state(manager->window->handle, manager->tree, MANAGER_CLIP_ICON, OS_DISABLED);
		change_state(manager->window->handle, manager->tree, MANAGER_CURRENT, OS_DISABLED);
		change_state(manager->window->handle, manager->tree, MANAGER_VIEW, OS_DISABLED);
		change_state(manager->window->handle, manager->tree, MANAGER_DELETE, OS_DISABLED);
	}

	/* Clear scrap.txt preview. */
	if (*manager->tree[MANAGER_CLIPBOARD].ob_spec.tedinfo->te_ptext != '\0')
	{
		*manager->tree[MANAGER_CLIPBOARD].ob_spec.tedinfo->te_ptext = '\0';

		if (manager->window->open == TRUE)
		{
			redraw_object(manager->window->handle, manager->tree, MANAGER_CLIPBOARD);
		}
	}

	/* Disable left side clip indices. */
	if ( ! (manager->tree[MANAGER_CLIP_N01].ob_state & OS_DISABLED))
	{
		for (i = MANAGER_CLIP_N01; i <= MANAGER_CLIP_N12; i++) {
			manager->tree[i].ob_state |= OS_DISABLED;
		}

		if (manager->window->open == TRUE)
		{
			redraw_object(manager->window->handle, manager->tree, MANAGER_CLIP_NO);
		}
	}

	/* Clear listbox. We only need to check the 1st item. */
	if (*manager->tree[MANAGER_CLIP01].ob_spec.tedinfo->te_ptext != '\0')
	{
		for (i = MANAGER_CLIP01; i <= MANAGER_CLIP12; i++) {
			*manager->tree[i].ob_spec.tedinfo->te_ptext = '\0';
		}

		if (manager->window->open == TRUE)
		{
			redraw_object(manager->window->handle, manager->tree, MANAGER_LISTBOX);
		}
	}

	/* Disable any selected clips. To avoid flicker, check 1st that there are selected slots. */
	for (i = MANAGER_CLIP01; i <= MANAGER_CLIP12; i++)
	{
		if (manager->tree[i].ob_state & OS_SELECTED)
		{
			sel = TRUE;
			break;
		}
	}

	if (sel == TRUE)
	{
		for (i = MANAGER_CLIP01; i <= MANAGER_CLIP12; i++)
		{
			if (manager->tree[i].ob_state & OS_SELECTED) {
				manager->tree[i].ob_state &= ~OS_SELECTED;
			}

			if (manager->window->open == TRUE) {
				redraw_object(manager->window->handle, manager->tree, i);
			}
		}
	}

	/* Clear preview box. */
	if (*manager->tree[MANAGER_PREV1].ob_spec.tedinfo->te_ptext != '\0')
	{
		for (i = MANAGER_PREV1; i <= MANAGER_PREV6; i++)
		{
			*manager->tree[i].ob_spec.tedinfo->te_ptext = '\0';
		}

		if (manager->window->open == TRUE)
		{
			redraw_object(manager->window->handle, manager->tree, MANAGER_PREVIEW_BOX);
		}
	}
}

/*
 * Resize the Manager window according to the preset in the popup menu. Take care of correcting for
 * some oversized objects. Returns the new window height.
 */
int resize_manager(GEMCLIP *manager, int item, USERCONF *conf)
{
	static int clipn = -1;   // Reset manager->clipn after switching from viewmodes that disable it.
	static int selected = 0; // Reset manager->selected after switching from viewmodes that disable it.
	GRECT rect;
	wind_get_grect(manager->window->handle, WF_CURRXYWH, &rect);

	/* We need to set the window to the full height to make correct computations below. */
	rect.g_h = manager->height;

	if (item == POP_VIEWMODE_STANDARD)
	{
		if (conf->clipmode == LIST_TEXT) {
			manager->tree[MANAGER_PREVIEW].ob_state |= OS_SELECTED;
		}
		else if (conf->clipmode == LIST_INFO) {
			manager->tree[MANAGER_PROPERTIES].ob_state |= OS_SELECTED;
		}

		manager->tree[MANAGER_PREVIEW].ob_state     &= ~OS_DISABLED;
		manager->tree[MANAGER_PROPERTIES].ob_state  &= ~OS_DISABLED;
		manager->tree[MANAGER_INFO_BOX].ob_flags    &= ~OF_HIDETREE;
		manager->tree[MANAGER_CLIP_ICON].ob_flags	&= ~OF_HIDETREE;
		manager->tree[MANAGER_CLIPBOARD].ob_flags	&= ~OF_HIDETREE;
		manager->tree[MANAGER_PREVIEW_BOX].ob_flags &= ~OF_HIDETREE;
		manager->tree[MANAGER_LISTBOX].ob_flags 	&= ~OF_HIDETREE;
		manager->tree[MANAGER_UPARROW].ob_flags		&= ~OF_HIDETREE;
		manager->tree[MANAGER_SCROLLBOX].ob_flags	&= ~OF_HIDETREE;
		manager->tree[MANAGER_CLIP_NO].ob_flags 	&= ~OF_HIDETREE;
		manager->tree[MANAGER_CURRENT].ob_state 	&= ~OS_DISABLED;
		manager->tree[MANAGER_DELETE].ob_state		&= ~OS_DISABLED;
		menu_ienable(manager->menu, MENUWIN_CURRENT, ENABLE);
		menu_ienable(manager->menu, MENUWIN_MERGE, ENABLE);
		menu_ienable(manager->menu, MENUWIN_ALL_SLOTS, ENABLE);
		menu_ienable(manager->menu, MENUWIN_ALL_CLIPS, ENABLE);
		menu_ienable(manager->menu, MENUWIN_DELETE, ENABLE);

		/* Reposition preview box y coord to its original place. */
		manager->tree[MANAGER_PREVIEW_BOX].ob_y = manager->tree[ROOT].ob_height
												- manager->tree[MANAGER_PREVIEW_BOX].ob_height
												- 9;
												
		if (clipn != -1)
		{
			manager->selected = selected;
			manager->clipn = clipn;
		}
	}
	else if (item == POP_VIEWMODE_MINIMAL)
	{
		int thickness = manager->tree[MANAGER_CLIPBOARD].ob_spec.tedinfo->te_thickness;

		/* Modify position of objects. We add some pixels since actual dimensions may be
		   larger than the object dimensions. We never change the root box dimensions. */
		rect.g_h = manager->widgeth + manager->tree[MANAGER_CURRENT].ob_height
									+ manager->tree[MANAGER_CLIPBOARD].ob_height
									+ thickness
									+ 24 + 4; // (3*8 px) + extra width of boxes.

		if (conf->clipmode == LIST_TEXT) {
			manager->tree[MANAGER_PREVIEW].ob_state |= OS_SELECTED;
		}
		else if (conf->clipmode == LIST_INFO) {
			manager->tree[MANAGER_PROPERTIES].ob_state |= OS_SELECTED;
		}

		manager->tree[MANAGER_PREVIEW].ob_state     &= ~OS_DISABLED;
		manager->tree[MANAGER_PROPERTIES].ob_state  &= ~OS_DISABLED;
		manager->tree[MANAGER_CLIP_ICON].ob_flags  &= ~OF_HIDETREE;
		manager->tree[MANAGER_CLIPBOARD].ob_flags  &= ~OF_HIDETREE;
		manager->tree[MANAGER_INFO_BOX].ob_flags    |= OF_HIDETREE;
		manager->tree[MANAGER_PREVIEW_BOX].ob_flags |= OF_HIDETREE;
		manager->tree[MANAGER_LISTBOX].ob_flags     |= OF_HIDETREE;
		manager->tree[MANAGER_CURRENT].ob_state     |= OS_DISABLED;
		manager->tree[MANAGER_DELETE].ob_state	    |= OS_DISABLED;

		menu_ienable(manager->menu, MENUWIN_CURRENT, DISABLE);
		menu_ienable(manager->menu, MENUWIN_MERGE, DISABLE);
		menu_ienable(manager->menu, MENUWIN_ALL_SLOTS, DISABLE);
		menu_ienable(manager->menu, MENUWIN_ALL_CLIPS, DISABLE);
		menu_ienable(manager->menu, MENUWIN_DELETE, DISABLE);
		
		selected = manager->selected;
		manager->selected = 0;
		clipn = manager->clipn;
		manager->clipn = SCRAP;
	}
	else if (item == POP_VIEWMODE_NOPREVIEW)
	{
		rect.g_h -= manager->tree[MANAGER_PREVIEW_BOX].ob_height + 8 + 1;

		if (conf->clipmode == LIST_TEXT) {
			manager->tree[MANAGER_PREVIEW].ob_state |= OS_SELECTED;
		}
		else if (conf->clipmode == LIST_INFO) {
			manager->tree[MANAGER_PROPERTIES].ob_state |= OS_SELECTED;
		}

		manager->tree[MANAGER_PREVIEW].ob_state     &= ~OS_DISABLED;
		manager->tree[MANAGER_PROPERTIES].ob_state  &= ~OS_DISABLED;
		manager->tree[MANAGER_INFO_BOX].ob_flags    &= ~OF_HIDETREE;
		manager->tree[MANAGER_PREVIEW_BOX].ob_flags |= OF_HIDETREE;
		manager->tree[MANAGER_CLIPBOARD].ob_flags	&= ~OF_HIDETREE;
		manager->tree[MANAGER_CLIP_ICON].ob_flags	&= ~OF_HIDETREE;
		manager->tree[MANAGER_LISTBOX].ob_flags 	&= ~OF_HIDETREE;
		manager->tree[MANAGER_SCROLLBOX].ob_flags	&= ~OF_HIDETREE;
		manager->tree[MANAGER_UPARROW].ob_flags		&= ~OF_HIDETREE;
		manager->tree[MANAGER_CLIP_NO].ob_flags 	&= ~OF_HIDETREE;
		manager->tree[MANAGER_CURRENT].ob_state 	&= ~OS_DISABLED;

		if (clipn != -1)
		{
			manager->clipn = clipn;
			manager->selected = selected;
		}
	}
	else if (item == POP_VIEWMODE_NOLIST)
	{
		rect.g_h -= manager->tree[MANAGER_LISTBOX].ob_height
				  + manager->tree[MANAGER_CLIPBOARD].ob_height + 16 + 6;

		/* Set preview box y coord just below the toolbar (instead of the clipboard box. */
		manager->tree[MANAGER_PREVIEW_BOX].ob_y = manager->tree[MANAGER_CLIPBOARD].ob_y;

		/* Since the list box is gone, the top preview box is redundant. */
		manager->tree[MANAGER_PREVIEW].ob_state      |= OS_DISABLED;
		manager->tree[MANAGER_PREVIEW].ob_state     &= ~OS_SELECTED;
		manager->tree[MANAGER_PROPERTIES].ob_state   |= OS_DISABLED;
		manager->tree[MANAGER_PROPERTIES].ob_state  &= ~OS_SELECTED;
		manager->tree[MANAGER_INFO_BOX].ob_flags     |= OF_HIDETREE;
		manager->tree[MANAGER_CLIP_ICON].ob_flags	 |= OF_HIDETREE;
		manager->tree[MANAGER_CLIPBOARD].ob_flags	 |= OF_HIDETREE;
		manager->tree[MANAGER_LISTBOX].ob_flags 	 |= OF_HIDETREE;
		manager->tree[MANAGER_SCROLLBOX].ob_flags	 |= OF_HIDETREE;
		manager->tree[MANAGER_UPARROW].ob_flags		 |= OF_HIDETREE;
		manager->tree[MANAGER_CLIP_NO].ob_flags 	 |= OF_HIDETREE;
		manager->tree[MANAGER_PREVIEW_BOX].ob_flags &= ~OF_HIDETREE;
		manager->tree[MANAGER_CURRENT].ob_state 	&= ~OS_DISABLED;

		/* Deselect any selected clip. */
		manager->selected = 1;
		select_all_clips(manager, MENUWIN_ALL_CLIPS);

		manager->selected = 0;
		clipn = manager->clipn;
		manager->clipn = SCRAP;
	}
	wind_set_grect(manager->window->handle, WF_CURRXYWH, &rect);
	
	/* Rewrite the infoline to (hopefully) fix a redraw glitch with MagiC. */
	wind_set_str(manager->window->handle, WF_INFO, manager->window->info);
	send_redraw_message(manager->window->handle);

	return rect.g_h;
}

/*
 * Update the inconified window with the preview of scrap.txt so that users always have a visual clue|
 * of what will be pasted. The display is centered within the iconified window.
 */
void update_icon_preview(GEMCLIP *manager)
{
	char empty[MBUFFER] = {0};
	char *buffer = NULL;
	short txtlen = 0;
	int pos 	 = 0;
	int len      = 0;
	GRECT rect;
	OBJECT *tree = get_gem_tree(ICONWIN);

	wind_get_grect(manager->window->handle, WF_WORKXYWH, &rect);
	tree[ICONWIN_BACKGRND].ob_x = (rect.g_w / 2) - (tree[ICONWIN_BACKGRND].ob_width / 2);
	tree[ICONWIN_BACKGRND].ob_y = (rect.g_h / 2) - (tree[ICONWIN_BACKGRND].ob_height / 2);
	tree[ROOT].ob_x = rect.g_x;
	tree[ROOT].ob_y = rect.g_y;

	if ((manager->clipn == NOSCRAP) && (manager->nclips == 0))
	{
		snprintf(empty, sizeof empty, "%s", get_gstring(STRINGS, STRINGS_EMPTY));
		buffer = empty;
		tree[ICONWIN_ICON].ob_state |= OS_SELECTED; // Will display the 'empty' picture.
	}
	else if (manager->clipn == NOSCRAP)
	{
		strcpy(empty, """");
		buffer = empty;
	}
	else {
		buffer = manager->clips[SCRAP]->preview;
	}

	if (buffer != NULL)
	{
		len = strlen(buffer) + 1;

		/* These object in the resource MUST be in incremental order. */
		for (int i = ICONWIN_PREV1; i <= ICONWIN_PREV4; i++)
		{
			/* Not all strings are the same size, so get size of each one. */
			txtlen = tree[i].ob_spec.tedinfo->te_txtlen;

			if (pos <= len) { // Do not go past the end of the preview.
				snprintf(tree[i].ob_spec.tedinfo->te_ptext, txtlen, "%s", &buffer[pos]);
			}
			else {
				*tree[i].ob_spec.tedinfo->te_ptext = '\0';
			}
			pos += txtlen - 1;
		}
	}

	snprintf(tree[ICONWIN_SIZE].ob_spec.tedinfo->te_ptext,
			 tree[ICONWIN_SIZE].ob_spec.tedinfo->te_txtlen,
			 "%ld %s",
			 manager->clips[SCRAP]->size,
			 get_gstring(STRINGS, STRINGS_BYTES));

	objc_wdraw(tree, ROOT, MAX_DEPTH, &rect, manager->window->handle);
}
