/*
 * GEMClip, a Clipboard Manager for FreeMiNT and MagiC.
 *
 * Copyright (C) 2011 Jean-Franois Lemaire (jflemaire@skynet.be)
 *
 * This file is part of GEMClip.
 *
 * This program is free software; redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "general.h"
#include "clips.h"
#include "manager.h"
#include "scroll.h"
#include "utils.h"

#define SCROLLBAR (manager->tree[MANAGER_SCROLLBAR].ob_height + 4)
#define SLIDER    (manager->tree[object].ob_height + 4)

/*
 * Handle scrolling of the listbox with the up/down arrow objects.
 */
void scroll_listbox_arrows(WINDOW *window, short message, OBJECT *tree)
{
	enum { HEIGHT = 17 }; // Height of a list box slot (16 for G_FTEXT + 1 for interline).

	short pxy[8]  = {0};
	MFDB src	  = {NULL, 0, 0, 0, 0, 0, 0, 0, 0};
	MFDB dst	  = {NULL, 0, 0, 0, 0, 0, 0, 0, 0};
	GRECT wrect   = {0, 0, 0, 0};
	GRECT rect	  = {0, 0, 0, 0};

	objc_offset(tree, MANAGER_LISTBOX, &wrect.g_x, &wrect.g_y);
	wrect.g_w = tree[MANAGER_LISTBOX].ob_width;
	wrect.g_h = tree[MANAGER_LISTBOX].ob_height;

	/* Remove 1 pixel from all 4 sides of the background box. */
	pxy[0] = pxy[4] = wrect.g_x + 1;
	pxy[2] = pxy[6] = wrect.g_x + wrect.g_w - 1;

	switch (message)
	{
	case WA_UPLINE:
		pxy[1] = wrect.g_y;
		pxy[3] = wrect.g_y + wrect.g_h - HEIGHT;
		pxy[5] = wrect.g_y + HEIGHT;
		pxy[7] = wrect.g_y + wrect.g_h;
		break;
	case WA_DNLINE:
		pxy[1] = wrect.g_y + HEIGHT + 1;
		pxy[3] = wrect.g_y + wrect.g_h - 1;
		pxy[5] = wrect.g_y + 1;
		pxy[7] = wrect.g_y + wrect.g_h - HEIGHT -1;
		break;
	}

	wind_get_grect(window->handle, WF_FIRSTXYWH, &rect);

	while (rect.g_w && rect.g_h)
	{
		if (rc_intersect(&wrect, &rect) != 0) {
			vro_cpyfm(window->wks, S_ONLY, pxy, &src, &dst);
		}

		wind_get_grect(window->handle, WF_NEXTXYWH, &rect);
	}

	/* Redraw top or bottom slot. */
	if (message == WA_UPLINE) {
		redraw_object(window->handle, tree, MANAGER_CLIP01);
	}
	else if (message == WA_DNLINE) {
		redraw_object(window->handle, tree, MANAGER_CLIP12);
	}
}

/*
 * Adapt scroll bar slider. We account for the extra thickness of some objects.
 * The slider y pos is 0 when at the top.  Since topclip is at least 1, we must
 * retract -1 to have the slider at 0 when topclip == 1.
 */
void adapt_slider(WINDOW *window, OBJECT *tree, int nclips, int topclip)
{
	int scrollbar = (tree[MANAGER_SCROLLBAR].ob_height + 4);
	int nclip     = topclip - 1;

	if (nclips <= 12)
	{
		tree[MANAGER_UPARROW].ob_state |= OS_DISABLED;
		tree[MANAGER_DNARROW].ob_state |= OS_DISABLED;
		tree[MANAGER_SLIDER].ob_state |= OS_DISABLED;
		tree[MANAGER_SCROLLBAR].ob_state |= OS_DISABLED;
		tree[MANAGER_SLIDER].ob_height = scrollbar - 4;
		tree[MANAGER_SLIDER].ob_y = 0;
	}
	else if (nclips == 99) // Max value means slider is at its smallest.
	{
		tree[MANAGER_SLIDER].ob_height = 16;

		if (topclip + NSLOTS == nclips) { // If at the bottom, set y pos relative to scrollbar.
			tree[MANAGER_SLIDER].ob_y = scrollbar - tree[MANAGER_SLIDER].ob_height - 4;
		}
		else {
			tree[MANAGER_SLIDER].ob_y = (nclip * 99) / (nclips - NSLOTS+1);
		}
	}
	else // Most cases.
	{
		tree[MANAGER_UPARROW].ob_state &= ~OS_DISABLED;
		tree[MANAGER_DNARROW].ob_state &= ~OS_DISABLED;
		tree[MANAGER_SLIDER].ob_state &= ~OS_DISABLED;
		tree[MANAGER_SCROLLBAR].ob_state &= ~OS_DISABLED;

		tree[MANAGER_SLIDER].ob_height = (int)((float)scrollbar / (float)nclips * (float)NSLOTS+1);

		if (topclip + NSLOTS == nclips) { // If at the bottom, set y pos relative to scrollbar.
			tree[MANAGER_SLIDER].ob_y = scrollbar - tree[MANAGER_SLIDER].ob_height - 4;
		}
		else {
			tree[MANAGER_SLIDER].ob_y = (nclip * 99) / (nclips - NSLOTS+1);
		}
	}
	redraw_object(window->handle, tree, MANAGER_SCROLLBOX);
	redraw_object(window->handle, tree, MANAGER_UPARROW);
	redraw_object(window->handle, tree, MANAGER_DNARROW);
}

/*
 * Account for the extra thickness of some objects
 */
void scroll_slider(GEMCLIP *manager, EVMULT_IN *evmult_in, short object)
{
	short x       = 0;
	short y       = 0;
	short my      = 0;
	short prevy   = 0;
	short mbutton = 0;
	short foo     = 0;
	short flags   = evmult_in->emi_flags;
	int pos       = 0;
	int topclip   = manager->topclip;

	graf_mouse(10, NULL);
	graf_mkstate(&foo, &prevy, &foo, &foo); // Get y position of mouse cursor before scrolling.

	manager->tree[object].ob_state |= OS_SELECTED;
	redraw_object(manager->window->handle, manager->tree, MANAGER_SCROLLBOX);

	evmult_in->emi_flags = MU_TIMER;
	evmult_in->emi_tlow = 100;

	do
	{
		graf_mkstate(&foo, &my, &mbutton, &foo);

		/* Did we actually scroll? */
		if (my != prevy)
		{
			objc_offset(manager->tree, object, &x, &y);
			y += 10; // Account for center of object.

			manager->tree[object].ob_y += my - y;

			if (manager->tree[object].ob_y < 0) {
				manager->tree[object].ob_y = 0;
			}

			if (manager->tree[object].ob_y > (SCROLLBAR - SLIDER)) {
				manager->tree[object].ob_y = SCROLLBAR - SLIDER;
			}

			/* Get the position as a value in the range 0-1000. */
			pos = (int)(((float)manager->tree[object].ob_y / ((float)SCROLLBAR - (float)SLIDER)) * (float)1000);
			pos *= (int)((float)SCROLLBAR - (float)SLIDER);
			pos /= 1000;

			/* Compute topclip index and do some adjustments. */
			manager->topclip = (int)((((float)pos + (float)10) / (float)SCROLLBAR) * (float)manager->nclips);

			if (manager->topclip < 1) { // Too high.
				manager->topclip = 1;
			}
			else if (manager->topclip > manager->nclips - NSLOTS) { //Too low.
				manager->topclip = manager->nclips - NSLOTS;
			}

			if (manager->topclip != topclip) // Did we actually scroll?
			{
				update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
				update_listbox_slots(manager);
				update_listbox(manager);
				redraw_object(manager->window->handle, manager->tree, MANAGER_LISTBOX);
				topclip = manager->topclip;
			}
			prevy = 0; // Reset value to avoid not scrolling in case the mouse returns to its original pos.
		}
	} while (mbutton & 0x01); // While button mouse depressed.

	/* Reset normal events and time values. */
	evmult_in->emi_flags = flags;
	evmult_in->emi_tlow = 1000;
	manager->tree[object].ob_state &= ~OS_SELECTED;
	redraw_object(manager->window->handle, manager->tree, MANAGER_SCROLLBOX);
	graf_mouse(ARROW, NULL);
}

/*
 * Handle scrolling up by one slot in the listbox.
 */
void scroll_uparrow(GEMCLIP *manager, EVMULT_OUT *evmult_out)
{
	if (manager->topclip > 1)
	{
		if (evmult_out->emo_mclicks == 2) { // Double-click.
			manager->topclip = 1; // At the top.
		}
		else {
			manager->topclip--;
		}
		strcpy(manager->tree[MANAGER_CLIP01].ob_spec.tedinfo->te_ptext,
			   manager->clips[manager->topclip]->preview);
		update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
		update_listbox_slots(manager);
		update_listbox(manager);
		scroll_listbox_arrows(manager->window, WA_UPLINE, manager->tree);
	}
}

/*
 * Handle scrolling down by one slot in the listbox.
 */
void scroll_dnarrow(GEMCLIP *manager, EVMULT_OUT *evmult_out)
{
	if ((manager->topclip + NSLOTS) < manager->nclips) // If not already at the bottom.
	{
		if (evmult_out->emo_mclicks == 2) { // Double-click.
			manager->topclip = manager->nclips - NSLOTS; // At the bottom.
		}
		else {
			manager->topclip++;
		}
		strcpy(manager->tree[MANAGER_CLIP12].ob_spec.tedinfo->te_ptext,
			   manager->clips[manager->topclip + NSLOTS]->preview);
		update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
		update_listbox_slots(manager);
		update_listbox(manager);
		scroll_listbox_arrows(manager->window, WA_DNLINE, manager->tree);
	}
}

/*
 * Move up/down a page after a click in the scrollbar. If the slider is currently above the clicked
 * position, page down, otherwise page up.
 */
void scroll_page(GEMCLIP *manager, EVMULT_OUT *evmult_out)
{
	short x = 0;
	short y = 0;

	objc_offset(manager->tree, MANAGER_SLIDER, &x, &y);
	y += manager->tree[MANAGER_SLIDER].ob_height;

	if (evmult_out->emo_mbutton & LEFT_BUTTON)
	{
		if (y < evmult_out->emo_mouse.p_y) // Page down.
		{
			if (manager->topclip + NSLOTS < manager->nclips)
			{
				/* Double click or go to bottom. */
				if ((evmult_out->emo_mclicks == 2) || (manager->topclip + NSLOTS + 1 >= manager->nclips - NSLOTS)) {
					manager->topclip = manager->nclips - NSLOTS; // At the bottom.
				}
				else if (manager->topclip + NSLOTS < manager->nclips - NSLOTS) {
					manager->topclip += NSLOTS;
				}
			}
		}
		else // Page up.
		{
			if (manager->topclip != 1)
			{
				/* Double click or go to top. */
				if ((evmult_out->emo_mclicks == 2) || (manager->topclip <= NSLOTS + 2)) {
					manager->topclip = 1; // At the top.
				}
				else if (manager->topclip - NSLOTS > 1) {
					manager->topclip -= NSLOTS;
				}
			}
		}
	}
	else // Reverse behaviour.
	{
		if (y < evmult_out->emo_mouse.p_y) // Page up.
		{
			if (manager->topclip != 1)
			{
				/* Double click or go to top. */
				if ((evmult_out->emo_mclicks == 2) || (manager->topclip <= NSLOTS + 2)) {
					manager->topclip = 1; // At the top.
				}
				else if (manager->topclip - NSLOTS > 1) {
					manager->topclip -= NSLOTS;
				}
			}
		}
		else // Page down.
		{
			if (manager->topclip + NSLOTS < manager->nclips)
			{
				/* Double click or go to bottom. */
				if ((evmult_out->emo_mclicks == 2) || (manager->topclip + NSLOTS + 1 >= manager->nclips - NSLOTS)) {
					manager->topclip = manager->nclips - NSLOTS; // At the bottom.
				}
				else if (manager->topclip + NSLOTS < manager->nclips - NSLOTS) {
					manager->topclip += NSLOTS;
				}
			}
		}
	}
}

/*
 * Circumvent XaAES's keyboard navigation for keyboard moving/selecting within the list box.
 */
void scroll_keyboard(GEMCLIP *manager, EVMULT_OUT *evmult_out)
{
	int clipn = SCRAP;
	int moved = TRUE; // Avoid flicker by not redrawing if we were already at the top/bottom.
	static short obj = MANAGER_CLIP01;
	
	if ((clipn = find_first_selected(manager)) > SCRAP)
	{
		// Find associated list box slot.
	}
	else {
		obj = MANAGER_CLIP01;
	}

	switch (evmult_out->emo_kreturn)
	{
	case SC_CLRHOME:
		if (manager->topclip > 1)
		{
			manager->topclip = 1;
			strcpy(manager->tree[MANAGER_CLIP01].ob_spec.tedinfo->te_ptext,
				   manager->clips[manager->topclip]->preview);
			update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
			update_listbox_slots(manager);
			update_listbox(manager);
			scroll_listbox_arrows(manager->window, WA_UPLINE, manager->tree);
		}
		else {
			moved = FALSE;
		}
		break;
	case SC_SHFT_CLRHOME:
		if ((manager->topclip + NSLOTS) < manager->nclips) // If not already at the bottom.
		{
			manager->topclip = manager->nclips - NSLOTS; // At the bottom.
			strcpy(manager->tree[MANAGER_CLIP12].ob_spec.tedinfo->te_ptext,
				   manager->clips[manager->topclip + NSLOTS]->preview);
			update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
			update_listbox_slots(manager);
			update_listbox(manager);
			scroll_listbox_arrows(manager->window, WA_DNLINE, manager->tree);
		}
		break;
	case SC_UPARROW:
		if (selection(manager->clips) == FALSE) {
			obj = MANAGER_CLIP12;
		}
		else
		{
			if (evmult_out->emo_kmeta & K_CTRL) {
				evmult_out->emo_kmeta |= (K_LSHIFT | K_RSHIFT); // Force multiple selection in handle_selection().
			}

			if (obj > MANAGER_CLIP01) {
				obj--;
			}
			else if (manager->topclip > 1)
			{
				manager->topclip--;
				strcpy(manager->tree[MANAGER_CLIP01].ob_spec.tedinfo->te_ptext,
					   manager->clips[manager->topclip]->preview);

				update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
				update_listbox_slots(manager);
				update_listbox(manager);
				scroll_listbox_arrows(manager->window, WA_UPLINE, manager->tree);
			}
			else {
				moved = FALSE;
			}
		}
		break;
	case SC_DNARROW:
		if (selection(manager->clips) == FALSE) {
			obj = MANAGER_CLIP01;
		}
		else
		{
			if (evmult_out->emo_kmeta & K_CTRL) {
				evmult_out->emo_kmeta |= (K_LSHIFT | K_RSHIFT); // Force multiple selection in handle_selection().
			}

 			/* Either move selection or scroll if not already at the bottom. */
 			if (obj < MANAGER_CLIP12) {
				obj++;
			}
			else if ((manager->topclip + NSLOTS) < manager->nclips)
			{
				manager->topclip++;
				strcpy(manager->tree[MANAGER_CLIP12].ob_spec.tedinfo->te_ptext,
					   manager->clips[manager->topclip + NSLOTS]->preview);
				update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
				update_listbox_slots(manager);
				update_listbox(manager);
				scroll_listbox_arrows(manager->window, WA_DNLINE, manager->tree);
			}
			else {
				moved = FALSE;
			}
		}
		break;
	case SC_SHFT_UPARROW:
 		/* Force a page up. */
		if (selection(manager->clips) == FALSE) {
			obj = 0;
		}
 		evmult_out->emo_mbutton |= LEFT_BUTTON;
		evmult_out->emo_mouse.p_y = 0;

		if (manager->topclip > 1)
		{
			scroll_page(manager, evmult_out);
			update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
			update_listbox_slots(manager);
			update_listbox(manager);
			redraw_object(manager->window->handle, manager->tree, MANAGER_LISTBOX);
		}
		else {
			moved = FALSE; // Avoids flickering of the preview box.
		}
		break;
	case SC_SHFT_DNARROW:
 		/* Force a page down. */
		if (selection(manager->clips) == FALSE) {
			obj = 0;
		}
 		evmult_out->emo_mbutton |= LEFT_BUTTON;
		evmult_out->emo_mouse.p_y = 9999;

		if ((manager->topclip + NSLOTS) < manager->nclips) // If not already at the bottom.
		{
			scroll_page(manager, evmult_out);
			update_indices(manager->clips, manager->topclip, manager->tree, manager->window->handle);
			update_listbox_slots(manager);
			update_listbox(manager);
			redraw_object(manager->window->handle, manager->tree, MANAGER_LISTBOX);
		}
		else {
			moved = FALSE;
		}
		break;
	default:
		moved = FALSE; // Avoid redrawing for a Left/Right arrow event.
		break;
	}

	if ((moved == TRUE)	                             &&
	  (!(manager->tree[obj].ob_flags & OS_DISABLED)) &&
	    (evmult_out->emo_kreturn != SC_CLRHOME)      &&
	    (evmult_out->emo_kreturn != SC_SHFT_CLRHOME) &&
	    (evmult_out->emo_kreturn != SC_SHFT_UPARROW) &&
	    (evmult_out->emo_kreturn != SC_SHFT_DNARROW) &&
	    (obj != 0))
	{
		manager->tree[obj].ob_state |= OS_SELECTED;
		manager->clipn = handle_selection(manager, obj, evmult_out);
		update_objects(manager);
		update_clips_infoline(manager);
		update_preview(manager);
	
		if (evmult_out->emo_kmeta & K_CTRL) {
			redraw_object(manager->window->handle, manager->tree, MANAGER_LISTBOX);
		}
	}
}
